/*
Program for generating the mean X, Y, and Z coordinates of a
single mass or a series of masses with the color 4 in the
someRandomColor.txt file to be specified in the command line.
Can be used to genreate MSD curves.
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sstream>
#include <fstream>
#include <vector>
#include <ctype.h>
#include "chromoTypes.h"
#include "analysis.h"

typedef std::vector<double> POINT;
typedef std::vector<POINT>  POINT_LINE;

static double sqr(double a) { 
	return (a)*(a); 
}

static double dist2(const POINT &a, const POINT &b)
{
	double dx = a[0] - b[0];
	double dy = a[1] - b[1];
	double dz = a[2] - b[2];
	return dx*dx + dy*dy + dz*dz;
}

static double dist(const POINT &a, const POINT &b)
{
	return sqrt(dist2(a, b));
}

// Find the mean x coord of
// the points at a given time.
static double find_mean_x_coord(const POINT_LINE &pts)
{
	// If we have no points, return zero.
	double mean = 0;
	if (pts.size() == 0) {
		return mean;
	}

	// Compute the mean distance of the all the points
	size_t i;
	for (i = 0; i < pts.size(); i++) {
		mean += (pts[i][0]);
	}
	mean /= pts.size();

	return mean;
}

// Find the mean y coord of
// the points at a given time.
static double find_mean_y_coord(const POINT_LINE &pts)
{
	// If we have no points, return zero.
	double mean = 0;
	if (pts.size() == 0) {
		return mean;
	}

	// Compute the mean distance of the all the points
	size_t i;
	for (i = 0; i < pts.size(); i++) {
		mean += (pts[i][1]);
	}
	mean /= pts.size();

	return mean;
}

// Find the mean z coord of
// the points at a given time.
static double find_mean_z_coord(const POINT_LINE &pts)
{
	// If we have no points, return zero.
	double mean = 0;
	if (pts.size() == 0) {
		return mean;
	}

	// Compute the mean distance of the all the points
	size_t i;
	for (i = 0; i < pts.size(); i++) {
		mean += (pts[i][2]);
	}
	mean /= pts.size();

	return mean;
}

static double find_mean_dist_from_Z(const POINT_LINE &pts)
{
	// If we have no points, return zero.
	double mean = 0;
	if (pts.size() == 0) {
		return mean;
	}

	// Compute the mean distance of the all the points
	size_t i;
	for (i = 0; i < pts.size(); i++) {
		mean += sqrt(pts[i][0] * pts[i][0] + pts[i][1] * pts[i][1]);
	}
	mean /= pts.size();

	return mean;
}

static double find_min_dist_from_Z(const POINT_LINE &pts)
{
	// If we have no points, return 1e6.
	double min = 1e6;
	if (pts.size() == 0) {
		return min;
	}

	// Compute the min distance of the all the points
	size_t i;
	for (i = 0; i < pts.size(); i++) {
		double dist = sqrt(pts[i][0] * pts[i][0] + pts[i][1] * pts[i][1]);
		if (dist < min) { min = dist; }
	}

	return min;
}

static double find_max_dist_from_Z(const POINT_LINE &pts)
{
	// If we have no points, return 0.
	double max = 0;
	if (pts.size() == 0) {
		return max;
	}

	// Compute the min distance of the all the points
	size_t i;
	for (i = 0; i < pts.size(); i++) {
		double dist = sqrt(pts[i][0] * pts[i][0] + pts[i][1] * pts[i][1]);
		if (dist > max) { max = dist; }
	}

	return max;
}

static double contour_length(const POINT_LINE &pts) {
	// If we don't have at least two points, return -1, which
	// indicates an error.
	if (pts.size() < 2) { return -1; }

	double length = 0;
	size_t i;
	for (i = 1; i < pts.size(); i++) {
		length += dist(pts[i - 1], pts[i]);
	}

	return length;
}

static double end_to_end_length(const POINT_LINE &pts) {
	// If we don't have at least two points, return -1, which
	// indicates an error.
	if (pts.size() < 2) { return -1; }

	return dist(pts[0], pts[pts.size() - 1]);
}

static double mean_square_displacement(const POINT_LINE &pts, int stride) {
	double sum = 0;
	int count = static_cast<int>(pts.size() - stride);
	for (int i = 0; i < count; i++) {
		for (int j = 0; j < 3; j++) {
			sum += sqr(pts[i][j] - pts[i + stride][j]);
		}
	}
	if (count > 0) {
		sum /= count;
	}

	return sum;
}

// compute value of the requested statistic for a set of points
static double computeStat(int statIndex, const POINT_LINE &pts) {
	switch (statIndex) {
	case 0: return find_mean_x_coord(pts);
	case 1: return find_mean_y_coord(pts);
	case 2: return find_mean_z_coord(pts);
	case 3: return find_mean_dist_from_Z(pts);
	case 4: return find_min_dist_from_Z(pts);
	case 5: return find_max_dist_from_Z(pts);
	case 6: return contour_length(pts);
	case 7: return end_to_end_length(pts);
	case 8: return mean_square_displacement(pts, 1);
	default: return -1;
	}
}

// functions to help read integers from csv
std::vector<int> &intsFromLine(const std::string &s, char delim, std::vector<int> &elems) {
	std::stringstream ss(s);
	std::string item;
	while (std::getline(ss, item, delim)) {
		elems.push_back(stoi(item));
	}
	return elems;
}

std::vector<int> intsFromLine(const std::string &s, char delim) {
	std::vector<int> elems;
	intsFromLine(s, delim, elems);
	return elems;
}

void Usage(const char *name)
{
	fprintf(stderr, "Usage: %s [-trim DIST] [-option s] [BeadFileName] [InputFileName] [OutFileDirectory]\n", name);
	fprintf(stderr, "       -trim: Trim points less than DIST from the center\n");
	fprintf(stderr, "       -meanx: output mean X positions for each bead\n");
	fprintf(stderr, "       -meany: output mean Y positions for each bead\n");
	fprintf(stderr, "       -meanz: output mean Z positions for each bead\n");
	fprintf(stderr, "       -meandist:\n");
	fprintf(stderr, "       -mindist:\n");
	fprintf(stderr, "       -maxdist:\n");
	fprintf(stderr, "       -contour:\n");
	fprintf(stderr, "       -endtoend:\n");
	fprintf(stderr, "       -rog:\n");
	fprintf(stderr, "       -msd: output mean square displacement (3D) for each bead\n");
	fprintf(stderr, "       BeadFileName: CSV File with numbers of beads to compute statistics for\n");
	fprintf(stderr, "       InputFileName: chromoShake output file with bead positions default stdin)\n");
	fprintf(stderr, "       OutFileDirectory: folder to put the output files in)\n");
	exit(-1);
}

int main(int argc, const char *argv[])
{
	const int numStats = 9;

	// statistics that may be computed
	bool getStat[numStats];
	for (int i = 0; i < numStats; i++) {
		getStat[i] = false;
	}

	// whether or not to output a summary file for each statistic
	bool getSum[numStats];
	for (int i = 0; i < numStats; i++) {
		getSum[i] = false;
	}

	std::string statName[numStats];
	statName[0] = "MeanX";
	statName[1] = "MeanY";
	statName[2] = "MeanZ";
	statName[3] = "MeanDist";
	statName[4] = "MinDist";
	statName[5] = "MaxDist";
	statName[6] = "ContourLength";
	statName[7] = "EndToEndLength";
	statName[8] = "MSD";

	// Bead file containing indicating beads of interest
	FILE *beadfile = NULL;
	const char *beadfile_name = "";

	// Get the input and output files (by default, just standard in and out)
	FILE *infile = stdin;
	const char *infile_name = "";
	char *outfile_dir = "";

	// If points are closer than this to the center, trim them from the
	// statistics.  This is because some models had colored the fixed
	// end-points of the DNA the same color as the cohesin rings.
	double trim_dist = 0.0;

	// Parse the command line.
	size_t i;
	unsigned real_params = 0;
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-trim")) {
			if (++i >= argc) { Usage(argv[0]); }
			trim_dist = atof(argv[i]);
		} else if (!strcmp(argv[i], "-meanx")) {
			getStat[0] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[0] = true; }
		} else if (!strcmp(argv[i], "-meany")) {
			getStat[1] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[1] = true; }
		} else if (!strcmp(argv[i], "-meanz")) {
			getStat[2] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[2] = true; }
		} else if (!strcmp(argv[i], "-meandist")) {
			getStat[3] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[3] = true; }
		} else if (!strcmp(argv[i], "-mindist")) {
			getStat[4] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[4] = true; }
		} else if (!strcmp(argv[i], "-maxdist")) {
			getStat[5] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[5] = true; }
		} else if (!strcmp(argv[i], "-contour")) {
			getStat[6] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[6] = true; }
		} else if (!strcmp(argv[i], "-endtoend")) {
			getStat[7] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[7] = true; }
		} else if (!strcmp(argv[i], "-msd")) {
			getStat[8] = true;
			if (++i > argc) { Usage(argv[0]); }
			if (!strcmp(argv[i], "s")) { getSum[8] = true; }
		} else if (argv[i][0] == '-') {
			Usage(argv[0]);
		}
		else switch (real_params++) {
		case 0:
			beadfile_name = argv[i];
			break;
		case 1:
			infile_name = argv[i];
			break;
		case 2:
			strcat(outfile_dir,argv[i]);
			strcat(outfile_dir, "/");
		default:
			Usage(argv[0]);
		}
	}

	// Open the input files.
	if (strlen(beadfile_name) == 0) {
		Usage(argv[0]);
	}
	else if (NULL == (beadfile = fopen(beadfile_name, "r"))) {
		perror("Cannot open color file for reading");
		exit(-2);
	}
	if ((strlen(infile_name) > 0) && (NULL == (infile = fopen(infile_name, "r")))) {
		perror("Cannot open input file for reading");
		exit(-2);
	}

	// Read the bead file. Each line gives either a single integer or 
	// comma-separate list of integers identifying beads to compute
	// statistics for at each timestep
	std::vector<std::vector<int>> beadLines;
	char    line[4096];
	line[sizeof(line) - 1] = '\0';    // Ensure null-termination
	
	while (fgets(line, sizeof(line) - 1, infile) != NULL) {
		std::string stdLine(line);
		std::vector<int> beadLine = intsFromLine(stdLine, ',');
		beadLines.push_back(beadLine);
	}

	// We repeat reading in blocks of points, each of which is preceded
	// by a line that says "Time T", where T is a floating-point number.
	// Then there are a number of lines of X Y Z which are floating-point
	// coordinates. Then there is a blank line after each block of points.
	// We only add points whose index is contained in the bead file
	
	
	// the times vector stores a list of "point lines" for each timestep
	// each of the 64 point lines contains one or more points corresponding
	// to the position of the bead indicated by the corresponding bead number
	// in the 64-line bead file. The beads are analyzed by line rather than
	// individually, which is why we group the points into lines rather than
	// a single long list
	std::vector<std::vector<POINT_LINE>>    times;

	// also have vector with the values of the time at each point
	std::vector< double > t_val;

	long int line_num = 0;
	bool foundCoordinates = false;
	while (true) {
		line_num++;
		// Read and parse the "Time" line.
		if (fgets(line, sizeof(line) - 1, infile) == NULL) {
			// Done reading lines, we got them all.
			break;
		}
		double t;
		if (sscanf(line, "Time %le", &t) <= 0) {
			if (foundCoordinates) {
				// If the line is empty, ignore it and move on.
				// Otherwise,we have a bad time line
				char c;
				int i = 0;
				while (line[i])
				{
					c = line[i];
					if (!isspace(c)) {
						fprintf(stderr, "Bad Time line %ld: %s\n", line_num, line);
						return(-1);
					}
					i++;
				}
			}
			else {
				// haven't arrived at the first timestep of coordinates
				continue;
			}
		}
		else if (!foundCoordinates) {
			foundCoordinates = true;
		}

		// save to t_val the current time
		t_val.push_back(t);

		// Read and parse point lines, adding them to our time.
		// Only add points that correspond to cohesin.
		std::vector<POINT_LINE> pointlines;
		times.push_back(pointlines);
		size_t point_count = 0;
		while (true) {
			// Read and parse the next line, which may be a points line.
			line_num++;
			if (fgets(line, sizeof(line) - 1, infile) == NULL) {
				// Done reading lines, we got them all.
				break;
			}
			double x, y, z;
			if (sscanf(line, "%le %le %le", &x, &y, &z) != 3) {
				// No more points for this line
				break;
			}

			// Trim points closer than the trim distance from the
			// Z axis
			if (trim_dist > sqrt(x*x + y*y)) {
				point_count++;
				continue;
			}

			for (int i = 0; i < beadLines.size(); i++) {
				if (std::find(beadLines[i].begin(), beadLines[i].end(), point_count) != beadLines[i].end()) {
					POINT p;
					p.push_back(x);
					p.push_back(y);
					p.push_back(z);
					times.back()[i].push_back(p);
					break;
				}
			}
			point_count++;
		}
	}
	fclose(infile);

	// Compute requested statistics
	for (int i = 0; i < beadLines.size(); i++) {
		std::ofstream outfile;
		std::string outfile_name(outfile_dir);
		outfile_name = outfile_name + "coord_summ" + std::to_string(i + 1) + ".csv";
		outfile.open(outfile_name, std::ofstream::out);
		outfile << "Time";
		for (int j = 0; j < numStats; j++) {
			if (getStat[j]) {
				outfile << "," << statName[j];
			}
		}
		outfile << std::endl;
		for (int t = 0; t < times.size(); t++) {
			outfile << t_val[t];
			for (int j = 0; j < numStats; j++) {
				if (getStat[j]) {
					outfile << "," << computeStat(j, times[t][i]);
				}
			}
			outfile << std::endl;
		}
		outfile.close();
	}

	for (int i = 0; i < numStats; i++) {
		if (getSum[i]) {
			std::ofstream summary;
			std::string summary_name(outfile_dir);
			summary_name = summary_name + statName[i] + "_summary.csv";
			summary.open(summary_name, std::ofstream::out);
			for (int t = 0; t < times.size(); t++) {
				for (int j = 0; j < times[t].size(); j++) {
					if (j == 0) {
						summary << computeStat(i, times[t][j]);
					}
					else {
						summary << "," << computeStat(i, times[t][j]);
					}
				}
				summary << std::endl;
			}
		}
	}


	return 0;
}
