/*
Model we want to test:
        The radius of gyration is the root mean of the squared
        distance from the center of mass to each point on the chain.

	The expected end-to-end distance is sqrt(Lp*Lc), where Lc
	is the contour length (the distance traced along the chain
	from one end to the other) and Lp is the persistence length
	(about 50nm for B-form DNA).

	The expected radius of gyration is sqrt(Lp*Lc/6).

        Rg = sqrt(<R(from center of mass)^2>)
        Lp = persistence length.  Lc = contour length.
        Lp*Lc = <R(end to end)^2>,  Also, Lp = E*I / (kB*T),
        which turns out to be when the correlation of the cosine of the angles goes below 1/e.
        It turns out that Rg^2 = Lp*Lc / 6.
	We want to compute Rg using the two-endpoints method.

        We'd like to have Lp much smaller than Lc.
        Lp = 50nm for B-form DNA, so we want Lc to be >= 1 micron.
        The end-to-end expected is 223nm, which is much less than this.
        Also try 10 microns to see if it works.
        We need to compute for a while to get this average,
        and let the system settle in before starting the average.
        We'll look at the time average over time and see when it settles down.
        It could take a lot of steps.
*/

#include <stdio.h>
#include <math.h>
#include <vector>
#include "chromoTypes.h"
#include "analysis.h"

typedef std::vector<CS_REAL3>  POINTS_FOR_A_GIVEN_TIME;

int main(int argc, const char *argv[])
{
    // Get the input and output files (for now, just standard in and out)
    FILE *infile = stdin;
    FILE *outfile = stdout;

    // Skip over the first block of text until we get to a line that
    // starts with '}', marking the end of the structure definition.
    // Then skip the line after that.
    char    line[4096];
    line[sizeof(line)-1] = '\0';    // Ensure null-termination
    do {
        if (fgets(line, sizeof(line)-1, infile) == NULL) {
            fprintf(stderr,"Did not find end of structure definition\n");
            return -1;
        }
    } while (line[0] != '}');
    if (fgets(line, sizeof(line)-1, infile) == NULL) {
        fprintf(stderr,"Did not find blank line after structure definition\n");
        return -1;
    }

    // Skip over the MassColors section of the file.
    // Then skip the line after that.
    do {
        if (fgets(line, sizeof(line)-1, infile) == NULL) {
            fprintf(stderr,"Did not find end of structure definition\n");
            return -1;
        }
    } while (!isspace(line[0]));

    // We repeat reading in blocks of points, each of which is preceded
    // by a line that says "Time T", where T is a floating-point number.
    // Then there are a number of lines of X Y Z which are floating-point
    // coordinates. Then there is a blank line after each block of points.
    std::vector<POINTS_FOR_A_GIVEN_TIME>    times;

    // we want to also record the actual times at which these were calculated
    // for accurate comparison between different runs (since SaveTrajectory
    // doesn't actually save every 0.1 seconds but only close to it)
    std::vector<double> t_vals;

    while (true) {
        // Read and parse the "Time" line.
        if (fgets(line, sizeof(line)-1, infile) == NULL) {
            // Done reading lines, we got them all.
            break;
        }
        double t;
        if (sscanf(line, "Time %lf", &t) <= 0) {
            fprintf(stderr, "Bad Time line: %s\n", line);
            return(-1);
        }
        // push the time to back of t_vals
        t_vals.push_back( t );

        // Read and parse point lines, adding them to our time.
        POINTS_FOR_A_GIVEN_TIME points;
        times.push_back(points);
        while (true) {
            // Read and parse the next line, which may be a points line.
            if (fgets(line, sizeof(line)-1, infile) == NULL) {
                // Done reading lines, we got them all.
                break;
            }
            double x,y,z;
            if (sscanf(line, "%lf %lf %lf", &x, &y, &z) != 3) {
                // No more points for this line
                break;
            }
            CS_REAL3 p;
            p.s[0] = static_cast<CS_REAL>(x);
            p.s[1] = static_cast<CS_REAL>(y);
            p.s[2] = static_cast<CS_REAL>(z);
            times.back().push_back(p);
        }
    }
    fclose(infile);

    // Compute the radius of gyration by looking at the motion of the points
    // in the chain over time.  Print out the radius of gyration at each time
    // step.
    size_t i;
    fprintf(outfile, "Time, RoG, EndToEnd, Contour\n");
    for (i = 0; i < times.size(); i++) {
	fprintf(outfile, "%lg, %lg, %lg, %lg\n",
                t_vals[i],
		compute_RoG(times[i]),
		sqrt(compute_squared_end_to_end(times[i])),
		compute_contour_length(times[i]) );
    }

    fclose(outfile);
    return 0;
}

