// chromoShake.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sstream>
#include "structureLoader.h"
#include "chromoSim.h"

void Usage(const char *s)
{
    fprintf( stderr, "Usage: %s [-GPU N] [-CPU] [-openCL_dir directory] [-profile pSteps] [-accelerate scheme] [-profileRandom outfile doint pSteps] [-save outfile saveIterations totalSaves] [ [[-relax TIMESTEP_ns COUNT] [infile]] | -continue]\n", s );
    fprintf( stderr, "       -GPU: Ask OpenCL to open GPU device N (default 0)\n" );
    fprintf( stderr, "       -CPU: Ask OpenCL to use the CPU instead of a GPU\n" );
    fprintf( stderr, "       -continue: Continue simulation from the last location specified in outfile (from -save or default) -- do not specify an input file.\n" );
    fprintf( stderr, "       -profile: Instead of simulating, profile the maximum displacements being made over pSteps time steps\n" );
    fprintf( stderr, "       -accelerate: Specify collision scheme (default 1)\n" );
    fprintf( stderr, "               scheme=-1 : no collision detection\n" );
    fprintf( stderr, "               scheme= 0 : brute force spherical collision detection\n" );
    fprintf( stderr, "               scheme= 1 : uniform volume subdivision spherical collision detection, GPUGems style\n" );
    fprintf( stderr, "               scheme= 2 : uniform volume subdivision spherical collision detection, NVidia example style\n" );
    fprintf( stderr, "       -profileRandom: save the random numbers generated in the first pSteps to outfile, if doint 0 -> try doing integers, if doint 1 -> do gaussians \n" );
    fprintf( stderr, "       -save: Save positions every saveIterations totalSaves times\n" );
    fprintf( stderr, "              (default temp.txt 100 100)\n" );
    fprintf( stderr, "       -relax: Rather than running the full sim, relax doing COUNT steps of TIMESTEP_ns and save that as the initial state in the output file\n" );
    fprintf( stderr, "       -openCL_dir: Open the specified directory to load the openCL kernels (default is ./)\n" );
    fprintf( stderr, "       infile: Configuration file (default chromoShake.cfg)\n" );
    exit(-1);
}

int main(int argc, const char *argv[])
{
    // Program version number.  Major, minor, bugfix
    const char *version_string = "01.02.00";

    // variables changed by command line argument
    bool useGPU = true; // Try to use a GPU for the calculations?
    unsigned whichGPU = 0;  // Try to open GPU device 0
    bool do_continue = false; // try to continue from save file?
    bool do_relax = false; // Run a relaxation run rather than an actual run?
    double relaxTimeNS;
    int relaxCount;
    bool do_profile = false;
    bool do_profileRandom = false;
    bool do_profileIntegers = false;
    int profileSteps = 0;
    int seed = 42; // magic number for seed
    double startTime = 0.0; // start time of the simulation
    double timeStep = 1; // time step to use if not specified in config file
    int saveIterations = 100; // the number of steps that we take each time before saving
    int totalIterations = 100; // the total number of times we iterate saveIterations times

    double collisionSpringConstant = 2;
    double springDamping = 0.0;

    // collision acceleration information
    int collisionScheme = 1;

    // File name to read, defaults to chromoShake.cfg
    const char * infile_name = "chromoShake.cfg";
    // File name to write to
    const char * outfile_name = "temp.txt";

    // Name of the openCL command file to load
    const char * openCL_file_dir = "./";

    // Command-line parsing
    int  i;
    unsigned  real_params = 0;  //< Number of non-flag parameters
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i],"-help")) {
            Usage(argv[0]);
        } else if (!strcmp(argv[i],"-GPU")) {
            if (++i >= argc) { Usage(argv[0]); }
            useGPU = true;
            whichGPU = static_cast<unsigned>(atoi( argv[i] ));
        } else if (!strcmp(argv[i],"-CPU")) {
            useGPU = false;
        } else if (!strcmp(argv[i],"-continue")) {
            do_continue = true;
	    infile_name = "";
        } else if (!strcmp(argv[i],"-profile")) {
            if (++i >= argc) { Usage(argv[0]); }
            do_profile = true;
            profileSteps = atoi( argv[i] );
        } else if (!strcmp(argv[i],"-accelerate")) {
            if (++i >= argc) { Usage(argv[0]); }
            collisionScheme = atoi( argv[i] );
        } else if (!strcmp(argv[i], "-profileRandom")) {
            do_profileRandom = true;
            if (++i >= argc) { Usage(argv[0]); }
            // save name of file to write to
            outfile_name = argv[i];
            if ( ++i >= argc ) { Usage( argv[0] ); }
            if ( !strcmp( argv[i], "0" ) ) { do_profileIntegers = true; }
            if (++i >= argc) { Usage(argv[0]); }
            // save the number of steps to profile
            profileSteps = atoi( argv[i] );
        } else if (!strcmp(argv[i],"-relax")) {
	    do_relax = true;
            if (++i >= argc) { Usage(argv[0]); }
            relaxTimeNS = atof( argv[i] );
            if (++i >= argc) { Usage(argv[0]); }
            relaxCount = atoi( argv[i] );
        } else if (!strcmp(argv[i],"-save")) {
            if (++i >= argc) { Usage(argv[0]); }
            // do outfile
            outfile_name = argv[i];
            if (++i >= argc) { Usage(argv[0]); }
            // do saveIterations
            saveIterations = atoi( argv[i] );
            if (++i >= argc) { Usage(argv[0]); }
            // do totalIterations
            totalIterations = atoi( argv[i] );
        } else if (!strcmp(argv[i],"-openCL_dir")) {
            if (++i >= argc) { Usage(argv[0]); }
            openCL_file_dir = argv[i];
        } else if (argv[i][0] == '-') {
            Usage(argv[0]);
        } else {
            switch (real_params++) {
            case 0:
                infile_name = argv[i];
                break;
            default:
                Usage(argv[0]);
            }
        }
    }

    // we don't want to both continue and specify an input file in one run.
    if (do_continue && (real_params > 0)) {
        std::cerr << "Either -continue or specify input file, not both..." << std::endl;
	Usage(argv[0]);
    }

    // we don't want to both continue and relax.
    if (do_continue && do_relax) {
        std::cerr << "Either -continue or -relax, not both..." << std::endl;
	Usage(argv[0]);
    }

    // we don't want to do both kinds of profiling in one run.
    if (do_profile && do_profileRandom) {
        std::cerr << "Please profile simulation and randomness separately..." << std::endl;
        exit(-22);
    }

    // Tell the version number.
    std::cout << "chromoShake version " << version_string << endl;

    // repeat command line arguments back to us
    std::cout << "Command line:";
    for (int i = 0; i < argc; ++i) {
        std::cout << " " << argv[i];
    }
    std::cout << std::endl;

    std::cout << "useGPU: ";
    if (useGPU) std::cout << "True, device " << whichGPU;
    else std::cout << "False";
    std::cout << std::endl;

    std::cout << "Using OpenCL directory: " << openCL_file_dir << std::endl;
    std::cout << "Using the number " << seed << " as seed value for random number generation" << std::endl;

    if (do_profile) {
        std::cout << "Profiling simulation details over " << profileSteps << " steps" << std::endl;
    } else if (do_profileRandom) {
        std::cout << "Saving random numbers over " << profileSteps << " to " << outfile_name << std::endl;
    } else  {
        if (do_continue) {
            std::cout << "Resuming from last state in " << outfile_name << std::endl;
        } else {
	    std::cout << "Reading configuration from file " << infile_name << std::endl;
        }
	if (do_relax) {
          std::cout << "Relaxing for " << relaxCount << " iterations of "
              << relaxTimeNS << " nanoseconds to file " << outfile_name << std::endl;
	} else {
          std::cout << "Saving every " << saveIterations << " iterations "
              << totalIterations << " times to file " << outfile_name << std::endl;
	}
    }

    // If we're continuing, then we read our input parameters from the output file,
    // rather than from a separate input input file.
    if (do_continue) {
        infile_name = outfile_name;
    }

    // Initialize structures for obtaining information from the configuration file
    MASS_node * mlist = NULL; // points to first node in mass list
    SPRING_node * slist = NULL; // points to first node in spring list
    HINGE_node * hlist = NULL; // points to first hinge in the hinge list
    double randomForce; // the random force
    std::vector<std::string> metaData;  // Metadata on the simulation

    // Read the structure description from the configuration file.
    FILE *f = fopen(infile_name, "r");
    if (f == NULL) {
        char s[1024];
        sprintf(s, "Could not open configuration file '%s' for reading", infile_name);
        perror(s);
        exit( -3 );
    }
    if (!parse_structure_from_file(f, &mlist, &slist, &hlist,
                                   &randomForce, &timeStep, &collisionSpringConstant,
                                   &springDamping, &seed,
                                   &metaData)) {
        fprintf( stderr, "Cannot parse %s\n", infile_name );
        exit( -4 );
    }
    fclose( f );

    // If we're continuing, we parse the collision scheme from the first
    // metadata line in the file.
    if (do_continue) {
        if (metaData.size() < 1) {
            std::cerr << "Did not find collision scheme in metadata during -continue" << endl;
            exit( -10 );
        }
        if (sscanf(metaData[0].c_str(), "meta collision_scheme %d", &collisionScheme) < 0) {
            std::cerr << "Could not parse collision scheme from metadata during -continue" << endl;
            exit( -11 );
        }
    }

    // Print out the metadata from the file.  Then the parameters we're using.
    for (i = 0; i < metaData.size(); i++) {
        std::cout << metaData[i];
    }
    std::cout << "Using simulation time step of " << timeStep << " per iteration" << std::endl;
    std::cout << "Using collision spring constant of " << collisionSpringConstant << std::endl;
    std::cout << "Using spring damping constant of " << springDamping << std::endl;
    std::cout << "Processing collisions using scheme " << collisionScheme << std::endl;
    
    // Process continue using code from old simulator (setting velocities to zero)
    // if do_continue then we want to change the mass coordinates
    if (do_continue) {
        // open the 'outfile,' which is really our continue file:
        FILE *cont_file = fopen(outfile_name, "r");
        if (cont_file == NULL) {
          char s[1024];
          sprintf(s, "Could not open old output file '%s' for reading", outfile_name);
          perror(s);
          fclose(cont_file);
          exit(-5);
        }
        // run function to parse out the new mass coordinates from this file
        int steps_already_taken;
        startTime = parse_continue(cont_file, &mlist, &steps_already_taken);
        if (startTime < 0.0 ) {
          fprintf(stderr,"Cannot parse %s\n", outfile_name);
          fclose(cont_file);
          exit(-6);
        }
        // Adjust the random-number seed based on the number of steps taken already
        seed += steps_already_taken;
        // close our continue file
        fclose(cont_file);
    }
    std::cout << "Random number seed " << seed << std::endl;
    std::cout << std::endl;

    // If we're relaxing, then we set the time step based on relaxation.
    if (do_relax) {
	timeStep = relaxTimeNS * 1e-9;
    }
    chromoSim simulation( mlist, slist, hlist, // structures from configuration files
              static_cast<CS_REAL>(randomForce), 
              static_cast<CS_REAL>(springDamping), // the default parameter in SOFA
              static_cast<CS_REAL>(collisionSpringConstant), // the spring constant for collisions
              static_cast<CS_REAL>(startTime), // what is the starting time of our simuation?
              static_cast<CS_REAL>(timeStep), // what is the time step of our simulation
              collisionScheme, // we haven't implemented cylindrical collisions yet
              useGPU, // are we using the GPU or CPU?
              seed, // use the random seed
              openCL_file_dir,
              4,   // TODO: Make this a parameter -- maximum number of collisions for a given particle
              whichGPU);

    if (!simulation.doing_okay()) {
	cerr << "Could not construct simulator." << endl;
	exit(-1);
    }
    
    if ( do_profile ) {
        simulation.profileSimulation( profileSteps );
    } else if ( do_profileRandom ) {
        simulation.profileRandom( profileSteps, do_profileIntegers, outfile_name );
    } else { // we are running to save the simulation
        // if we are continuing from previous simulation, we do not need to add initial state,
        // but if we are not, we should save initial state first...
        if (do_continue) {
            std::cout << "Continuing from simulation time " << simulation.getCurrentTime()
                << ", 0 percent complete." << std::endl;
        } else {
            // We are constructing a new output file.
            std::ofstream outFile;
            outFile.open( outfile_name, std::ios::out );
            if (!outFile.is_open()) {
                std::cerr << "Could not open " << outfile_name << endl;
                return -1;
            }

            // Note the collision scheme that was used.
            outFile << "meta collision_scheme " << collisionScheme << endl;

            // Copy the input file into the beginning of the output file.
            // Insert a blank link afterwards.
            char line[1024];
            line[sizeof(line)-1] = '\0';
            FILE *f = fopen(infile_name, "r");
            while (fgets( line, sizeof(line) - 1, f ) != NULL) {
                outFile << line;
            }
            outFile << endl;
            fclose(f);

            // Store the mass-color information
            outFile << "MassColors" << endl;
            for (MASS_node *m = mlist; m != NULL; m = m->next) {
                outFile << m->colorNumber << endl;
            }   
            outFile.flush();
            outFile.close();

	    // If we're relaxing, then we do the requested number of steps
	    // before saving the initial step.
	    if (do_relax) {
		simulation.simulateTimeSteps( relaxCount );
	    }

            // Append the initial position to the file.
            simulation.saveState( outfile_name, true );
            std::cout << "Saving initial state at simulation time " << simulation.getCurrentTime() 
                << ", 0 percent complete." << std::endl;
        }
	if (do_relax) {
		totalIterations = 0;
	}
        for ( int i = 1; i <= totalIterations; ++i ) {
            simulation.simulateTimeSteps( saveIterations );
            simulation.saveState( outfile_name, true );
            std::cout << "Saving at simulation time " << simulation.getCurrentTime() 
                << ", " << (100. * i) / totalIterations << " percent complete." << std::endl;
        }
    }

    return 0;
}

