// chromoSim.h
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif
#include "structureLoader.h"
#include <iostream>
#include <string>
#include <vector>

#include "chromoTypes.h"
// solver
#include "genericSolver.h"
#include "solverIMEXEuler.h"
#include "solverImplicitEuler.h"
// forces
#include "genericForce.h" // the generic force class
// the forces we are using
#include "forceHingesCL.h"
#include "forceSpringsCL.h"
#include "forceRandomCPU.h"
#include "forceRandomGPU.h"
#include "forceCollisionsBruteSphere.h"
#include "forceCollisionsUniformSphere.h"
#include "forceCollisionsUniformSphere2.h"
#include "forceMassDamping.h"

#include "sumForces.h"


#ifndef CHROMOSIM_H
#define CHROMOSIM_H


class chromoSim
{
private :
    bool broken;	// Is the simulator broken for some reason?
    bool verbose;

    // simulation time
    double initialTime;
    // time step
    CS_REAL simTimeStep;
    // time steps made so far
    int timeStepsMade;

    // stuff for working with OpenCL
    cl_platform_id platformId; // platform ID
    cl_device_id deviceId; // device ID
    cl_context context; // the context we create
    cl_command_queue queue; // the command queue

    // our solver class
    genericSolver * solver;

    // what we use to calculate our forces
    genericForce * gDamping;
    genericForce * gBrownian;
    genericForce * gSpring;
    genericForce * gHinge;
    genericForce * gCollision;

    // what we use to add our forces
    sumForces * totalForce;

    // function for initializing the aspects of OpenCL unrelated to the simulation.
    // It is told whether to try and run on a GPU device (otherwise, it selects a
    // CPU device) and which device number to use.  It defaults to using the first
    // GPU device.
    // Returns true if was able to initialize, false if not.
    bool initializeOpenCL ( bool useGPU = true, unsigned whichDevice = 0 );

    // function for checking and printing OpenCL errors.
    // Returns true on error, false on no error.
    static bool CheckErr ( cl_int err, const char * name ) {
        if (err != CL_SUCCESS) {
            std::cerr << "ERROR: " << name << " (" << err << ")" << std::endl;
	    return true;
        }
	return false;
    }


public :
    // most general constructor
    chromoSim ( MASS_node * mh, SPRING_node * sh, HINGE_node * hh,
        CS_REAL randomForce,
        CS_REAL springDamping, // the SOFA-esque damping constant associated with relative velocity over springs. F = -k * springDamping * relativeVelocity
        CS_REAL collisionSpringConstant, // the spring constant for collisions
        CS_REAL startingTime, // the time that we start off as saying we're at
        CS_REAL timeStep, // the time step used for each iteration
        int collisionScheme, // collision scheme (0 is brute force spheres, 1 is GPUGems spheres spatial subdivision, 2 is NVidia spheres spatial subdivision...)
        bool useGPU, // asks openCL for a GPU if true (default), asks for a CPU if false
        int randomSeed, // asks for a random seed to use
        std::string clFilesDir // the file being used for OpenCL
	, cl_ulong maximumInteractions = 4  // Maximum particles interacting with a single particle in one step.
        , int whichDevice = 0   // Which device to attempt to open?
        );

    // destructor
    virtual ~chromoSim ( );

    // Is everything okay with the simulator?
    bool doing_okay(void) const { return !broken; }

    // get the current time of our simulation based on the startingTime and the number of time steps made since then
    double getCurrentTime ( ) {
        return initialTime + timeStepsMade * simTimeStep;
    }

    // simulate a step forward
    bool simulateTimeStep ( );

    // simulate several time steps forward
    bool simulateTimeSteps ( int numSteps ) {
        for (int i = 0; i < numSteps; ++i) {
            if (!simulateTimeStep( )) {
		return false;
	    }
        }
        return true;
    }

    // save the current state
    bool saveState ( std::string fileName, // the name/path of the file that we will be saving to
        bool append // if true, append to fileName, if false, overwrite fileName
        );

    // Read the current state (position and velocity of all particles)
    bool readState ( std::vector<CS_REAL3> &pos, std::vector<CS_REAL3> &vel );

    // for purposes of debugging, in a sense
    bool printForces ( );
    bool checkNAN ( );
    bool profileSimulation ( int numSteps );
    bool profileRandom( int numSteps, bool do_integers, std::string fileName );
};

#endif
