#include <QtGui>
#include <QtOpenGL>
#include <QFileDialog>
#include <QMessageBox>
#include <math.h>
#include "myglwidget.h"

MyGLWidget::MyGLWidget(QWidget *parent)
    : QGLWidget(QGLFormat(QGL::SampleBuffers), parent)
    , d_play_timer(NULL)
    , d_yRotSpeed(0)
{
    d_xRot = 0;
    d_yRot = 0;
    d_zRot = 0;

    // Construct the colors to be used for the various model parts.
    // They correspond to the color indices.
    // The fifth one should be white, to match the expectation
    // from the original brownianMotion code.
    // TODO: Consider making them isoluminant.
    d_colors.push_back(QColor::fromRgb(255,128,128));
    d_colors.push_back(QColor::fromRgb(128,128,255));
    d_colors.push_back(QColor::fromRgb(128,255,128));
    d_colors.push_back(QColor::fromRgb(255,128,255));
    d_colors.push_back(QColor::fromRgb(255,255,255));

    // Make all of the objects visible.
    for (int i = 0; i < d_colors.size(); i++) {
        d_visible.push_back(true);
    }

    // Background color.
    d_background = QColor::fromRgb(50,50,75);

    computeBounds();

    // Ask for the keyboard focus.
    this->setFocusPolicy(Qt::StrongFocus);

    // Connect the play timer to our slot.  Go ahead and
    // start the timer to be called every 50ms.
    d_currentTau = -1;
    d_deltaTau = 1;
    d_currentIndex = 0;
    d_play_timer = new QTimer(this);
    connect(d_play_timer, SIGNAL(timeout()), SLOT(playTime()));
    d_play_timer->start(50);
}

void MyGLWidget::playTime()
{
    // Draw things if we have any masses
    if (d_massLocations.size() > 1) {
        // Play forwards or backwards, then make sure we're not
        // out of bounds.  Then find the current time index based
        // on truncating the floating-point time.
        d_currentTau += d_deltaTau;
        while (d_currentTau >= d_massLocations.size()) {
            d_currentTau -= d_massLocations.size();
        }
        while (d_currentTau < 0) {
            d_currentTau += d_massLocations.size();
        }
        d_currentIndex = static_cast<int>(d_currentTau) % d_massLocations.size();
        updateGL();
    }

    // Rotate the viewpoint slowly around the vertical axis.
    setYRotation(d_yRot + d_yRotSpeed);
}

MyGLWidget::~MyGLWidget()
{
}

QSize MyGLWidget::minimumSizeHint() const
{
    return QSize(300, 300);
}

QSize MyGLWidget::sizeHint() const
{
    return QSize(400, 400);
}

static void qNormalizeAngle(int &angle)
{
    while (angle < 0) angle += 360 * 16;
    while (angle > 360 * 16) angle -= 360 * 16;
}

void MyGLWidget::loadSimulationOutput(void)
{
    QString fileName = QFileDialog::getOpenFileName(this,
        tr("Open Simulation File"), "", tr("Simulation output files (*.out *.txt)"));
    if (fileName.length() > 0) {
        if (!loadSimulationOutputFromFile(fileName)) {
            QMessageBox::warning(this, tr("Failure"),
                                            tr("Could not load ") + fileName,
                                            QMessageBox::Ok);
        }
    }
    updateGL();
}

void MyGLWidget::setYRotationSpeed(int speed)
{
    d_yRotSpeed = speed;
}


bool MyGLWidget::loadSimulationOutputFromFile(const QString &name)
{
    // Open the file for reading.
    FILE *f = fopen(name.toStdString().c_str(), "r");
    if (f == NULL) {
        return false;
    }

    // Clear all information from the previous simulation.
    d_metaInfo.clear();

    // Read the meta information from the file.
    // Then skip the rest of the header.
    char line[4096];
    fgets(line, sizeof(line), f);
    while (!strncmp(line, "meta", strlen("meta"))) {
        d_metaInfo += QString(&line[strlen("meta ")]);
        fgets(line, sizeof(line), f);
    }
    while (line[0] != '}') {
        if (fgets(line, sizeof(line), f) == NULL) {
            return false;
        }
    }

    // Read the color information from the file.
    // First find the color line.
    while (0 != strncmp(line, "MassColors", strlen("MassColors"))) {
        if (fgets(line, sizeof(line), f) == NULL) {
            return false;
        }
    }
    unsigned colorIndex;
    if (fgets(line, sizeof(line), f) == NULL) {
        return false;
    }
    while (sscanf(line, "%u", &colorIndex) > 0) {
        d_massColors.push_back(colorIndex);
        if (fgets(line, sizeof(line), f) == NULL) {
            return false;
        }
    }

    // Read the position information from the file,
    // adding a new time for each Time line in the
    // file.
    double time_secs;
    double x,y,z;
    if (fgets(line, sizeof(line), f) == NULL) {
        return false;
    }
    while (sscanf(line, "Time %lg", &time_secs) > 0) {
        // Found another Time line; gobble up all of the
        // position lines and add them to the list.
        printf("Reading Time %lg\n", time_secs);
        MassList ml;
        ml.time_secs = time_secs;
        d_massLocations.push_back(ml);
        if (fgets(line, sizeof(line), f) == NULL) {
            return false;
        }
        while (sscanf(line, "%lg %lg %lg",&x, &y, &z) == 3) {
            Point3D p;
            p.x = x; p.y = y; p.z = z;
            d_massLocations.back().points.push_back(p);
            if (fgets(line, sizeof(line), f) == NULL) {
                break;
            }
        }

        // Prime the read for the next time line after skipping
        // the blank line between times.
        fgets(line, sizeof(line), f);
    }
    computeBounds();
    d_metaInfo += QString("\n");
    d_metaInfo += QString::number(d_massColors.size()) + " masses\n";
    d_metaInfo += QString::number(d_massLocations.size()) + " times\n";
    d_metaInfo += QString("\n");
    d_metaInfo += QString("Bounds (um): ")
            + QString::number(d_minX*1e6) + ","
            + QString::number(d_minY*1e6) + ","
            + QString::number(d_minZ*1e6) + " - "
            + QString::number(d_maxX*1e6) + ","
            + QString::number(d_maxY*1e6) + ","
            + QString::number(d_maxZ*1e6) + "\n";

    // Done.  Close things up and tell about the new info.
    fclose(f);
    emit fileNameChanged(name);
    emit metaInfoChanged(d_metaInfo);
    updateGL();
    return true;
}

void MyGLWidget::computeBounds()
{
    // If we have no masses, we set some defaults.
    if ( (d_massLocations.size() == 0) ||
         (d_massLocations[0].points.size() == 0) ) {
        d_minX = d_minY = d_minZ = -1;
        d_maxX = d_maxY = d_maxZ = 1;
        return;
    }

    // Look over all time and find out how far any point travels.
    d_minX = d_maxX = d_massLocations[0].points[0].x;
    d_minY = d_maxY = d_massLocations[0].points[0].y;
    d_minZ = d_maxZ = d_massLocations[0].points[0].z;
    int i;
    for (i = 0; i < d_massLocations.size(); i++) {
        const QVector<Point3D> &points = d_massLocations[i].points;
        int j;
        for (j = 0; j < points.size(); j++) {
            double x = points[j].x;
            double y = points[j].y;
            double z = points[j].z;
            if (x < d_minX) { d_minX = x; }
            if (x > d_maxX) { d_maxX = x; }
            if (y < d_minY) { d_minY = y; }
            if (y > d_maxY) { d_maxY = y; }
            if (z < d_minZ) { d_minZ = z; }
            if (z > d_maxZ) { d_maxZ = z; }
        }
    }
}

void MyGLWidget::setXRotation(int angle)
{
    qNormalizeAngle(angle);
    if (angle != d_xRot) {
        d_xRot = angle;
        emit xRotationChanged(angle);
        updateGL();
    }
}

void MyGLWidget::setYRotation(int angle)
{
    qNormalizeAngle(angle);
    if (angle != d_yRot) {
        d_yRot = angle;
        emit yRotationChanged(angle);
        updateGL();
    }
}

void MyGLWidget::setZRotation(int angle)
{
    qNormalizeAngle(angle);
    if (angle != d_zRot) {
        d_zRot = angle;
        emit zRotationChanged(angle);
        updateGL();
    }
}

void MyGLWidget::initializeGL()
{
    qglClearColor(d_background);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    static GLfloat lightPosition[4] = { 0.5, 5.0, 7.0, 1.0 };
    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
    glEnable(GL_NORMALIZE);
    //glDisable(GL_TEXTURE);
    //glEnable(GL_COLOR_MATERIAL);
    emit versionChanged("Version: 1.2.0");
}

void MyGLWidget::paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

    // Move our viewpoint so that we have a look at our
    // unit volume from a distance that puts it into the
    // center of the viewport, and within the clipping
    // planes.
    glTranslatef(0.0, 0.0, -10.0);

    // Scale so that our viewport fills the range
    // (-0.5, 0.5) in X and Y, with a little extra.
    double scale = d_maxX - d_minX;
    if (scale < d_maxY - d_minY) { scale = d_maxY - d_minY; }
    if (scale < d_maxZ - d_minZ) { scale = d_maxZ - d_minZ; }
    scale *= 1.1;
    glScaled(1/scale, 1/scale, 1/scale);

    // Rotate around the world center.
    glRotatef(d_xRot / 16.0, 1.0, 0.0, 0.0);
    glRotatef(d_yRot / 16.0, 0.0, 1.0, 0.0);
    glRotatef(d_zRot / 16.0, 0.0, 0.0, 1.0);

    // Translate to put the center of the region where
    // there are masses at the center of the world.
    glTranslated( -(d_minX+d_maxX)/2,
                  -(d_minY+d_maxY)/2,
                  -(d_minZ+d_maxZ)/2);

    // If we have no time steps, we're done.
    if (d_massLocations.size() == 0) { return; }

    // Draw the spheres for the current simulation time step
    GLUquadricObj   *q = gluNewQuadric();
    GLfloat color[3];
    gluQuadricDrawStyle(q, GLU_FILL);
    gluQuadricNormals(q, GLU_SMOOTH);
    gluQuadricOrientation(q, GLU_OUTSIDE);
    const MassList &ml = d_massLocations[d_currentIndex];
    int count = ml.points.size();
    for (int i = 0; i < count; i++) {
        unsigned colorID = (d_massColors[i] - 1) % d_colors.size();
        if (d_visible[colorID]) {
            color[0] = d_colors[colorID].red() / 255.0;
            color[1] = d_colors[colorID].green() / 255.0;
            color[2] = d_colors[colorID].blue() / 255.0;
            glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
            glPushMatrix();
                glTranslated(ml.points[i].x, ml.points[i].y, ml.points[i].z);
                gluSphere(q, 6e-9, 7, 7);
            glPopMatrix();
        }
    }
    gluDeleteQuadric(q);
}

void MyGLWidget::resizeGL(int width, int height)
{
    glViewport(0, 0, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    double side = qMin(width, height);
    glOrtho(-0.5 * width/side, +0.5 * width/side,
            -0.5 * height/side , +0.5 * height/side,
            4.0, 15.0);
    glMatrixMode(GL_MODELVIEW);
}

void MyGLWidget::mousePressEvent(QMouseEvent *event)
{
    d_lastPos = event->pos();
}

void MyGLWidget::mouseMoveEvent(QMouseEvent *event)
{
    int dx = event->x() - d_lastPos.x();
    int dy = event->y() - d_lastPos.y();

    if (event->buttons() & Qt::LeftButton) {
        setXRotation(d_xRot + 8 * dy);
        setYRotation(d_yRot + 8 * dx);
    } else if (event->buttons() & Qt::RightButton) {
        setXRotation(d_xRot + 8 * dy);
        setZRotation(d_zRot + 8 * dx);
    }
    d_lastPos = event->pos();
}

void MyGLWidget::keyPressEvent(QKeyEvent *keyEvent)
{
    switch (keyEvent->key()) {
    case Qt::Key_Up:
        d_deltaTau += 0.25;
        break;
    case Qt::Key_Down:
        d_deltaTau -= 0.25;
        break;
    case Qt::Key_Right:
        d_currentTau += 1;
        d_deltaTau = 0;
        break;
    case Qt::Key_Left:
        d_currentTau -= 1;
        d_deltaTau = 0;
        break;
    case '1':   // Make objects visible or invisible
    case '2':
    case '3':
    case '4':
    case '5':
        d_visible[keyEvent->key() - '1'] = !d_visible[keyEvent->key() - '1'];
        break;
    default:
        printf("XXX Unrecognized key\n");
        keyEvent->ignore();
        return;
    };

    // If we handled the key, we accept it.
    keyEvent->accept();

    // We may need to update the display
    updateGL();
}
