// forceCollisionsBruteSphere.cl
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II
// OpenCL file for forceCollisionsBruteSphere

#include <chromoTypes.cl>


// apply spherical collision models
__kernel void applySphericalCollisions( // applies spring collision forces using spheres
    const CS_REAL collisionSpringConstant,
    const ulong numParticles,
    __global CS_REAL * massRadius,
    __global CS_REAL3 * position,
    __global CS_REAL3 * fCollision )
{
    ulong ndx = get_global_id(0); // we can worry about groups later

    if (ndx >= numParticles)
    {
        return;
    }

    CS_REAL3 cforce = 0; // set to zero
    CS_REAL3 pos = position[ndx]; // keep the current position as a local variable
    
    // See how brute force works in parallel
    CS_REAL fixedOffset = massRadius[ndx];
    for (ulong i = 0; i < numParticles; ++i)
    {
        if (i != ndx) {
            CS_REAL3 displacement = pos - position[i];
            // TODO add additional buffer between balls set by user
            CS_REAL contactDistance = fixedOffset + massRadius[i] - length( displacement );
            // if contactDistance > 0, then we have contact, thus
            if (contactDistance > 0) {
                cforce += collisionSpringConstant * contactDistance * normalize( displacement );
            }
        }
    }
    fCollision[ndx] = cforce;
}

// apply spherical collision models, calculate Jacobian
__kernel void applySphericalCollisionsAndJacobian( // applies spring collision forces using spheres, gets Jacobian
    const CS_REAL collisionSpringConstant,
    const ulong numParticles,
    __global CS_REAL * massRadius,
    __global CS_REAL3 * position,
    __global CS_REAL3 * fCollision,
    __global CS_REAL3 * collisionJPosX,
    __global CS_REAL3 * collisionJPosY,
    __global CS_REAL3 * collisionJPosZ,
    __global ulong * collisionInteractionIndex, // what is being interacted with
    __global ulong * collisionNumInteractions // how many have we interacted with so far )
    )
{
    ulong ndx = get_global_id(0); // we can worry about groups later

    if (ndx >= numParticles)
    {
        return;
    }

    CS_REAL3 cforce = 0; // set to zero
    CS_REAL3 pos = position[ndx]; // keep the current position as a local variable
    ulong numInteractions = 1; // how many interactions do we currently have? (say self at first)
    // keep track of (negative of) self position Jacobian
    CS_REAL3 selfJPosX = 0;
    CS_REAL3 selfJPosY = 0;
    CS_REAL3 selfJPosZ = 0;
    
    // See how brute force works in parallel
    CS_REAL fixedOffset = massRadius[ndx];
    for (ulong i = 0; i < numParticles; ++i)
    {
        if (i != ndx) {
            // displacement vector
            CS_REAL3 d = position[i] - pos;
            // distance from this
            CS_REAL dLength = length(d);
            // unitless scale of if we're stretching or compressed
            CS_REAL scale = 1 - (fixedOffset + massRadius[i]) / dLength;
            if (scale < 0) // we have a collision
            {
                // compute outer product of displacement with itself
                CS_REAL3 outerX = d.x * d / (dLength*dLength);
                CS_REAL3 outerY = d.y * d / (dLength*dLength);
                CS_REAL3 outerZ = d.z * d / (dLength*dLength);
                // add the force
                cforce += collisionSpringConstant * scale * d;
                // compute the Jacobian
                CS_REAL3 jx = collisionSpringConstant * ( outerX + scale * ( ((float3) (1,0,0)) - outerX ) );
                CS_REAL3 jy = collisionSpringConstant * ( outerY + scale * ( ((float3) (0,1,0)) - outerY ) );
                CS_REAL3 jz = collisionSpringConstant * ( outerZ + scale * ( ((float3) (0,0,1)) - outerZ ) );
                // we will add to selfJPosX these values although they are negative
                // of actual self interaction Jacobian. This means that when we update
                // the __global variable, we put it in later as negative
                selfJPosX += jx;
                selfJPosY += jy;
                selfJPosZ += jz;
                // update cross term (ndx to i)
                ulong cross_ndx = numInteractions*numParticles + ndx;
                collisionJPosX[cross_ndx] = jx;
                collisionJPosY[cross_ndx] = jy;
                collisionJPosZ[cross_ndx] = jz;
                collisionInteractionIndex[cross_ndx] = i;
                // we have just added one more interaction
                ++numInteractions;
            }
        }
    }
    // update the collision force
    fCollision[ndx] = cforce;
    // update the number of interactions (number of collisions + 1)
    collisionNumInteractions[ndx] = numInteractions;
    // update the self collision Jacobian
    collisionJPosX[ndx] = -selfJPosX;
    collisionJPosY[ndx] = -selfJPosY;
    collisionJPosZ[ndx] = -selfJPosZ;
}