// forceCollisionsUniformSphere.cl
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II
// OpenCL file for forceCollisionsBruteSphere
// following example from GPUGems 3

#include <chromoTypes.cl>


typedef struct POSITION_HASH
{
    ulong hash;
    ulong index;
    char homeCellType; // what is the home cell of the particle?
    char thisCellType; // which cell type does this particular entry correspond to?
    char inCellType; // really 8 bits that will be used like a 8 bool array
} POSITION_HASH;

typedef struct COLLISION_CELL
{
    ulong homeCellStart; // start index of collision cells, starts with home cells
    ulong phantomCellStart; // end index of home cells, start index of phantom cells
    ulong phantomCellEnd; // end index of phantom cells (and thus the group with the same hash)
} COLLISION_CELL;
    

// function that assigns a grid coordinate to a given position coordinate
int3 getGridCoordinate( CS_REAL3 posCoordinate, CS_REAL3 cellDimensions )
{
    return convert_int3( posCoordinate / cellDimensions );
}

// function that assigns a cell type to a given grid coordinate
char getCellType( int3 gridCoordinate )
{
    return (gridCoordinate.x & 1) + 2*( (gridCoordinate.y & 1) + 2*(gridCoordinate.z & 1) );
}

// function that assigns a positive value from a grid coordinate (zero value is thus invalid)
ulong getGridHash( int3 gridCoordinate, uint3 gridScale )
{
    return 1 + convert_ulong( 
        gridCoordinate.x * gridScale.x
        + gridCoordinate.y * gridScale.y
        + gridCoordinate.z * gridScale.z
        );
}

// kernel function for getting the hashes of the position vectors
__kernel void calculateHashedGrid(
    const uint3 gridScale,
    const CS_REAL3 cellDimensions,
    const ulong numParticles,
    __global CS_REAL3 * position, // INPUT: the current position 
    __global CS_REAL * boundingRadius, // INPUT: a multiple of the mass radius used for broad phase collision detection (typically 1.5 * mass radius)
    __global POSITION_HASH * hashedPositions // OUTPUT: hash array of length 8 * numParticles
    )
{
    ulong ndx = get_global_id(0); // we can worry about groups later

    if (ndx >= numParticles)
    {
        return;
    }

    CS_REAL3 pos = position[ndx];
    CS_REAL rad = boundingRadius[ndx];

    int3 gridCoord = getGridCoordinate( pos, cellDimensions );
    char homeType = getCellType( gridCoord );
    // get the hashed coordinates for the eight different possible intersections
    ulong hashedCoord[8];
    char thisCellType[8];
    hashedCoord[0] = getGridHash( gridCoord, gridScale );
    thisCellType[0] = homeType;
    for (ulong i = 1; i < 8; ++i)
    {
        hashedCoord[i] = 0;
        thisCellType[i] = 254;
    }
    // we will keep the order of the cell types with the exception of the home cell. The home
    // cell will be kept at index zero -- and cell type zero goes to whatever the home cell type is.
    // this way, the home cell will be in our array first, so our stable sort will keep it in the front.

    // inCellType, however, will keep track of the cellTypes without such a switch
    char inCellType = 0;
    inCellType |= (1 << homeType);

    // loop over the neighboring possible ways we could intersect and see if we do
    for (int dx = -1; dx <= 1; ++dx)
    {
    for (int dy = -1; dy <= 1; ++dy)
    {
    for (int dz = -1; dz <= 1; ++dz)
    {
        CS_REAL3 dpos = pos + ( (CS_REAL3) ( dx * rad, dy * rad, dz * rad ) );
        int3 dGridCoord = getGridCoordinate( dpos, cellDimensions );
        char dType = getCellType( dGridCoord );
        if (dType == homeType) // we did not go into another cell in this direction
        {
            continue;
        }
        // otherwise...
        inCellType |= (1 << dType);
        ulong hash_index = dType == 0 ? homeType : dType;
        hashedCoord[hash_index] = getGridHash( dGridCoord, gridScale );
        thisCellType[hash_index] = dType;
    }
    }
    }

    // set values for hashedPositions...
    for (ulong i = 0; i < 8; ++i)
    {
        ulong hNdx = i * numParticles + ndx;
        hashedPositions[hNdx].hash = hashedCoord[i];
        hashedPositions[hNdx].thisCellType = thisCellType[i];
        hashedPositions[hNdx].index = ndx;
        hashedPositions[hNdx].homeCellType = homeType;
        hashedPositions[hNdx].inCellType = inCellType;
    }
}

// kernel function for using sorted hashes to calculate the collision forces
__kernel void applyCollisions(
    const CS_REAL collisionSpringConstant,
    const ulong numParticles,
    const ulong numCollisionCells,
    __global COLLISION_CELL * cells,
    __global POSITION_HASH * hashes,
    __global CS_REAL * massRadius,
    __global CS_REAL3 * position,
    __global CS_REAL3 * fCollision // assume that fCollision is zeroed out between every step
    )
{
    ulong ndx = get_global_id(0);

    if (ndx >= numCollisionCells)
    {
        return;
    }

    // get indices to iterate over on hashes
    ulong hStart = cells[ndx].homeCellStart;
    ulong pStart = cells[ndx].phantomCellStart;
    ulong pEnd = cells[ndx].phantomCellEnd;

    // get the type of this collision cell
    char cellType = hashes[hStart].homeCellType;

    
    // loop over the home cells and then all the rest...
    for (ulong i = hStart; i < pStart; ++i)
    {
        POSITION_HASH hashi = hashes[i]; // get the hash associated with i
        ulong ndx_i = hashi.index; // the index of the position of this hash
        CS_REAL fixedOffset = massRadius[ndx_i]; // offset distance from particle i
        CS_REAL3 positioni = position[ndx_i]; // keep this position in local memory for time being
        for (ulong j = i + 1; j < pEnd; ++j)
        {
            POSITION_HASH hashj = hashes[j];
            char jType = hashj.homeCellType;

            if ( jType < cellType && (hashi.inCellType & (1 << jType)) )
            {
                // we should get this when the particle associated with hashj is in
                // its home cell
                continue;
            }

            // otherwise, calculate the force if necessary
            ulong ndx_j = hashj.index;
            CS_REAL3 displacement = positioni - position[ndx_j];
            CS_REAL contactDistance = fixedOffset + massRadius[ndx_j] - length( displacement );
            // if contact distance > 0, then we have contact so
            if (contactDistance > 0)
            {
                // CS_REAL3 force = collisionSpringConstant * contactDistance * normalize( displacement );
                CS_REAL3 force = collisionSpringConstant * contactDistance * normalize( displacement );
                fCollision[ndx_i + cellType*numParticles] += force;
                fCollision[ndx_j + cellType*numParticles] -= force;
            }
        }
    }
}
    