// forceCollisionsUniformSphere.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "forceCollisionsUniformSphere.h"
#include <algorithm> // for stable_sort

// what's the name of this force?
const char * forceCollisionsUniformSphere::getName( )
{
    return "Force UVSD Spherical Collision";
}

// constructor
forceCollisionsUniformSphere::forceCollisionsUniformSphere(
    std::string sim_CLdir, // the path of directory with openCL source
    cl_context sim_context, // openCL context to use
    cl_command_queue sim_queue, // openCL queue to use
    cl_ulong sim_numParticles, // number of particles in the simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    cl_device_id sim_deviceId, // the deviceId we are running on
    cl_mem sim_clPosition, // the positions
    CS_REAL * sim_MassRadius, // the mass radius information
    CS_REAL collisionSpringConstant,
    CS_REAL boundingRadiusScale, // the proportion of bounding radius to mass radius (~ 1.5)
    CS_REAL cellDimensionScale, // the proportion of cell length size to the largest mass radius (~ 5)
    cl_uint gridScale // parameter for hash function: gx + gridScale*( gy + gridScale * gz )
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, sim_totalMaxInteractions, true )
{
    // set collisionConstant
    collisionConstant = collisionSpringConstant;
    
    // there may be 8 threads at a time trying to access a given force coordinate
    columns = 8;

    // variable for getting errors from OpenCL API
    cl_int error = 0;

    // keep track of the maximum massRadius
    CS_REAL maxMassRadius = 0;

    // copy the mass radius array to device
    clMassRadius = clCreateBuffer( context, // put into our class' context
        CL_MEM_READ_ONLY, // make the buffer read only
        numParticles * sizeof( CS_REAL ), // allocate space for the springs
        NULL, &error);
    checkErr( error, "clCreateBuffer (massRadius)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clMassRadius, CL_TRUE, 0,
            numParticles * sizeof( CS_REAL ), sim_MassRadius,
            0, NULL, NULL),
        "clEnqueueWriteBuffer (massRadius)"
        );

    // get the bounding radius array to device
    CS_REAL * boundingRadius = new CS_REAL[numParticles];
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        CS_REAL mr = sim_MassRadius[i];
        boundingRadius[i] = mr * boundingRadiusScale;
        // check if the mass radius is larger than the current known maximum
        if (mr > maxMassRadius)
        {
            maxMassRadius = mr;
        }
    }
    clBoundingRadius = clCreateBuffer( context, CL_MEM_READ_ONLY, numParticles * sizeof( CS_REAL ),
        NULL, &error);
    checkErr( error, "clCreateBuffer(boundingRadius)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clBoundingRadius, CL_TRUE, 0,
            numParticles * sizeof( CS_REAL ), boundingRadius,
            0, NULL, NULL),
        "clEnqueueWriteBuffer (boundingRadius)"
        );
    delete[] boundingRadius;

    // create structure for hashes of positions in the grid, send to device
    positionHash = new POSITION_HASH[numParticles*8];
    for (int i = 0; i < numParticles*8; ++i)
    {
        positionHash[i].hash = 0;
        positionHash[i].index = 0;
        positionHash[i].homeCellType = 0;
        positionHash[i].thisCellType = 0;
        for (int j = 0; j < 8; ++j)
        {
            positionHash[i].inCellType = 0;
        }
    }
    clPositionHash = clCreateBuffer( context, CL_MEM_READ_WRITE,
        numParticles*8 * sizeof( POSITION_HASH ), NULL, &error);
    checkErr( error, "clCreateBuffer (positionHash)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clPositionHash, CL_TRUE, 0,
            numParticles*8 * sizeof( POSITION_HASH ), positionHash,
            0, NULL, NULL),
        "clEnqueueWriteBuffer (positionHash)"
        );

    // there will never be more than numParticles number of collision cells
    collisionCell = new COLLISION_CELL[numParticles];
    clCollisionCell = clCreateBuffer( context, CL_MEM_READ_WRITE,
        numParticles * sizeof( COLLISION_CELL ), NULL, &error);
    checkErr( error, "clCreateBuffer (collisionCell)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clCollisionCell, CL_TRUE, 0,
            numParticles * sizeof( COLLISION_CELL ), collisionCell,
            0, NULL, NULL),
        "clEnqueueWriteBuffer (collisionCell)"
        );

    // create a force vector and set it to zero on host and device
    initForcesZero( );

    // create additional zero force vector to zero out collisions each time
    fZero = new CS_REAL3[numParticles*columns];
    for (int i = 0; i < numParticles*columns; ++i)
    {
        fZero[i].s[0] = 0;
        fZero[i].s[1] = 0;
        fZero[i].s[2] = 0;
    }

    // load the program "forceCollisionsUniformSphere.cl"
    loadProgram( sim_deviceId, "forceCollisionsUniformSphere.cl" );

    // load the kernel to get the hashes
    clKernelGetHash = clCreateKernel( program,
        "calculateHashedGrid", &error );
    checkErr( error, "clCreateKernel (clKernelGetHash)" );

    // pass arguments into the kernel to get the hashes
    cl_uint3 clGridScale;
    clGridScale.s[0] = 1;
    clGridScale.s[1] = gridScale;
    clGridScale.s[2] = gridScale*gridScale;
    error = clSetKernelArg( clKernelGetHash, 0,
        sizeof( cl_uint3 ), &clGridScale );
    CS_REAL3 clCellDimensions;
    clCellDimensions.s[0] = maxMassRadius * cellDimensionScale;
    clCellDimensions.s[1] = maxMassRadius * cellDimensionScale;
    clCellDimensions.s[2] = maxMassRadius * cellDimensionScale;
    error |= clSetKernelArg( clKernelGetHash, 1,
        sizeof( CS_REAL3 ), &clCellDimensions );
    error |= clSetKernelArg( clKernelGetHash, 2,
        sizeof( cl_ulong ), &numParticles );
    error |= clSetKernelArg( clKernelGetHash, 3,
        sizeof( cl_mem ), &sim_clPosition );
    error |= clSetKernelArg( clKernelGetHash, 4,
        sizeof( cl_mem ), &clBoundingRadius );
    error |= clSetKernelArg( clKernelGetHash, 5,
        sizeof( cl_mem ), &clPositionHash );
    checkErr( error, "clSetKernelArg (clKernelGetHash)" );

    // get global work item size for this
    clKernelGetHashGlobalSize = getKernelPaddedGlobalSize( 
        sim_deviceId, clKernelGetHash, numParticles );

    // TODO: parallelize the sort and determination of collision cells phases

    // load the kernel to compute the forces
    clKernelCollision = clCreateKernel( program,
        "applyCollisions", &error );
    checkErr( error, "clCreateKernel (clKernelCollision)" );

    // pass arguments into the kernel
    error = clSetKernelArg( clKernelCollision, 0, 
        sizeof(CS_REAL), &collisionConstant );
    error |= clSetKernelArg( clKernelCollision, 1,
        sizeof(cl_ulong), &numParticles );
    error |= clSetKernelArg( clKernelCollision, 2,
        sizeof(cl_ulong), &numParticles ); // eventually number of collision cells
    error |= clSetKernelArg( clKernelCollision, 3,
        sizeof(cl_mem), &clCollisionCell );
    error |= clSetKernelArg( clKernelCollision, 4,
        sizeof(cl_mem), &clPositionHash );
    error |= clSetKernelArg( clKernelCollision, 5,
        sizeof(cl_mem), &clMassRadius );
    error |= clSetKernelArg( clKernelCollision, 6,
        sizeof(cl_mem), &sim_clPosition );
    error |= clSetKernelArg( clKernelCollision, 7,
        sizeof(cl_mem), &clForce );
    checkErr( error, "clSetKernelArg (clKernelCollision)" );
    
    // getting corrected global size for clKernelCollision done each step
    // need to hold onto deviceId to be able to do the calculation
    deviceId = sim_deviceId;
}

// destructor
forceCollisionsUniformSphere::~forceCollisionsUniformSphere( )
{
    delete[] positionHash;
    delete[] collisionCell;
    delete[] fZero;
    
    clReleaseKernel( clKernelGetHash );
    clReleaseKernel( clKernelCollision );
    clReleaseMemObject( clMassRadius );
    clReleaseMemObject( clBoundingRadius );
    clReleaseMemObject( clPositionHash );
    clReleaseMemObject( clCollisionCell );
}

// calculating the force
void forceCollisionsUniformSphere::calculateForce( )
{
    cl_int error = 0;
    
    // get hash information from current positions
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // we are adding it to our command queue
            clKernelGetHash, // the kernel we are running (integrate)
            1, // we have a 1 dimensional work dimension
            NULL, // we aren't using any offsets for the work dimensions
            &clKernelGetHashGlobalSize, // our global work space is the number of particles
            NULL, // local item size chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel (clKernelGetHash)");

    // TODO: parallelize the sorting and cell detection stages

    // get the hash information from the device
    checkErr(
        clEnqueueReadBuffer( queue, clPositionHash, CL_TRUE, 0,
            numParticles*8 * sizeof( POSITION_HASH ), positionHash,
            0, NULL, NULL ),
        "clEnqueueReadBuffer (positionHash)" );

    // sort the hash information in descending order (zeroes at the end) with respect to hash
    // TODO
    std::stable_sort( positionHash, positionHash + (numParticles*8), comparisonPOSITION_HASH );

    // get the information about where the collision cells begin and end
    cl_ulong numCollisionCells = 0;
    cl_ulong last_hash = 0;
    bool wasHomeCell = false; // was the previous step a home cell?
    bool homeCellSet = false;
    for (cl_ulong i = 0; i < 8*numParticles; ++i)
    {
        cl_ulong cur_hash = positionHash[i].hash;
        if (last_hash != cur_hash)
        {
            // we have a new hash
            last_hash = cur_hash;

            // if we have a previously set homeCell set, we must set the previous end to this index
            if (homeCellSet)
            {
                collisionCell[numCollisionCells - 1].phantomCellEnd = i;
                // if our last was a home cell, we also need to set the phantom start to here as well
                if (wasHomeCell)
                {
                    collisionCell[numCollisionCells - 1].phantomCellStart = i;
                }
                homeCellSet = false;
            }

            // if the current hash is zero, we stop
            if (cur_hash == 0)
            {
                break;
            }

            // Is the particle we are looking at in its home cell?
            if (positionHash[i].homeCellType == positionHash[i].thisCellType)
            {
                ++numCollisionCells; // we are on a new collision cell
                collisionCell[numCollisionCells - 1].homeCellStart = i; // mark this beginning
                wasHomeCell = true; // mark that we are currently on home cells
                homeCellSet = true;
            }
        }
        else // we have the same value as before
        {
            if (wasHomeCell && positionHash[i].homeCellType != positionHash[i].thisCellType)
            {
                // this is the first phantom cell for the current hash
                collisionCell[numCollisionCells - 1].phantomCellStart = i;
                wasHomeCell = false;
            }
        }

        // the current hash becomes the last hash
        last_hash = cur_hash;
    }

    // send the sorted hash table to device
    checkErr(
        clEnqueueWriteBuffer( queue, clPositionHash, CL_TRUE,
            0, numParticles * 8 * sizeof( POSITION_HASH ),
            positionHash, 0, NULL, NULL),
        "clEnqueueWriteBuffer (positionHash)"
        );

    // send the information about the collision cells to device
    checkErr(
        clEnqueueWriteBuffer( queue, clCollisionCell, CL_TRUE,
            0, numParticles * sizeof( COLLISION_CELL ),
            collisionCell, 0, NULL, NULL),
        "clEnqueueWriteBuffer (collisionCell)"
        );

    // make sure that the force vectors are zeroed out
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clForce, // write to clForce
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            columns * numParticles * sizeof( CS_REAL3 ), // size of the memory we are sending
            fZero, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (fZero)"
        );

    clKernelCollisionGlobalSize = getKernelPaddedGlobalSize(
        deviceId, clKernelCollision, numCollisionCells );
    error = clSetKernelArg( clKernelCollision, 2,
        sizeof(cl_ulong), &numCollisionCells );
    checkErr( error, "clSetKernelArg (numCollisionCells)" );
    
    // calculate the forces
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // we are adding it to our command queue
            clKernelCollision, // the kernel we are running (integrate)
            1, // we have a 1 dimensional work dimension
            NULL, // we aren't using any offsets for the work dimensions
            &clKernelCollisionGlobalSize, // our global work space is the number of particles
            NULL, // local item size chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel");

    // the host and device arrays are no longer in synchronization
    updated = false;
}

bool forceCollisionsUniformSphere::comparisonPOSITION_HASH( const POSITION_HASH & left, 
    const POSITION_HASH & right )
{
    return left.hash > right.hash;
}
