// forceCollisionsUniformSphere2.cl
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II
// OpenCL file for forceCollisionsUniformSphere2
// following example from NVidia OpenCL example

#include <chromoTypes.cl>
   

// function that assigns a grid coordinate to a given position coordinate
int3 getGridCoordinate( CS_REAL3 posCoordinate, CS_REAL3 cellDimensions )
{
    return convert_int3( posCoordinate / cellDimensions );
}

// function that assigns a positive value from a grid coordinate (zero value is thus invalid)
uint getGridHash( int3 gridCoordinate, uint3 gridScale, uint maxHash )
{
    return convert_uint(
        gridCoordinate.x * gridScale.x
        + gridCoordinate.y * gridScale.y
        + gridCoordinate.z * gridScale.z)
        % maxHash;
}

// kernel function for getting the hashes of the position vectors
__kernel void calculateHashedPosition(
    const uint3 gridScale,
    const CS_REAL3 cellDimensions,
    const uint maxHash,
    const ulong numParticles,
    __global CS_REAL3 * position, // INPUT: the current position 
    __global uint2 * hashedPositions // OUTPUT: hash array of length numParticles (.x is hash, .y is index)
    )
{
    uint ndx = get_global_id(0); // we can worry about groups later

    if (ndx >= numParticles)
    {
        return;
    }

    CS_REAL3 pos = position[ndx];
    
    int3 gridCoord = getGridCoordinate( pos, cellDimensions );
    uint gridHash = getGridHash( gridCoord, gridScale, maxHash );

    // hashedPositions[ndx].hash = gridHash;
    // hashedPositions[ndx].index = ndx;
    uint2 retval = (uint2) ( gridHash, ndx );
    hashedPositions[ndx] = retval;
}

__kernel void findCellBounds(
    __global uint2 * hashes, // input hashes to read in
    const ulong numParticles, // the number of particles
    __local uint * localHashes, // local store of hashes for better memory access, get_group_size(0)+1 elements
    __global uint * collisionCellStart, // the start index in the hashes array of a given hash value
    __global uint * collisionCellEnd // the end index in the hashes array of a given hash value
    )
{
    uint hash;

    uint ndx = get_global_id(0);
    uint localNdx = get_local_id(0);

    if (ndx < numParticles)
    {
        hash = hashes[ndx].x;

        // load hash into local memory, storing this value into next index
        // in local memory
        localHashes[localNdx + 1] = hash;
        // we are trying to make it so that localHashes[localNdx] is the
        // hash that precedes hash = hashes[ndx]. We must access global
        // memory to do this if localNdx is zero:
        if (ndx > 0 && localNdx == 0)
        {
            localHashes[0] = hashes[ndx - 1].x;
        }
    }

    // synchronize these changes in local memory
    barrier(CLK_LOCAL_MEM_FENCE);

    // quit if ndx invalid
    if (ndx >= numParticles)
    {
        return;
    }

    // deal with the case where ndx is zero
    if (ndx == 0)
    {
        collisionCellStart[hash] = 0;
    }
    else // we have a previous value too
    {
        // if current hash value and previous hash value are different, mark it
        if (hash != localHashes[localNdx])
        {
            collisionCellStart[hash] = collisionCellEnd[localHashes[localNdx]] = ndx;
        }
    }

    // deal with the case where ndx is numParticles-1
    if (ndx == numParticles - 1)
    {
        collisionCellEnd[hash] = numParticles;
    }
}
    

// kernel function for using sorted hashes to calculate the collision forces
__kernel void applyCollisions(
    const CS_REAL collisionSpringConstant,
    const uint3 gridScale,
    const CS_REAL3 cellDimensions,
    const uint maxHash,
    const ulong numParticles,
    __global uint * collisionCellStart,
    __global uint * collisionCellEnd,
    __global uint2 * hashes,
    __global CS_REAL * massRadius,
    __global CS_REAL3 * position,
    __global CS_REAL3 * fCollision
    )
{
    uint ndx = get_global_id(0);

    if (ndx >= numParticles)
    {
        return;
    }

    CS_REAL3 pos = position[ndx];
    int3 gridCoord = getGridCoordinate( pos, cellDimensions );

    CS_REAL3 cforce = 0;

    CS_REAL fixedOffset = massRadius[ndx];
    for (int dx = -1; dx <= 1; ++dx)
    {
        for (int dy = -1; dy <= 1; ++dy)
        {
            for (int dz = -1; dz <= 1; ++dz)
            {
                uint hash = getGridHash( gridCoord + (int3) (dx, dy, dz), gridScale, maxHash );
                uint start = collisionCellStart[hash];
                uint end = collisionCellEnd[hash];
                if (start >= numParticles)
                {
                    continue;
                }

                for (uint i = start; i < end; ++i)
                {
                    uint ndx2 = hashes[i].y;
                    if (ndx == ndx2)
                    {
                        continue;
                    }
                    // what's the displacement vector between the two positions
                    CS_REAL3 displacement = pos - position[ndx2];
                    // is the displacement longer than the two radii plus a small buffer?
                    CS_REAL contactDistance = fixedOffset + massRadius[ndx2] - length( displacement );
                    // if contact distance > 0, then we have contact, thus
                    if (contactDistance > 0)
                    {
                        cforce += collisionSpringConstant * contactDistance * normalize( displacement );
                    }
                }
            }
        }
    }

    // apply force
    fCollision[ndx] = cforce;
}

// kernel function for using sorted hashes to calculate the collision forces
__kernel void applyCollisionsAndJacobian(
    const CS_REAL collisionSpringConstant,
    const uint3 gridScale,
    const CS_REAL3 cellDimensions,
    const uint maxHash,
    const ulong numParticles,
    __global uint * collisionCellStart,
    __global uint * collisionCellEnd,
    __global uint2 * hashes,
    __global CS_REAL * massRadius,
    __global CS_REAL3 * position,
    __global CS_REAL3 * fCollision,
    __global CS_REAL3 * collisionJPosX,
    __global CS_REAL3 * collisionJPosY,
    __global CS_REAL3 * collisionJPosZ,
    __global ulong * collisionInteractionIndex, // what is being interacted with
    __global ulong * collisionNumInteractions // how many we have interacted with so far
    )
{
    uint ndx = get_global_id(0);

    if (ndx >= numParticles)
    {
        return;
    }

    CS_REAL3 pos = position[ndx];
    int3 gridCoord = getGridCoordinate( pos, cellDimensions );

    CS_REAL3 cforce = 0;

    // stuff for implicit
    ulong numInteractions = 1; // how many interactions do we currently have? (say self at first)
    // keep track of (negative of) self position Jacobian
    CS_REAL3 selfJPosX = 0;
    CS_REAL3 selfJPosY = 0;
    CS_REAL3 selfJPosZ = 0;

    CS_REAL fixedOffset = massRadius[ndx];
    for (int dx = -1; dx <= 1; ++dx)
    {
        for (int dy = -1; dy <= 1; ++dy)
        {
            for (int dz = -1; dz <= 1; ++dz)
            {
                uint hash = getGridHash( gridCoord + (int3) (dx, dy, dz), gridScale, maxHash );
                uint start = collisionCellStart[hash];
                uint end = collisionCellEnd[hash];
                if (start >= numParticles)
                {
                    continue;
                }

                for (uint i = start; i < end; ++i)
                {
                    uint ndx2 = hashes[i].y;
                    if (ndx == ndx2)
                    {
                        continue;
                    }
                    // displacement vector
                    CS_REAL3 d = position[ndx2] - pos;
                    // distance from this
                    CS_REAL dLength = length(d);
                    // unitless scale of if we're stretching or compressed (collided)
                    CS_REAL scale = 1 - (fixedOffset + massRadius[ndx2]) / dLength;
                    if (scale < 0) // we have a collision
                    {
                        // compute outer product of displacement with itself
                        CS_REAL3 outerX = d.x * d / (dLength*dLength);
                        CS_REAL3 outerY = d.y * d / (dLength*dLength);
                        CS_REAL3 outerZ = d.z * d / (dLength*dLength);
                        // add the force
                        cforce += collisionSpringConstant * scale * d;
                        // compute the Jacobian
                        CS_REAL3 jx = collisionSpringConstant * ( outerX + scale * ( ((float3) (1,0,0)) - outerX ) );
                        CS_REAL3 jy = collisionSpringConstant * ( outerY + scale * ( ((float3) (0,1,0)) - outerY ) );
                        CS_REAL3 jz = collisionSpringConstant * ( outerZ + scale * ( ((float3) (0,0,1)) - outerZ ) );
                        // add to negative of self Jacobian
                        selfJPosX += jx;
                        selfJPosY += jy;
                        selfJPosZ += jz;
                        // update cross term (ndx to ndx2)
                        // TODO: Guard this to not overwrite memory when too many interactions
                        // TODO: Return this info and handle it on the CPU code
                        ulong cross_ndx = numInteractions*numParticles + ndx;
                        collisionJPosX[cross_ndx] = jx;
                        collisionJPosY[cross_ndx] = jy;
                        collisionJPosZ[cross_ndx] = jz;
                        collisionInteractionIndex[cross_ndx] = ndx2;
                        // we have just added one more interaction
                        ++numInteractions;
                    }
                }
            }
        }
    }

    // apply force
    fCollision[ndx] = cforce;
    // update the number of interactions (number of collisions + 1)
    collisionNumInteractions[ndx] = numInteractions;
    // update the self collision Jacobian
    collisionJPosX[ndx] = -selfJPosX;
    collisionJPosY[ndx] = -selfJPosY;
    collisionJPosZ[ndx] = -selfJPosZ;
}