// forceHinges.cl
// OpenCL file for forceHingesCL
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II


#include <chromoTypes.cl>


typedef struct simHINGE {
    ulong m1, m2, m3; // indices for particles spring attached to
    ulong ndx1, ndx2, ndx3; // indices for force vector
    CS_REAL hingeConstant;
} simHINGE;


// TODO: this hinge force doesn't satisfy Newton's 3rd law (extra torque)
__kernel void calculateHingeForce( // calculates the force exerted by each of the hinges
    __global simHINGE * hinge, // all the hinges
    __global CS_REAL3 * position, // all position of all the particles
    __global CS_REAL3 * fHinge, // the force applied on the masses by the hinges
    const ulong numParticles, // the number of particles in the simulation
    const ulong numHinges // the total number of hinges
    )
{
    ulong ndx = get_global_id(0); // we can worry about work groups later

    if (ndx >= numHinges) {
        return;
    }

    // indices of the attached masses
    ulong m1 = hinge[ndx].m1;
    ulong m2 = hinge[ndx].m2;
    ulong m3 = hinge[ndx].m3;
    ulong ndx1 = hinge[ndx].ndx1;
    ulong ndx2 = hinge[ndx].ndx2;
    ulong ndx3 = hinge[ndx].ndx3;

    // locations of the attached masses
    CS_REAL3 v1 = position[m1];
    CS_REAL3 v2 = position[m2];
    CS_REAL3 v3 = position[m3];

    // adopting the same terminology as in "HingeForceField.inl"...
    
    // calculate unit directions out from center of hinge (m2)
    CS_REAL3 left = normalize( v1 - v2 );
    CS_REAL3 right = normalize( v3 - v2 );

    // calculate the angle (in degrees) of the hinge
    //CS_REAL angle = fabs( 1 - acospi( dot( left, right ) ) );

    // calculate the angle (in radians) of the hinge
    CS_REAL angle = fabs( M_PI - acos( dot( left, right ) ) );

    // calculate the force on the middle of the hinge
    CS_REAL3 f = hinge[ndx].hingeConstant * angle * normalize( left + right );

    if ( any( isnan( f ) ) || angle < 0.001 ) {
        f = (CS_REAL3) ( 0, 0, 0 );
    }
    
    // apply to our three masses appropriately
    fHinge[m1 + ndx1*numParticles] = -f/2;
    fHinge[m2 + ndx2*numParticles] = f;
    fHinge[m3 + ndx3*numParticles] = -f/2;
}
