// forceHingesCL.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "forceHingesCL.h"

// what's the name of this force?
const char * forceHingesCL::getName( )
{
    return "Force Hinge";
}

// constructor
forceHingesCL::forceHingesCL(
    std::string sim_CLdir, // the path of directory with our openCL source code
    cl_context sim_context, // OpenCL context passed in from simulation
    cl_command_queue sim_queue, // OpenCL queue passed in from simulation,
    cl_ulong sim_numParticles, // number of particles in the simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    cl_device_id sim_deviceId, // the deviceId we are running on
    cl_mem sim_clPosition, // the openCL pointer for the particles' positions
    HINGE_node * hh // the spring nodes from the configuration (MassMesh)
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, sim_totalMaxInteractions, true )
{
    simHINGE * hinge; // temporary array containing information about the hinges

    // get the total number of hinges
    numHinges = 0;
    
   // get node to iterate over the hinges
    HINGE_node *h_it = hh;
    while (h_it != NULL)
    {
        h_it = h_it->next;
        numHinges++;
    }
    // initialize array for storing springs that will be put to the device
    hinge = new simHINGE[numHinges];
    // hCount will keep track of how many hinges are attached to each particle
    cl_ulong * hCount = new cl_ulong[numParticles];
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        hCount[i] = 0;
    }
    // loop over the hinges and give them values.
    h_it = hh;
    for (cl_ulong k = 0; k < numHinges; ++k)
    {
        // store indices of connected masses
        cl_ulong m1 = h_it->m1->sofaID;
        hinge[k].m1 = m1; // the hinge has one end at m1
        hinge[k].ndx1 = hCount[m1]; // how many times has m1 had a hinge?
        ++hCount[m1];
        cl_ulong m2 = h_it->m2->sofaID;
        hinge[k].m2 = m2; // the hinge has its pivot at m2
        hinge[k].ndx2 = hCount[m2]; // how many times has m2 had a hinge?
        ++hCount[m2];
        cl_ulong m3 = h_it->m3->sofaID;
        hinge[k].m3 = m3; // the hinge has one end at m3
        hinge[k].ndx3 = hCount[m3]; // how many times has m3 had a hinge?
        ++hCount[m3];
        
        // get remaining information about the hinge, constant in units of radians.
        hinge[k].hingeConstant = static_cast<CS_REAL>(h_it->k);

        // hinge constant is in units of radians, so we just store it.
        //hinge[k].hingeConstant *= 180;
        
        // next iteration please
        h_it = h_it->next;
    }
    // determine the maximum number of hinges we had attached to any one hinge
    columns = 1; // even if we have no hinges, we don't want to have an empty array
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        if (hCount[i] > columns)
        {
            columns = hCount[i];
        }
    }
    // create a force vector and set it to zero on host and device
    initForcesZero( );
    // if we have any hinges to speak of,
    if (numHinges > 0)
    {
        cl_int error = 0;

        // send to clHinge on device
        clHinge = clCreateBuffer( context, CL_MEM_READ_ONLY,
            numHinges * sizeof( simHINGE ), NULL, &error );
        checkErr( error, "clCreateBuffer" );
        checkErr(
            clEnqueueWriteBuffer( queue, clHinge, CL_TRUE, 0,
                numHinges * sizeof( simHINGE ), hinge,
                0, NULL, NULL),
            "clEnqueueWriteBuffer (clHinge/hinge)"
            );

        // load the program "forceHinges.cl"
        loadProgram( sim_deviceId, "forceHinges.cl" );

        // load the hinge calculation kernel from program
        clKernelHinge = clCreateKernel( program, "calculateHingeForce", &error );
        checkErr( error, "clCreateKernel" );

        // pass arguments into kernel
        error = clSetKernelArg( clKernelHinge, 0, sizeof(cl_mem), &clHinge );
        error |= clSetKernelArg( clKernelHinge, 1, sizeof(cl_mem), &sim_clPosition );
        error |= clSetKernelArg( clKernelHinge, 2, sizeof(cl_mem), &clForce );
        error |= clSetKernelArg( clKernelHinge, 3, sizeof(cl_ulong), &numParticles );
        error |= clSetKernelArg( clKernelHinge, 4, sizeof(cl_ulong), &numHinges );
        checkErr( error, "clSetKernelArg" );

        // get optimal global item size
        clKernelHingeGlobalSize = getKernelPaddedGlobalSize( 
            sim_deviceId, clKernelHinge, numHinges );
    }

    // free temporary memory we allocated
    delete[] hCount;
    delete[] hinge;
}

// destructor
forceHingesCL::~forceHingesCL( )
{
    if (numHinges > 0)
    {
        clReleaseKernel( clKernelHinge );
        clReleaseMemObject( clHinge );
    }
}

// calculating the force
void forceHingesCL::calculateForce( )
{
    if (numHinges > 0)
    {
        checkErr(
            clEnqueueNDRangeKernel( // run the kernel
                queue, // use the queue
                clKernelHinge, // the kernel we are running
                1, // we have a 1 dimensional work dimension
                NULL, // we aren't using any offsets for our work dimensions
                &clKernelHingeGlobalSize, // our global work space size is defined above
                NULL, // local items chosen by implementation
                0, NULL, NULL // we aren't waiting for any events
                ),
            "clEnqueueNDRangeKernel");

        // the host and device arrays are no longer in synchronization
        updated = false;
    }
}
