// forceMassDamping.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "forceMassDamping.h"

// what's the name of this force?
const char * forceMassDamping::getName( )
{
    return "Force Mass Damping";
}

// destructor
forceMassDamping::~forceMassDamping( )
{
    clReleaseKernel( clKernelMassDrag );
    clReleaseMemObject( clMassDragConstant );
}

// constructor
forceMassDamping::forceMassDamping(
    std::string sim_CLdir, // the path of our directory
    cl_context sim_context, // OpenCL context
    cl_command_queue sim_queue, // OpenCL command queue
    cl_ulong sim_numParticles, // number of particles in simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    cl_device_id sim_deviceId, // the deviceId we are running on
    cl_mem sim_clVelocity, // the OpenCL pointer for particles' velocities
    CS_REAL * mass, // the mass of the particles
    CS_REAL * massDamping // the mass damping constants (drag = mass * massDamping)
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, sim_totalMaxInteractions, true )
{
    // we have one column
    columns = 1;
    // initialize zeroed out force vector
    initForcesZero( );
    // mass damping only comes from self interaction
    jacobianMaxInteractions = 1;
    // initialize a zeroed out Jacobian
    initJacobianZero( );

    // handle errors with OpenCL API
    cl_int error;
    
    // set up the drag buffer (clMassDragConstant)
    CS_REAL * drag = new CS_REAL[numParticles];
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        drag[i] = mass[i] * massDamping[i]; // in the future, we may change mass damping to just drag...
    }
    // push this to the device
    clMassDragConstant = clCreateBuffer(
        context, // put into CL context
        CL_MEM_READ_ONLY, // the buffer will not be changed in the kernel
        numParticles * sizeof( CS_REAL ), // allocate space for the springs
        NULL, &error );
    checkErr( error, "clCreateBuffer (clMassDragConstant)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clMassDragConstant, CL_TRUE,
            0, numParticles * sizeof( CS_REAL ), drag, 0, NULL, NULL),
        "clEnqueueWriteBuffer (clMassDragConstant)" );

    // if we allocated a Jacobian, set it up as a diagonal matrix wrt velocity
    if (clJacobianLoaded)
    {
        // update the Jacobian in the appropriate way
        // how many interactions per particle?
        cl_ulong * countInteraction = new cl_ulong[numParticles];
        // what is the interaction matrix and with whom? (since only one per particle, same number)
        CS_REAL3 * idX = new CS_REAL3[numParticles];
        CS_REAL3 * idY = new CS_REAL3[numParticles];
        CS_REAL3 * idZ = new CS_REAL3[numParticles];
        cl_ulong * interactionNdx = new cl_ulong[numParticles];
        for (cl_ulong i = 0; i < numParticles; ++i)
        {
            // the velocity Jacobian is the identity matrix scaled by drag
            idX[i].s[0] = - drag[i]; idX[i].s[1] = 0; idX[i].s[2] = 0;
            idY[i].s[0] = 0; idY[i].s[1] = - drag[i]; idY[i].s[2] = 0;
            idZ[i].s[0] = 0; idZ[i].s[1] = 0; idZ[i].s[2] = - drag[i];
            // we are interacting on the diagonal
            interactionNdx[i] = i;
            // we have exactly one interaction for this particle
            countInteraction[i] = 1;
        }
        // pass these into our Jacobian
        checkErr( 
            clEnqueueWriteBuffer( queue, clJacobian.numInteractions,
                CL_TRUE, 0, numParticles * sizeof( cl_ulong ),
                countInteraction, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (numInteractions)" );
        checkErr( 
            clEnqueueWriteBuffer( queue, clJacobian.interactionParticleIndex,
                CL_TRUE, 0, numParticles * sizeof( cl_ulong ),
                interactionNdx, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (interactionIndex)" );
        checkErr( 
            clEnqueueWriteBuffer( queue, clJacobian.j_VelX,
                CL_TRUE, 0, numParticles * sizeof( CS_REAL3 ),
                idX, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (j_VelX)" );
        checkErr( 
            clEnqueueWriteBuffer( queue, clJacobian.j_VelY,
                CL_TRUE, 0, numParticles * sizeof( CS_REAL3 ),
                idY, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (j_VelY)" );
        checkErr( 
            clEnqueueWriteBuffer( queue, clJacobian.j_VelZ,
                CL_TRUE, 0, numParticles * sizeof( CS_REAL3 ),
                idZ, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (j_VelZ)" );

        // delete remaining memory
        delete[] countInteraction;
        delete[] idX;
        delete[] idY;
        delete[] idZ;
        delete[] interactionNdx;
    }

    // delete the temporary drag array
    delete[] drag;

    // we just have to set up the kernel to calculate the drag force at each step

    // now, load the program so we can build the kernel
    loadProgram( sim_deviceId, "forceMassDamping.cl" );

    // load the drag force calculation kernel from program
    clKernelMassDrag = clCreateKernel( program, "calculateDragForce", &error );
    checkErr( error, "clCreateKernel" );

    // pass arguments into the kernel
    error = clSetKernelArg( clKernelMassDrag, 0, sizeof(cl_mem), &clMassDragConstant );
    error |= clSetKernelArg( clKernelMassDrag, 1, sizeof(cl_mem), &sim_clVelocity );
    error |= clSetKernelArg( clKernelMassDrag, 2, sizeof(cl_mem), &clForce );
    error |= clSetKernelArg( clKernelMassDrag, 3, sizeof(cl_ulong), &numParticles );

    // set global item size
    clKernelMassDragGlobalSize = getKernelPaddedGlobalSize(
        sim_deviceId, clKernelMassDrag, numParticles );
}

// calculate force with Jacobian
void forceMassDamping::calculateForceAndJacobian( )
{
    // we don't actually need to update the Jacobian
    calculateForce( );
}

// calculate the force
void forceMassDamping::calculateForce( )
{
    // launch kernel
    checkErr(
        clEnqueueNDRangeKernel( // run on the kernel
            queue,
            clKernelMassDrag, // the kernel we are running
            1, // 1d work dimension
            NULL,
            &clKernelMassDragGlobalSize, // our global work space
            NULL, // local items are chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel");

    // no longer in synchronization
    updated = false;
}