// forceRandomGPU.cl
// OpenCL file for forceRandomGPU
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include <chromoTypes.cl>
#include <Random123/threefry.h>
#include <Random123/u01.h>

// Random123 described in the paper
// http://www.thesalmons.org/john/random123/papers/random123sc11.pdf
// using Random123 was helped by the following webpages
// http://efreedom.net/Question/1-11255266/Random123-Generating-Random-Numbers-Opencl-Using-Visual-Studio
// http://efreedom.net/Question/1-11268023/Random-Numbers-OpenCL-Using-Random123
// http://www.thesalmons.org/john/random123/releases/1.06/docs/

// box muller transform to take 
// CS_REAL2 uniformly distributed on (0,1] x (0,1]
// to CS_REAL2 independent standard normal random variables
CS_REAL2 boxMullerTransform( CS_REAL2 U ) // the input 2d uniform rv
{
    // get the individual components
    CS_REAL u1 = U.x;
    CS_REAL u2 = U.y;

    // we solve for sqrt(-2*ln(u1))*cos(2*pi*u2)
    // and sqrt(-2*ln(u1))*sin(2*pi*u2)

    // get the term R = sqrt(-2*ln(u1))
    CS_REAL R = sqrt( -2 * log( u1 ) );

    // get the cos and sin terms
    CS_REAL cs, sn;
    cs = cospi( 2 * u2 ); // (cospi(x) = cos(pi*x))
    sn = sinpi( 2 * u2 ); // (sinpi(x) = sin(pi*x))

    // combine the values to get our normally distributed values and return
    CS_REAL2 Z;
    Z.x = R * cs;
    Z.y = R * sn;

    return Z;
}

// the kernel function for creating random forces
__kernel void getRandomForce(
    __global threefry4x32_key_t * randomKey, // the unchanged seed for some particle
    __global threefry4x32_ctr_t * randomCounter, // the incremented counter for each particle
    const CS_REAL theRandomForce, // the random force parameter
    __global CS_REAL3 * fBrownian, // output array of random forces
    const ulong numParticles // the total number of particles we are getting forces for
    )
{
    ulong ndx = get_global_id(0);

    // don't run if we were extra padding
    if (ndx >= numParticles) {
        return;
    }

    // make private copies of key and counter
    threefry4x32_key_t key = randomKey[ndx];
    threefry4x32_ctr_t counter = randomCounter[ndx];

    // increment counter -- XXX I think that since it's unsigned, we get automatic wrapping back to zero, right?
    ++counter.v[0];
    if (counter.v[0] == 0) {
        ++counter.v[1];
        if (counter.v[1] == 0) {
            ++counter.v[2];
            if (counter.v[2] == 0) {
                ++counter.v[3];
            }
        }
    } 
    // I believe that this should make it so that we only repeat counters
    // every 2**128

    // get four random integers back using current state of counter and key
    union {
        threefry4x32_ctr_t c;
        int4 i;
    } u;
    u.c = threefry4x32( counter, key );

    // get out four random floats on (0,1] from this
    // XXX we need to make this appropriate when we are using floats vs doubles...
    CS_REAL2 uniform12, uniform34;
    // u01_open_closed_32_24 converts 32 bit integer to 24 bit float on (0,1]
    uniform12.x = u01_open_closed_32_24( u.i.x );
    uniform12.y = u01_open_closed_32_24( u.i.y );
    uniform34.x = u01_open_closed_32_24( u.i.z );
    uniform34.y = u01_open_closed_32_24( u.i.w );

    // get four random normally distributed values from this
    CS_REAL2 normal12, normal34;
    normal12 = boxMullerTransform( uniform12 );
    normal34 = boxMullerTransform( uniform34 );

    // use three of the values, throw away the fourth
    CS_REAL3 normal3D;
    normal3D.x = normal12.x;
    normal3D.y = normal12.y;
    normal3D.z = normal34.x;

    // update the force
    CS_REAL3 force = theRandomForce * normal3D;
    fBrownian[ndx] = force;

    // update the counter in global memory
    randomCounter[ndx] = counter;
}

// the kernel function for creating random forces
__kernel void getRandomIntegers(
    __global threefry4x32_key_t * randomKey, // the unchanged seed for some particle
    __global threefry4x32_ctr_t * randomCounter, // the incremented counter for each particle
    __global int4 * randomNumbers, // output array of random forces
    const ulong numParticles // the total number of particles we are getting forces for
    )
{
    ulong ndx = get_global_id(0);

    // don't run if we were extra padding
    if (ndx >= numParticles) {
        return;
    }

    // make private copies of key and counter
    threefry4x32_key_t key = randomKey[ndx];
    threefry4x32_ctr_t counter = randomCounter[ndx];

    // increment counter -- XXX I think that since it's unsigned, we get automatic wrapping back to zero, right?
    ++counter.v[0];
    if (counter.v[0] == 0) {
        ++counter.v[1];
        if (counter.v[1] == 0) {
            ++counter.v[2];
            if (counter.v[2] == 0) {
                ++counter.v[3];
            }
        }
    } 
    // I believe that this should make it so that we only repeat counters
    // every 2**128

    // get four random integers back using current state of counter and key
    union {
        threefry4x32_ctr_t c;
        int4 i;
    } u;
    u.c = threefry4x32( counter, key );

    randomNumbers[ndx] = u.i;

    // update the counter in global memory
    randomCounter[ndx] = counter;
}
