// forceSpringsCL.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "forceSpringsCL.h"

// what's the name of this force?
const char * forceSpringsCL::getName( )
{
    return "Force Spring";
}

// constructor
forceSpringsCL::forceSpringsCL(
    std::string sim_CLdir, // the path of directory with our openCL source code
    cl_context sim_context, // OpenCL context passed in from simulation
    cl_command_queue sim_queue, // OpenCL queue passed in from simulation,
    cl_ulong sim_numParticles, // number of particles in the simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    cl_device_id sim_deviceId, // the deviceId we are running on
    cl_mem sim_clPosition, // the openCL pointer for the particles' positions
    cl_mem sim_clVelocity, // the openCL pointer for the particles' velocities
    SPRING_node * sh, // the spring nodes from the configuration (MassMesh)
    CS_REAL springDamping // the springDrag = springDamping * springConstant (same definition as in SOFA)
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, sim_totalMaxInteractions, true )
{
    simSPRING * spring; // temporary array containing information about the springs

    // get the total number of springs
    numSprings = 0;
    SPRING_node *s_it = sh;
    while (s_it != NULL)
    {
        s_it = s_it->next;
        ++numSprings;
    }
    // initialize array for storing springs that will be put to the device
    spring = new simSPRING[numSprings];
    // sCount will keep track of how many springs are attached to each particle
    cl_ulong * sCount = new cl_ulong[numParticles];
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        sCount[i] = 0;
    }
    // loop over the springs and give them values.
    s_it = sh;
    for (cl_ulong j = 0; j < numSprings; ++j)
    {
        // store indices of connected masses
        cl_ulong m1 = s_it->m1->sofaID;
        spring[j].m1 = m1; // this spring is attached to m1
        spring[j].ndx1 = sCount[m1]; // how many times has m1 had a spring?
        ++sCount[m1];
        cl_ulong m2 = s_it->m2->sofaID;
        spring[j].m2 = m2;
        spring[j].ndx2 = sCount[m2];
        ++sCount[m2];
        
        // get remaining information about the spring
        spring[j].restLength = static_cast<CS_REAL>(s_it->rest);
        spring[j].springConstant = static_cast<CS_REAL>(s_it->k);
        spring[j].springDrag = spring[j].springConstant * springDamping;
        
        // next iteration, please
        s_it = s_it->next;
    }
    // determine the maximum number of springs we had attached to any one mass
    columns = 1; // even if we have no springs, we don't want to have empty array
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        if (sCount[i] > columns)
        {
            columns = sCount[i];
        }
    }
    // create a force vector and set it to zero on host and device
    initForcesZero( );
    // if we have any springs to speak of, 
    if (numSprings > 0)
    {
        cl_int error = 0;

        // send to clSpring on device
        clSpring = clCreateBuffer( context, // put into our class' context
            CL_MEM_READ_ONLY, // make the buffer read only
            numSprings * sizeof( simSPRING ), // allocate space for the springs
            NULL, &error);
        checkErr( error, "clCreateBuffer" );
        checkErr(
            clEnqueueWriteBuffer( queue, clSpring, CL_TRUE, 0,
                numSprings * sizeof( simSPRING ), spring,
                0, NULL, NULL),
            "clEnqueueWriteBuffer (clSpring/spring)"
            );
        
        // load the program "forceSprings.cl"
        loadProgram( sim_deviceId, "forceSprings.cl" );

        // load the spring calculation kernel from program
        clKernelSpring = clCreateKernel( program, "calculateSpringForce", &error );
        checkErr( error, "clCreateKernel" );

        // pass arguments into kernel
        error = clSetKernelArg( clKernelSpring, 0, sizeof(cl_mem), &clSpring );
        error |= clSetKernelArg( clKernelSpring, 1, sizeof(cl_mem), &sim_clPosition );
        error |= clSetKernelArg( clKernelSpring, 2, sizeof(cl_mem), &sim_clVelocity );
        error |= clSetKernelArg( clKernelSpring, 3, sizeof(cl_mem), &clForce );
        error |= clSetKernelArg( clKernelSpring, 4, sizeof(cl_ulong), &numParticles );
        error |= clSetKernelArg( clKernelSpring, 5, sizeof(cl_ulong), &numSprings );
        checkErr( error, "clSetKernelArg" );

        // get optimal global item size
        clKernelSpringGlobalSize = getKernelPaddedGlobalSize( 
            sim_deviceId, clKernelSpring, numSprings );

        // set up for calculations of the Jacobians as well
        jacobianMaxInteractions = columns + 1;

        // set up the buffers for the Jacobian
        initJacobianZero( );
        // did we successfully allocate a Jacobian?
        if (clJacobianLoaded)
        {
            // looks like it so go ahead and fill the Jacobian
            // with what we can do for now
            // first, we increment sCount by 1 for each index and note that it serves perfectly
            // for numInteractions.
            for (int i = 0; i < numParticles; ++i)
            {
                ++sCount[i];
            }
            checkErr( clEnqueueWriteBuffer( queue, clJacobian.numInteractions, CL_TRUE,
                0, numParticles * sizeof( cl_ulong ), sCount, 0, NULL, NULL ),
                "clEnqueueWriteBuffer (numInteractions)" );
            // then, we set up interactionParticleIndex by setting the diagonal to the first column/row
            // and setting the others according to ndx1/ndx2 per spring
            cl_ulong * interactionIndex = new cl_ulong[numParticles * jacobianMaxInteractions];
            // set up implicit spring velocity damping: constant over the simulation
            CS_REAL3 * jvX = new CS_REAL3[numParticles * jacobianMaxInteractions];
            CS_REAL3 * jvY = new CS_REAL3[numParticles * jacobianMaxInteractions];
            CS_REAL3 * jvZ = new CS_REAL3[numParticles * jacobianMaxInteractions];
            // zero out the velocity Jacobian
            for (cl_ulong i = 0; i < numParticles * jacobianMaxInteractions; ++i)
            {
                for (int j = 0; j < 3; ++j)
                {
                    jvX[i].s[j] = 0;
                    jvY[i].s[j] = 0;
                    jvZ[i].s[j] = 0;
                }
            }
            // loop through the springs and apply them to our Jacobian
            for (cl_ulong i = 0; i < numParticles; ++i)
            {
                interactionIndex[i] = i; // the diagonal
            }
            for (cl_ulong s = 0; s < numSprings; ++s)
            {
                simSPRING sp = spring[s];
                interactionIndex[(sp.ndx1+1)*numParticles + sp.m1] = sp.m2;
                interactionIndex[(sp.ndx2+1)*numParticles + sp.m2] = sp.m1;
                // do the velocity Jacobians
                jvX[(sp.ndx1+1)*numParticles + sp.m1].s[0] = sp.springDrag;
                jvY[(sp.ndx1+1)*numParticles + sp.m1].s[1] = sp.springDrag;
                jvZ[(sp.ndx1+1)*numParticles + sp.m1].s[2] = sp.springDrag;
                jvX[(sp.ndx2+1)*numParticles + sp.m2].s[0] = sp.springDrag;
                jvY[(sp.ndx2+1)*numParticles + sp.m2].s[1] = sp.springDrag;
                jvZ[(sp.ndx2+1)*numParticles + sp.m2].s[2] = sp.springDrag;
                jvX[sp.m1].s[0] -= sp.springDrag;
                jvY[sp.m1].s[1] -= sp.springDrag;
                jvZ[sp.m1].s[2] -= sp.springDrag;
                jvX[sp.m2].s[0] -= sp.springDrag;
                jvY[sp.m2].s[1] -= sp.springDrag;
                jvZ[sp.m2].s[2] -= sp.springDrag;
            }

            checkErr( clEnqueueWriteBuffer( queue, clJacobian.interactionParticleIndex, CL_TRUE,
                0, numParticles * jacobianMaxInteractions * sizeof( cl_ulong ), interactionIndex,
                0, NULL, NULL ),
                "clEnqueueWriteBuffer (interactionIndex)" );
            checkErr( clEnqueueWriteBuffer( queue, clJacobian.j_VelX, CL_TRUE,
                0, numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), jvX,
                0, NULL, NULL ),
                "clEnqueueWriteBuffer (Jacobian vX)" );
            checkErr( clEnqueueWriteBuffer( queue, clJacobian.j_VelY, CL_TRUE,
                0, numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), jvY,
                0, NULL, NULL ),
                "clEnqueueWriteBuffer (Jacobian vY)" );
            checkErr( clEnqueueWriteBuffer( queue, clJacobian.j_VelZ, CL_TRUE,
                0, numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), jvZ,
                0, NULL, NULL ),
                "clEnqueueWriteBuffer (Jacobian vZ)" );
            delete[] interactionIndex;
            delete[] jvX;
            delete[] jvY;
            delete[] jvZ;
            
            

            // set up kernels for the Jacobian calculations
            // load the spring/Jacobian calculation kernel from program
            clKernelSpringJacobian = clCreateKernel( program, "calculateSpringForceAndJacobian", &error );
            checkErr( error, "clCreateKernel (Jacobian)" );

            // pass arguments into kernel
            error = clSetKernelArg( clKernelSpringJacobian, 0, sizeof(cl_mem), &clSpring );
            error |= clSetKernelArg( clKernelSpringJacobian, 1, sizeof(cl_mem), &sim_clPosition );
            error |= clSetKernelArg( clKernelSpringJacobian, 2, sizeof(cl_mem), &sim_clVelocity );
            error |= clSetKernelArg( clKernelSpringJacobian, 3, sizeof(cl_mem), &clForce );
            error |= clSetKernelArg( clKernelSpringJacobian, 4, sizeof(cl_mem), &clJacobian.j_PosX);
            error |= clSetKernelArg( clKernelSpringJacobian, 5, sizeof(cl_mem), &clJacobian.j_PosY);
            error |= clSetKernelArg( clKernelSpringJacobian, 6, sizeof(cl_mem), &clJacobian.j_PosZ);
            error |= clSetKernelArg( clKernelSpringJacobian, 7, sizeof(cl_ulong), &numParticles );
            error |= clSetKernelArg( clKernelSpringJacobian, 8, sizeof(cl_ulong), &numSprings );
            checkErr( error, "clSetKernelArg (Jacobian)" );

            // get optimal global item size
            clKernelSpringJacobianGlobalSize = getKernelPaddedGlobalSize( 
                sim_deviceId, clKernelSpringJacobian, numSprings );

            // load the kernel that fills the diagonal
            clKernelGetDiagonal = clCreateKernel( program, "calculateSpringDiagonal", &error );
            checkErr( error, "clCreateKernel (Diagonal)" );

            // pass arguments into kernel
            error |= clSetKernelArg( clKernelGetDiagonal, 0, sizeof(cl_mem), &clJacobian.j_PosX);
            error |= clSetKernelArg( clKernelGetDiagonal, 1, sizeof(cl_mem), &clJacobian.j_PosY);
            error |= clSetKernelArg( clKernelGetDiagonal, 2, sizeof(cl_mem), &clJacobian.j_PosZ);
            error |= clSetKernelArg( clKernelGetDiagonal, 3, sizeof(cl_ulong), &numParticles );
            error |= clSetKernelArg( clKernelGetDiagonal, 4, sizeof(cl_ulong), &jacobianMaxInteractions );
            checkErr( error, "clSetKernelArg (Diagonal)" );

            // get optimal global item size
            clKernelGetDiagonalGlobalSize = getKernelPaddedGlobalSize( 
                sim_deviceId, clKernelGetDiagonal, numParticles );

        }
    }

    // free temporary memory we allocated
    delete[] sCount;
    delete[] spring;
}

// destructor
forceSpringsCL::~forceSpringsCL( )
{
    if (numSprings > 0)
    {
        clReleaseKernel( clKernelSpring );
        clReleaseMemObject( clSpring );
    }
}

// calculating the force
void forceSpringsCL::calculateForce( )
{
    if (numSprings > 0)
    {
        checkErr(
            clEnqueueNDRangeKernel( // run the kernel
                queue, // we are adding it to our command queue
                clKernelSpring, // the kernel we are running (integrate)
                1, // we have a 1 dimensional work dimension
                NULL, // we aren't using any offsets for the work dimensions
                &clKernelSpringGlobalSize, // our global work space is the number of springs
                NULL, // local items are chosen by implementation
                0, NULL, NULL // we aren't waiting for any events
                ),
            "clEnqueueNDRangeKernel");

        // the host and device arrays are no longer in synchronization
        updated = false;
    }
}

// calculate the force and Jacobian
void forceSpringsCL::calculateForceAndJacobian( )
{   
    if (!clJacobianLoaded)
    {
        // then we should just calculate the force
        calculateForce( );
    }
    else
    {
        // we could check for the number of springs, but we only allocate the Jacobian
        // if we have numSprings > 0 anyway
        
        checkErr(
            clEnqueueNDRangeKernel( // run the kernel
                queue, // we are adding it to our command queue
                clKernelSpringJacobian, // the kernel we are running (integrate)
                1, // we have a 1 dimensional work dimension
                NULL, // we aren't using any offsets for the work dimensions
                &clKernelSpringJacobianGlobalSize, // our global work space is the number of springs
                NULL, // local items are chosen by implementation
                0, NULL, NULL // we aren't waiting for any events
                ),
            "clEnqueueNDRangeKernel (Jacobian)");
        // some kind of synchronization?
        if ( jacobianMaxInteractions > 1)
        {
            checkErr(
                clEnqueueNDRangeKernel( // run the kernel
                    queue, // we are adding it to our command queue
                    clKernelGetDiagonal, // the kernel we are running (integrate)
                    1, // we have a 1 dimensional work dimension
                    NULL, // we aren't using any offsets for the work dimensions
                    &clKernelGetDiagonalGlobalSize, // our global work space is the number of springs
                    NULL, // local items are chosen by implementation
                    0, NULL, NULL // we aren't waiting for any events
                    ),
                "clEnqueueNDRangeKernel (GetDiagonal)");
        }

        // the host and device arrays are no longer in synchronization
        updated = false;
    }
}
