// genericSolver.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "genericSolver.h"
#include "clErrorToString.h"
#include <math.h>

// name of this solver
const char * genericSolver::getName( )
{
    return "Generic Solver";
}

// get number of particles
cl_ulong genericSolver::getNumParticles( )
{
    return numParticles;
}

cl_mem genericSolver::getCLPosition ( )
{
    return clPosition;
}

cl_mem genericSolver::getCLVelocity ( )
{
    return clVelocity;
}

// destructor
genericSolver::~genericSolver( )
{
    // delete the allocated arrays
    delete[] position;
    delete[] velocity;

    // release the memory on device
    clReleaseMemObject( clPosition );
    clReleaseMemObject( clVelocity );
    if ( programLoaded )
    {
        // release the program that we allocated
        clReleaseProgram( program );
    }
}

// constructor
genericSolver::genericSolver( 
    std::string sim_CLdir, // the path of directory with our openCL source code
    cl_context sim_context, // OpenCL context passed in from simulation
    cl_command_queue sim_queue, // OpenCL queue passed in from simulation
    CS_REAL sim_timeStep, // the time step of the simulation
    CS_REAL3 * sim_position, // the initial configuration of positions
    cl_ulong sim_numParticles // the number of particles in the simulation
    )
{
    // Not yet broken...
    broken = false;

    // we haven't loaded any programs yet...
    programLoaded = false;
    // we haven't attached the forces yet...
    forcesAttached = false;
    // copy the context and the queue
    context = sim_context;
    queue = sim_queue;
    // copy the directory for our CL source code
    CLdir = sim_CLdir;
    // set the time step
    timeStep = sim_timeStep;

    // set the count of particles
    numParticles = sim_numParticles;

    // initialize an array for storing the positions and velocities
    position = new CS_REAL3[numParticles];
    velocity = new CS_REAL3[numParticles];
    // loop over the particles and update position and velocity
    for (cl_ulong i = 0; i < numParticles; ++i) {
        // Note that we should have that i equals m_it->sofaID. If not, bad luck on us...
        // populate the position
        position[i].s[0] = sim_position[i].s[0];
        position[i].s[1] = sim_position[i].s[1];
        position[i].s[2] = sim_position[i].s[2];
        // populate velocity with zeroes
        velocity[i].s[0] = 0;
        velocity[i].s[1] = 0;
        velocity[i].s[2] = 0;
    }

    // update position and velocity onto the device
    positionUpdated = false;
    velocityUpdated = false;

    cl_int error;
    // create buffer for position
    clPosition = clCreateBuffer( 
        context, // CL context
        CL_MEM_READ_WRITE, // we will read and write from kernels
        numParticles * sizeof( CS_REAL3 ), // how many particles/space to allocate
        NULL, // we aren't copying from existing device buffer
        &error );
    checkErr( error, "clCreateBuffer (clPosition)" );
    // create buffer for velocity
    clVelocity = clCreateBuffer(
        context, CL_MEM_READ_WRITE, numParticles * sizeof( CS_REAL3 ),
        NULL, &error );
    checkErr( error, "clCreateBuffer (clVelocity)" );

    // now that we've allocated the buffers on the device, update...
    updateDevicePosition( );
    updateDeviceVelocity( );
}

// integrate a step forward in time, update position and velocity
void genericSolver::integrateStep( )
{
    // we are not updating our position or our velocity, but to keep consistent
    // with other solvers, which use theForces to either solve for the forces
    // or the forces and Jacobian, automatically picking the appropriate one
    // for the solution method, we need to calculate the forces.
    if ( !forcesAttached ) {
        checkErr( -100, "Attempting to integrate when forces unattached." );
    }
    theForces->calculateForce( );

    // but we still don't do anything with it...
}

void genericSolver::attachForces ( 
    sumForces * totalForce
    )
{
    // set theForces
    theForces = totalForce;

    // we have now attached our forces
    forcesAttached = true;
}

void genericSolver::updateDevicePosition( )
{
    cl_int error = 0;
    // set to clPosition the values of position
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clPosition, // write to clPosition
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * sizeof( CS_REAL3 ), // size of the memory we are sending
            position, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (clPosition)"
        );

    // since the device and host values for the position should be the same...
    positionUpdated = true;
}

void genericSolver::updateDeviceVelocity( )
{
    cl_int error = 0;
    // set to clVelocity the values of velocity
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clVelocity, // write to clVelocity
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * sizeof( CS_REAL3 ), // size of the memory we are sending
            velocity, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (clVelocity)"
        );

    // since the device and host values for the velocity should be the same...
    velocityUpdated = true;
}

void genericSolver::updateHostPosition( )
{
    checkErr(
        clEnqueueReadBuffer( queue, clPosition, CL_TRUE, 0,
            numParticles * sizeof( CS_REAL3 ), position,
            0, NULL, NULL ),
        getName( ) );
    // they've been updated
    positionUpdated = true;
}

void genericSolver::updateHostVelocity( )
{
    checkErr(
        clEnqueueReadBuffer( queue, clVelocity, CL_TRUE, 0,
            numParticles * sizeof( CS_REAL3 ), velocity,
            0, NULL, NULL ),
        getName( ) );
    // they've been updated
    velocityUpdated = true;
}

std::string genericSolver::helperPrintCoordinates( CS_REAL3 * coords )
{
    // assume that coords is updated
    std::ostringstream ret_stream;
    for (cl_ulong i = 0; i < numParticles; ++i) {
        ret_stream << coords[i].s[0] << " "
            << coords[i].s[1] << " "
            << coords[i].s[2] << "\n";
    }

    return ret_stream.str();
}

std::string genericSolver::printPosition( double time )
{
    if (!positionUpdated) {
        updateHostPosition( );
    }
    std::ostringstream ret_stream;
    ret_stream << "Time " << time << "\n" 
        << helperPrintCoordinates( position );

    return ret_stream.str();
}

std::string genericSolver::printVelocity( )
{
    if (!velocityUpdated) {
        updateHostVelocity( );
    }
    std::ostringstream ret_stream;
    ret_stream << "Velocity\n"
        << helperPrintCoordinates( velocity );

    return ret_stream.str();
}

void genericSolver::readPosition( std::vector<CS_REAL3> &pos )
{
    if (!positionUpdated) {
        updateHostPosition( );
    }
    pos.clear();
    for (size_t i = 0; i < numParticles; i++) {
        pos.push_back( position[i] );
    }
}

void genericSolver::readVelocity( std::vector<CS_REAL3> &vel )
{
    if (!positionUpdated) {
        updateHostPosition( );
    }
    vel.clear();
    for (size_t i = 0; i < numParticles; i++) {
        vel.push_back( velocity[i] );
    }
}

bool genericSolver::anyNAN( )
{
    if (!positionUpdated)
    {
        updateHostPosition( );
    }
    if (!velocityUpdated)
    {
        updateHostVelocity( );
    }
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        if (position[i].s[0] != position[i].s[0] 
            || position[i].s[1] != position[i].s[1] 
            || position[i].s[2] != position[i].s[2]
            || velocity[i].s[0] != velocity[i].s[0] 
            || velocity[i].s[1] != velocity[i].s[1]
            || velocity[i].s[2] != velocity[i].s[2])
        {
            return true;
        }
    }

    return false;
}

CS_REAL genericSolver::maxVelocity ( )
{
    if (!velocityUpdated)
    {
        updateHostVelocity( );
    }

    CS_REAL retval = 0;
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        CS_REAL vi = 0;
        for (int j = 0; j < 3; ++j)
            vi += velocity[i].s[j] * velocity[i].s[j];
        if (vi > retval)
            retval = vi;
    }

    return sqrt( retval );
}

void genericSolver::checkErr ( cl_int err, const char * name )
{
    if (err != CL_SUCCESS)
    {
        std::cerr << "ERROR: " << name << " in " << getName( )
            << " (" << clErrorToString(err) << ")" << std::endl;
        exit( EXIT_FAILURE );
    }
}

bool genericSolver::loadProgram ( cl_device_id deviceId, const char * file )
{
    // construct path to source file using CLdir and file
    std::ostringstream path;
    path << CLdir << "/" << file;
    // read in source file
    std::ifstream sourceFile( path.str() );
    if ( !sourceFile.is_open() ) {
        std::cerr << "ERROR: Could not open " << path.str() << " for kernel functions." << std::endl;
	return false;
    }
    std::string sourceCode(
        std::istreambuf_iterator<char>(sourceFile),
        (std::istreambuf_iterator<char>()));
    const char * src = sourceCode.c_str();
    size_t length = sourceCode.length();
    // create the program and build it
    cl_int error;
    // create the program
    program = clCreateProgramWithSource( context, 1, // XXX: not described in documentation, but example used magic number 1
        &src, &length, &error );
    checkErr( error, "clCreateProgramWithSource" );
    
    // set up the include directory to ensure consistent typedefs in kernel code (CS_REAL, CS_REAL3)
    std::ostringstream options;
    options << "-I " << CLdir;

    error = clBuildProgram( program, 1, // the number of devices
        &deviceId, options.str().c_str(), 
        NULL, NULL // stuff for notifications?
        );
    if (error != CL_SUCCESS) {
        // First call to know the proper size
        size_t log_size;
        clGetProgramBuildInfo(program, deviceId, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
        char *build_log = (char* )malloc((log_size+1));

        // Second call to get the log
        clGetProgramBuildInfo(program, deviceId, CL_PROGRAM_BUILD_LOG, log_size, build_log, NULL);
        build_log[log_size] = '\0';
        std::cerr << "--- Build log ---\n ";
        std::cerr << build_log << "\n";
        free(build_log);
    }
    checkErr( error, "clBuildProgram" );
    // we have loaded the program, so it will eventually need to be unloaded
    programLoaded = true;
    return true;
}

// returns a padded global size using kernel preferred local work group size multiple
size_t genericSolver::getKernelPaddedGlobalSize( 
    cl_device_id sim_deviceId, // the device id
    cl_kernel userKernel, // the kernel we are doing this for
    size_t minimumSize // the minimum value for the global size
    )
{
    // get preferred work group size multiple
    size_t preferredSize = getKernelMinimumPreferredLocalSize( sim_deviceId, userKernel );

    // now, set the global size to be the smallest size_t larger than minimumSize
    // and a multiple of preferredSize
    return (preferredSize - ( minimumSize % preferredSize )) + minimumSize;
}

size_t genericSolver::getKernelMinimumPreferredLocalSize(
    cl_device_id sim_deviceId, // the device id
    cl_kernel userKernel // the kernel we are doing this for
    )
{
    // get preferred work group size multiple
    size_t preferredSize = 0;
    checkErr( clGetKernelWorkGroupInfo(
            userKernel, sim_deviceId, 
            CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE,
            sizeof( preferredSize ), &preferredSize,
            NULL ),
        "clGetKernelWorkGroupInfo" );

    return preferredSize;
}
