// genericSolver.h
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif
#include "sumForces.h"
#include "structureLoader.h"
#include "chromoTypes.h"
#include <iostream>
#include <string>
#include <sstream>
#include <fstream>
#include <vector>

#ifndef GENERICSOLVER_H
#define GENERICSOLVER_H

// base class for evolving the positions (and perhaps velocities) of the particles
// damping is handled in this class

class genericSolver
{
protected :
    // Becomes true if we are broken.
    bool broken;

    // information about the simulation
    CS_REAL timeStep;

    // information about the particles
    // Position
    cl_mem clPosition; // the position in device memory
    CS_REAL3 * position; // the position in host memory
    bool positionUpdated; // has it been updated since last integration step?
    // Velocity
    cl_mem clVelocity; // the velocities of our masses (will be zero in case of brownian dynamics)
    CS_REAL3 * velocity; // the velocity in host memory
    bool velocityUpdated; // has it been updated since last integration step?
    // number of particles
    cl_ulong numParticles;

    // information about the forces
    sumForces * theForces; // the forces that will be updated at each integrate step
    bool forcesAttached; // keep track of whether the forces have been attached

    // OpenCL details
    cl_context context; // get from constructor
    cl_command_queue queue; // get from constructor
    cl_program program; // a program that may or may not be created yet
    bool programLoaded; // has the program been loaded yet? (default false)

    std::string CLdir; // the path of the directory with our OpenCL source code

    // retrieve the position from the device and put to host
    void updateHostPosition ( );

    // take position from host and put on device
    void updateDevicePosition ( );

    // retrieve the velocity from the device and put to host
    void updateHostVelocity ( );

    // take velocity from host and put on device
    void updateDeviceVelocity ( );

    // helper for print position and velocity
    std::string helperPrintCoordinates ( CS_REAL3 * coords );

    // check for error in calling some OpenCL function
    void checkErr ( cl_int err, const char * name );

    // load a given .cl file to program and build
    bool loadProgram ( cl_device_id deviceId, const char * file );

    // returns a padded global size using kernel preferred local work group size multiple
    size_t getKernelPaddedGlobalSize( 
        cl_device_id sim_deviceId, // the device id
        cl_kernel userKernel, // the kernel we are doing this for
        size_t minimumSize // the minimum value for the global size
        );

    size_t getKernelMinimumPreferredLocalSize(
        cl_device_id sim_deviceId, // the device id
        cl_kernel userKernel // the kernel we are doing this for
        );

public :
    // constructor
    genericSolver (
        std::string sim_CLdir, // the path of directory with our openCL source code
        cl_context sim_context, // OpenCL context passed in from simulation
        cl_command_queue sim_queue, // OpenCL queue passed in from simulation
        CS_REAL sim_timeStep, // the time step of the simulation
        CS_REAL3 * sim_position, // the initial configuration of positions
        cl_ulong sim_numParticles // the number of particles in the simulation
        );

    // destructor
    virtual ~genericSolver ( );

    // Are we doing okay?
    bool doing_okay(void) const { return !broken; }

    // name of the solver
    virtual const char * getName ( );

    // integrate one step
    virtual void integrateStep ( );

    // attach forces to the solver
    virtual void attachForces ( 
        sumForces * totalForce
        );

    // print position or velocity
    std::string printPosition ( double time );
    std::string printVelocity ( );

    // Read position or velocity into a vector
    void readPosition( std::vector<CS_REAL3> &pos );
    void readVelocity( std::vector<CS_REAL3> &vel );

    // getter for the number of particles
    cl_ulong getNumParticles ( );
    
    // getter for the current position on the device
    cl_mem getCLPosition ( );

    // getter for the current velocity on the device
    cl_mem getCLVelocity ( );

    // checks if any of the positions or velocities are NAN
    bool anyNAN ( );
    // the maximum magnitude of the velocities
    CS_REAL maxVelocity ( );
};







#endif
