#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <fstream>
#include "unitConversion.h"

void usage(const char *name)
{
	fprintf(stderr,"Usage: %s [-option value] [outfile]\n", name);
	fprintf(stderr,"     outfile: Name of the output file to write to (default stdout)\n");
	fprintf(stderr, "     -bead_directory: Name of directory for files with important bead numbers\n");
	fprintf(stderr, "     -no_bead_output: Don't output files with important bead numbers\n");
	fprintf(stderr,"     -chain_length (microns) (default 1)\n");
	fprintf(stderr,"     -node_separation (microns) (default 0.01)\n");
	fprintf(stderr, "     -brushes (default 0)\n");
	fprintf(stderr, "     -brush_length (microns) (default 0.2)\n");
	fprintf(stderr,"     -temperature (Celsius) (default 25)\n");
	fprintf(stderr,"     -viscosity (centiPoise) (default 1.5e-5)\n");
        fprintf(stderr,"         (Water = 1, but we use this to scale time)\n");
        fprintf(stderr,"         (Smaller runs faster but can make the simulation unstable)\n");
	fprintf(stderr,"     -damping_radius_factor: Multiplied by node_separation to determine effective radius for damping (ratio) (default 0.8)\n");
	fprintf(stderr,"     -time_step (nanoseconds) (default 2)\n");
        fprintf(stderr,"         (Longer time steps run faster but can make the simulation unstable\n");
	fprintf(stderr,"     -modulus: Young's modulus of DNA (gigaPascal) (default 2)\n");
	fprintf(stderr,"     -dna_radius: Radius of B-form DNA (nanometers) (default 0.6)\n");
        fprintf(stderr,"         (The collision/visual radius will be larger than this)\n");
	fprintf(stderr,"     -collision_radius_factor: Fraction of distance collision spheres will cover (default 0.9)\n");
        fprintf(stderr,"     -collision_force_factor: Number of times the spring constant to use for collision response (default 0.25)\n");
        fprintf(stderr,"     -spring_damping_factor: Ratio of drag constant from relative velocity of springs to spring constants (default 0)\n");
        fprintf(stderr, "    -seed: Specify the seed value for the random number generator\n" );
        fprintf(stderr, "        (default is 42)\n" );
	exit(-1);
}

int main (int argc, char *argv[])
{
	// Constants (could become command-line arguments).
    const char *outfile_name = NULL;
	double	chain_length_microns = 1;
	double	mass_separation_microns = 0.01;
	double  brush_length_microns = 0.2;
	int brushes = 0;
    double  temperature_Celsius = 25;
	double  viscosity_centiPoise = 1;
    double  time_step_nanoseconds = 2;
    double  modulus_gigaPascal = 2;
	double  dna_radius_nanometers = 0.6;
    double  collision_radius_factor = 0.9;
    double  collision_force_factor = 0.25;
    double  spring_damping_factor = 0.0;
    double  damping_radius_factor = 0.8;
    int random_number_seed = 42;
	bool beadOutput = true;
	std::string beadFileName;

	// Parse the command line to adjust default values.
	int i;
        unsigned  real_params = 0;  //< Number of non-flag parameters
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-bead_directory") == 0) {
			if (++i > argc) { usage(argv[0]); }
			beadFileName = argv[i];
			beadFileName += "/chainbeads.csv";
		}
		else if (strcmp(argv[i], "-no_bead_output") == 0) {
			beadOutput = false;
		} else if (!strcmp(argv[i], "-chain_length")) {
			if (++i >= argc) { usage(argv[0]); }
			chain_length_microns = atof(argv[i]);
		} else if (!strcmp(argv[i], "-node_separation")) {
			if (++i >= argc) { usage(argv[0]); }
			mass_separation_microns = atof(argv[i]);
		} else if (!strcmp(argv[i], "-brushes")) {
			if (++i >= argc) { usage(argv[0]); }
			brushes = atoi(argv[i]);
		} else if (!strcmp(argv[i], "-brush_length")) {
			if (++i >= argc) { usage(argv[0]); }
			brush_length_microns = atof(argv[i]);
		} else if (!strcmp(argv[i], "-temperature")) {
			if (++i >= argc) { usage(argv[0]); }
			temperature_Celsius = atof(argv[i]);
		} else if (!strcmp(argv[i], "-viscosity")) {
			if (++i >= argc) { usage(argv[0]); }
			viscosity_centiPoise = atof(argv[i]);
		} else if (!strcmp(argv[i], "-damping_radius_factor")) {
			if (++i >= argc) { usage(argv[0]); }
			damping_radius_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-time_step")) {
			if (++i >= argc) { usage(argv[0]); }
			time_step_nanoseconds = atof(argv[i]);
		} else if (!strcmp(argv[i], "-modulus")) {
			if (++i >= argc) { usage(argv[0]); }
			modulus_gigaPascal = atof(argv[i]);
		} else if (!strcmp(argv[i], "-dna_radius")) {
			if (++i >= argc) { usage(argv[0]); }
			dna_radius_nanometers = atof(argv[i]);
		} else if (!strcmp(argv[i], "-collision_radius_factor")) {
			if (++i >= argc) { usage(argv[0]); }
			collision_radius_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-collision_force_factor")) {
			if (++i >= argc) { usage(argv[0]); }
			collision_force_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-spring_damping_factor")) {
			if (++i >= argc) { usage(argv[0]); }
			spring_damping_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-random_number_seed")) {
			if (++i >= argc) { usage(argv[0]); }
			random_number_seed = atoi(argv[i]);
		} else {
                    switch (real_params++) {
                    case 0:
                        outfile_name = argv[i];
                        break;
                    default:
                        usage(argv[0]);
                    }
		}
	}

        // Compute the derived values that we feed to the simulator.
        double  time_step_seconds = nanoseconds_to_seconds(time_step_nanoseconds);
        double  chain_length_meters = microns_to_meters(chain_length_microns);
        double  node_separation_meters = microns_to_meters(mass_separation_microns);
		double  brush_length_meters = microns_to_meters(brush_length_microns);
	double	mass_kg = dna_mass_in_kg(node_separation_meters);
        double  temperature_kelvin = Celsius_to_Kelvin(temperature_Celsius);
        double  viscosity_Pascal_seconds = centiPoise_to_Pascal_seconds(viscosity_centiPoise);
        double  dna_radius_meters = nanometers_to_meters(dna_radius_nanometers);
	double	effective_damping_radius = damping_radius_factor * node_separation_meters;

	double	spring_constant = dna_tensile_spring_constant(
                        node_separation_meters,
                        gigaPascal_to_Pascal(modulus_gigaPascal),
                        dna_radius_meters
                        );
	double	hinge_constant = dna_bending_spring_constant(
                        node_separation_meters,
                        gigaPascal_to_Pascal(modulus_gigaPascal),
                        dna_radius_meters
                        );
	double	random_force = Brownian_force_equivalent(
                        node_separation_meters,
                        temperature_kelvin,
                        viscosity_Pascal_seconds,
                        time_step_seconds
                        );
	double	mass_damping = mass_damping_equivalent(
			viscosity_Pascal_seconds,
			effective_damping_radius,
			mass_kg);
	double	mass_radius = collision_radius_factor * (node_separation_meters / 2.0);
        double  collision_spring_constant = spring_constant * collision_force_factor;

        // Figure out where to put the output file.  Open it if needed.
	FILE *out = stdout;
        if (outfile_name != NULL) {
            out = fopen(outfile_name, "wb");
            if (out == NULL) {
                perror("Could not open output file");
                return -1;
            }
        }

	// Print out the header information and general parameters
	// for the simulation.  This includes medatadata lines and then
        // the beginning of the structure with its parameters
        fprintf(out, "meta temperature_Celsius %lg\n", temperature_Celsius);
        fprintf(out, "meta viscosity_centiPoise %lg\n", viscosity_centiPoise);
        fprintf(out, "meta effective_damping_radius %lg\n", effective_damping_radius);
        fprintf(out, "meta dna_modulus_gigaPascal %lg\n", modulus_gigaPascal);
        fprintf(out, "meta dna_radius_nanometers %lg\n", dna_radius_nanometers);
        fprintf(out, "meta damping_radius_factor %lg\n", damping_radius_factor);
	fprintf(out, "structure {\n");
	fprintf(out, "  random_force %lg\n", random_force);
	fprintf(out, "  mass_damping %lg\n", mass_damping);
	fprintf(out, "  mass_radius %lg\n", mass_radius);
	fprintf(out, "  time_step %lg\n", time_step_seconds);
	fprintf(out, "  collision_spring_constant %lg\n", collision_spring_constant);
	fprintf(out, "  spring_damping_factor %lg\n", spring_damping_factor);
	fprintf(out, "  random_number_seed %d\n", random_number_seed);
	fprintf(out, "  color %d\n", 1);

	// Put the masses into the system.  Attach a spring between each
	// pair of masses.  Attach a hinge between each triplet of
	// masses.  Stop when we get more than a fraction of the mass-
	// separation beyond the requested length.
	i = 0;
	double x;
	double epsilon = node_separation_meters / 10.0;
	for (x = 0.0; x < chain_length_meters+epsilon; x += node_separation_meters, i++) {

		fprintf(out, "  mass %u\t%g\t%g 0 0\n", i, mass_kg, x);
		if (i >= 1) {
			fprintf(out, "  spring %u %u %lg %lg\n", i-1, i,
				node_separation_meters, spring_constant);
		}
		if (i >= 2) {
			fprintf(out, "  hinge %u %u %u %lg\n", i-2, i-1, i,
				hinge_constant);
		}
	}
	int last_center_chain_bead = i;
	
	if (brushes > 0) {
		const int numbeads = i;

		// Add evenly spaced brushes attached to beads along the chain
		if (brushes >= numbeads) {
			brushes = numbeads;
		}
		int brushes_added = 0;
		double y, z;
		int angle = 0, color = 2;

		double bead_step = (double)numbeads / (double)(brushes + 1);
		if (bead_step < 1) { bead_step = 1; }
		angle = 0;
		for (double bead = bead_step; bead < numbeads && brushes_added < brushes; bead += bead_step, i++) {
			int cur_bead = floor(bead);
			x = cur_bead * node_separation_meters;
			y = 0;
			z = 0;
			if (angle < 3) {
				angle++;
				color++;
			}
			else {
				angle = 0;
				color = 2;
			}
			int beads_per_brush = floor(brush_length_meters / node_separation_meters);
			for (int brush_bead = 0; brush_bead < beads_per_brush; brush_bead++, i++) {
				if (angle == 0) { y += node_separation_meters; }
				else if (angle == 1) { z += node_separation_meters; }
				else if (angle == 2) { y -= node_separation_meters; }
				else if (angle == 3) { z -= node_separation_meters; }
				fprintf(out, "  mass %u\t%g\t%g %g %g %u\n", i, mass_kg, x, y, z, color);
				if (brush_bead == 0) {
					fprintf(out, "  spring %u %u %lg %lg\n", cur_bead, i,
						node_separation_meters, spring_constant);
				}
				if (brush_bead >= 1) {
					fprintf(out, "  spring %u %u %lg %lg\n", i - 1, i,
						node_separation_meters, spring_constant);
				}
				if (brush_bead == 1) {
					fprintf(out, "  hinge %u %u %u %lg\n", cur_bead, i - 1, i,
						hinge_constant);
				}
				if (brush_bead > 2) {
					fprintf(out, "  hinge %u %u %u %lg\n", i - 2, i - 1, i,
						hinge_constant);
				}
			}
			brushes_added++;
		}
	}

	if (beadOutput) {
		std::ofstream beadFile;
		beadFile.open(beadFileName, std::ofstream::out);
		for (int j = 0; j < last_center_chain_bead; j++) {
			if (j == 0) {
				beadFile << (i - j - 1);
			}
			else {
				beadFile << "," << (i - j - 1);
			}
		}
	}

	// Close off the structure.
	fprintf(out, "}\n");

	fclose(out);
	return 0;
}

