#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <xtgmath.h>
#include <vector>
#include "unitConversion.h"

// Constants (could become command-line arguments).
double	mass_separation_microns = 0.01;
double  temperature_Celsius = 25;
double  viscosity_centiPoise = 1;
double  time_step_nanoseconds = 2;
double  modulus_gigaPascal = 2;
double  dna_radius_nanometers = 0.6;
double  collision_radius_factor = 0.9;
double  collision_force_factor = 0.25;
double  spring_damping_factor = 0.0;
double  damping_radius_factor = 0.8;
int random_number_seed = 42;

// Computed quantities needed by the structure-construction code
double  time_step_seconds;
double  node_separation_meters;
double	mass_kg;
double  temperature_kelvin;
double  viscosity_Pascal_seconds;
double  dna_radius_meters;
double	effective_damping_radius;
double	spring_constant;
double	hinge_constant;
double	random_force;
double	mass_damping;
double	mass_radius;
double  collision_spring_constant;


void Usage(const char *name)
{
	fprintf(stderr, "Usage: %s [-cohesin_offset_microns O] [-cohesin_x_offset_microns O] [-loop_expansion f] "
		"[-basis_order k][-missing_condensin x1, x2, x3..] "
		"[-nochromatin] [-nocohesin] [-nocondensin] "
		"[-bead_directory d] [-option value] [outfile]\n", name);
	fprintf(stderr, "     outfile: Name of the output file to write to (default stdout)\n");
	fprintf(stderr, "     -bead_directory: Name of directory for files with important bead numbers\n");
	fprintf(stderr, "     -no_bead_output: Don't output files with important bead numbers\n");
	fprintf(stderr, "     -cohesin_offet_microns: Specify offset (default 0.22)\n");
	fprintf(stderr, "     -cohesin_x_offet_microns: Specify offset (default 0, 0.0255 is radius of cohesin)\n");
	fprintf(stderr, "	 -loop_expansion: Specify factor f by which to expand loops radially from spindle center (default 1)\n");
	fprintf(stderr, "     -basis_order: Specify the spline basis order (default 4)\n");
	fprintf(stderr, "     -missing_condensin: Specify the condensin number, separate by comma (range 1-128)\n");
	fprintf(stderr, "     -nochromatin: Turn off DNA loops\n");
	fprintf(stderr, "     -nocohesin: Turn off cohesin loops\n");
	fprintf(stderr, "     -nocondensin: Turn off condensin staples\n");
	fprintf(stderr, "   Other options:\n");
	fprintf(stderr, "     -node_separation (microns) (default 0.01\n");
	fprintf(stderr, "     -temperature (Celsius) (default 25)\n");
	fprintf(stderr, "     -viscosity (centiPoise) (default 1)\n");
	fprintf(stderr, "         (Smaller runs faster but can make the simulation unstable)\n");
	fprintf(stderr, "     -damping_radius_factor: Multiplied by node_separation to determine effective radius for damping (ratio) (default 0.8)\n");
	fprintf(stderr, "     -time_step (nanoseconds) (default 2)\n");
	fprintf(stderr, "         (Longer time steps run faster but can make the simulation unstable\n");
	fprintf(stderr, "     -modulus: Young's modulus of DNA (gigaPascal) (default 2)\n");
	fprintf(stderr, "     -dna_radius: Radius of B-form DNA (nanometers) (default 0.6)\n");
	fprintf(stderr, "         (The collision/visual radius will be larger than this)\n");
	fprintf(stderr, "     -collision_radius_factor: Fraction of distance collision spheres will cover (default 0.9)\n");
	fprintf(stderr, "     -collision_force_factor: Number of times the spring constant to use for collision response (default 0.25)\n");
	fprintf(stderr, "     -spring_damping_factor: Ratio of drag constant from relative velocity of springs to spring constants (default 0)\n");
	fprintf(stderr, "    -seed: Specify the seed value for the random number generator\n");
	fprintf(stderr, "        (default is 42)\n");
	exit(-1);
}

// Number of Condensin included so far.
int g_num_condensin = 0;

/*
Subroutine to generate a B-spline open knot vector with multiplicity
equal to the order at the ends.

c            = order of the basis function
n            = the number of defining polygon vertices
nplus2       = index of x() for the first occurence of the maximum knot vector value
nplusc       = maximum value of the knot vector -- $n + c$
x()          = array containing the knot vector
*/
void knot(int n, int c, int x[])
{
	int nplusc, nplus2, i;

	nplusc = n + c;
	nplus2 = n + 2;

	x[1] = 0;
	for (i = 2; i <= nplusc; i++){
		if ((i > c) && (i < nplus2))
			x[i] = x[i - 1] + 1;
		else
			x[i] = x[i - 1];
	}
}


/*  Subroutine to generate rational B-spline basis functions--open knot vector

C code for An Introduction to NURBS
by David F. Rogers. Copyright (C) 2000 David F. Rogers,
All rights reserved.

Name: rbasis
Language: C
Subroutines called: none
Book reference: Chapter 4, Sec. 4. , p 296

c        = order of the B-spline basis function
d        = first term of the basis function recursion relation
e        = second term of the basis function recursion relation
h[]	     = array containing the homogeneous weights
npts     = number of defining polygon vertices
nplusc   = constant -- npts + c -- maximum number of knot values
r[]      = array containing the rationalbasis functions
r[1] contains the basis function associated with B1 etc.
t        = parameter value
temp[]   = temporary array
x[]      = knot vector
*/
void rbasis(int c, float t, int npts, int x[], float h[], float r[])
{
	int nplusc;
	int i, k;
	float d, e;
	float sum;
	float temp[36];

	nplusc = npts + c;

	/*		printf("knot vector is \n");
	for (i = 1; i <= nplusc; i++){
	printf(" %d %d \n", i,x[i]);
	}
	printf("t is %f \n", t);
	*/

	/* calculate the first order nonrational basis functions n[i]	*/

	for (i = 1; i <= nplusc - 1; i++){
		if ((t >= x[i]) && (t < x[i + 1]))
			temp[i] = 1;
		else
			temp[i] = 0;
	}

	/* calculate the higher order nonrational basis functions */

	for (k = 2; k <= c; k++){
		for (i = 1; i <= nplusc - k; i++){
			if (temp[i] != 0)    /* if the lower order basis function is zero skip the calculation */
				d = ((t - x[i])*temp[i]) / (x[i + k - 1] - x[i]);
			else
				d = 0;

			if (temp[i + 1] != 0)     /* if the lower order basis function is zero skip the calculation */
				e = ((x[i + k] - t)*temp[i + 1]) / (x[i + k] - x[i + 1]);
			else
				e = 0;

			temp[i] = d + e;
		}
	}

	if (t == (float)x[nplusc]){		/*    pick up last point	*/
		temp[npts] = 1;
	}
	/*
	printf("Nonrational basis functions are \n");
	for (i=1; i<= npts; i++){
	printf("%f ", temp[i]);
	}
	printf("\n");
	*/
	/* calculate sum for denominator of rational basis functions */

	sum = 0.;
	for (i = 1; i <= npts; i++){
		sum = sum + temp[i] * h[i];
	}

	/* form rational basis functions and put in r vector */

	for (i = 1; i <= npts; i++){
		if (sum != 0){
			r[i] = (temp[i] * h[i]) / (sum);
		}
		else
			r[i] = 0;
	}
}

//interpolation function: interpolates from current point to next point & saves
//the value in point.
void lerp(float distanceToGo, float currentPoint[], float nextPoint[], float point[])
{
	//v = p1 - p0 = (x1-x0, y1-y0, z1-z0) = (x,y,z);
	float x = nextPoint[0] - currentPoint[0];
	float y = nextPoint[1] - currentPoint[1];
	float z = nextPoint[2] - currentPoint[2];

	float mag = sqrt(x*x + y*y + z*z);

	x = (distanceToGo / mag) * x;
	y = (distanceToGo / mag) * y;
	z = (distanceToGo / mag) * z;

	point[0] = currentPoint[0] + x;
	point[1] = currentPoint[1] + y;
	point[2] = currentPoint[2] + z;
}

//function to check the distance
float distance(float a[], float b[])
{
	float dist = sqrt((b[0] - a[0])*(b[0] - a[0]) + (b[1] - a[1])*(b[1] - a[1]) + (b[2] - a[2])*(b[2] - a[2]));
	return dist;
}

/*  Subroutine to generate a rational B-spline curve using an uniform open knot vector

C code for An Introduction to NURBS
by David F. Rogers. Copyright (C) 2000 David F. Rogers,
All rights reserved.

Name: rbspline.c
Language: C
Subroutines called: knot.c, rbasis.c, fmtmul.c
Book reference: Chapter 4, Alg. p. 297

b[]         = array containing the defining polygon vertices
b[1] contains the x-component of the vertex
b[2] contains the y-component of the vertex
b[3] contains the z-component of the vertex
h[]			= array containing the homogeneous weighting factors
k           = order of the B-spline basis function
nbasis      = array containing the basis functions for a single value of t
nplusc      = number of knot values
npts        = number of defining polygon vertices
p[,]        = array containing the curve points
p[1] contains the x-component of the point
p[2] contains the y-component of the point
p[3] contains the z-component of the point
p1          = number of points to be calculated on the curve
t           = parameter value 0 <= t <= npts - k + 1
x[]         = array containing the knot vector
//returns massCount!
*/
int rbspline(FILE *ofp2, int npts, int k, int p1, float b[], float h[], float p[], float finalP[],
	int massCount, int current_chromosome, int springMasses[], int* missing_condensin,
	int doDNA, int doStaples, int doRings,
	double cohesin_offset_microns, double cohesin_x_offset_microns,
	std::vector<int>& ends, std::vector<int>& tips, std::vector<int>& bases,
	std::vector<int>& halfloops, std::vector<int>& outerloops,
	std::vector<int>& innerloops, std::vector<int>& cohesin_beads,
	std::vector<int>& array6p8, std::vector<int>& colorBeads)
{

	//file stuff below until end of fprintf statements.
	// Constants (could become command-line arguments).
	// All specified in microns.
	unsigned num_lines = 2;

	double	mass = mass_kg;
	double bigMass = mass * 100000;
	double mass_separation = node_separation_meters;

	//	ofp2 = fopen("NURBS.txt","a+"); /* apend file (add text to 
	//	  a file or create a file if it does not exist.*/

	//For debugging and only creating the end part of the model
	// if(current_chromosome > 13) {
	// 	fprintf(ofp2, "start");
	// }
	int colorNumber = 1 + (current_chromosome % 4);

	int i, j, icount, jcount;
	int i1;
	int nplusc;

	float step;
	float t;
	float *nbasis = new float[npts + 1];
	float temp;

	nplusc = npts + k;
	int *x = new int[nplusc + 1];

	/*  zero and redimension the knot vector and the basis array */

	for (i = 0; i <= npts; i++) {
		nbasis[i] = 0.;
	}

	for (i = 0; i <= nplusc; i++) {
		x[i] = i;
	}

	/* generate the uniform open knot vector */

	knot(npts, k, x);

	/*
	printf("The knot vector is ");
	for (i = 1; i <= nplusc; i++){
	printf(" %d ", x[i]);
	}
	printf("\n");
	*/

	icount = 0;

	/*    calculate the points on the rational B-spline curve */

	t = 0;
	step = ((float)x[nplusc]) / ((float)(p1 - 1));

	for (i1 = 1; i1 <= p1; i1++) {

		if ((float)x[nplusc] - t < 5e-12){
			t = (float)x[nplusc];
		}

		rbasis(k, t, npts, x, h, nbasis);      /* generate the basis function for this value of t */
		/*
		printf("t = %f \n",t);
		printf("nbasis = ");
		for (i = 1; i <= npts; i++){
		printf("%f  ",nbasis[i]);
		}
		printf("\n");
		*/
		//fills out p with all of the points along the curve.  Do something with those next
		for (j = 1; j <= 3; j++){      /* generate a point on the curve */
			jcount = j;
			p[icount + j] = 0.;

			for (i = 1; i <= npts; i++){ /* Do local matrix multiplication */
				temp = nbasis[i] * b[jcount];
				p[icount + j] = p[icount + j] + temp;
				/*
				printf("jcount,nbasis,b,nbasis*b,p = %d %f %f %f %f\n",jcount,nbasis[i],b[jcount],temp,p[icount+j]);
				*/
				jcount = jcount + 3;
				//
			}
		}
		/*
		printf("icount, p %d %f %f %f \n",icount,p[icount+1],p[icount+2],p[icount+3]);
		*/
		icount = icount + 3;
		t = t + step;
	}

	//you just got the points -- they're in p[icount+1], p[icount+2], p[icount+3]
	float currentPoint[3];
	float nextPoint[3];
	float ptDistance;
	float maxDistanceFromCenter = 0;

	float springLength = static_cast<float>(mass_separation);

	int pointStep = 0;
	int finalStep = 0;

	while (pointStep < p1) {
		//first point
		currentPoint[0] = p[3 * pointStep + 1];
		currentPoint[1] = p[3 * pointStep + 2];
		currentPoint[2] = p[3 * pointStep + 3];
		pointStep++;

		//go ahead & put this in the final point array at the 1st position
		finalP[3 * finalStep + 0] = currentPoint[0];
		finalP[3 * finalStep + 1] = currentPoint[1];
		finalP[3 * finalStep + 2] = currentPoint[2];

		nextPoint[0] = p[3 * pointStep + 1];
		nextPoint[1] = p[3 * pointStep + 2];
		nextPoint[2] = p[3 * pointStep + 3];
		pointStep++;

		ptDistance = distance(currentPoint, nextPoint);

		while (ptDistance < springLength) {
			currentPoint[0] = nextPoint[0];
			currentPoint[1] = nextPoint[1];
			currentPoint[2] = nextPoint[2];

			//step nextPoint along
			nextPoint[0] = p[3 * pointStep + 1];
			nextPoint[1] = p[3 * pointStep + 2];
			nextPoint[2] = p[3 * pointStep + 3];
			pointStep++;

			//add on to the point distance
			ptDistance = ptDistance + distance(currentPoint, nextPoint);
		}

		//at this point you've gone too far along the line
		pointStep--;

		//you now need to interpolate along current point to previous point at
		//the length of distance to go
		float distanceToGo = springLength - (ptDistance - distance(currentPoint, nextPoint));
		//now interpolate b/n currentPoint & nextPoint
		//point is the actual new point.
		float *point = new float[3 * npts];

		//TODO: fix float* the point, not right (?)
		//look at the way that this is being done by p & finalP
		lerp(distanceToGo, currentPoint, nextPoint, point);
		finalP[0 + 3 * finalStep] = point[0];
		finalP[1 + 3 * finalStep] = point[1];
		finalP[2 + 3 * finalStep] = point[2];
		finalStep++;

		float ptXY[3] = { point[0], point[1], 0 };
		float zero[3] = { 0, 0, 0 };
		float currentDist = distance(ptXY, zero);
		if (currentDist > maxDistanceFromCenter) {
			maxDistanceFromCenter = currentDist;
		}

		//put finalP into the right index of p so that it gets
		//picked up as the current point, but make sure you're not
		//writing over any important data.
		p[3 * pointStep + 1] = point[0];
		p[3 * pointStep + 2] = point[1];
		p[3 * pointStep + 3] = point[2];

		delete[] point;
	}

	// Keep track of the first mass in this structure, so we don't
	// tie it to the last one with a spring and a hinge.
	unsigned first_mass = massCount;
	unsigned most_recent_staple = massCount;	// Keep track of when we add condensin staples
	//steps by three because of weird storage mechanism introduced by original code.
	if (doDNA) {
		bool base1_found = false, base2_found = false;
		float tip1_dist = 0, tip2_dist = 0;
		int tip1_bead = 0, tip2_bead = 0;
		bool onLoop = false; // whether or not the last bead was on the upper half of a loop
		bool onOuterLoop = true;
		bool on6p8 = false;
		bool firstLoopBead = true;

		// calculate start and end of 6p8 arrays, starting (1.8/50) of the way through the 
		// strand and continuing for 1/5 of the strand.
		int startfirst6p8 = floor(first_mass + (finalStep * 1.8 / 50.0));
		int endfirst6p8 = startfirst6p8 + floor(finalStep / 5.0);
		int endsecond6p8 = ceil(first_mass + finalStep - (finalStep * 1.8 / 50.0));
		int startsecond6p8 = endsecond6p8 - floor(finalStep / 5.0);

		for (i = 0; i <= 3 * finalStep - 3; i = i + 3, massCount++) {
			if ((massCount >= startfirst6p8 && massCount <= endfirst6p8) || (massCount >= startsecond6p8 && massCount <= endsecond6p8)) {
				array6p8.push_back(massCount);
				if (!on6p8) { on6p8 = true; }
			}
			else if (on6p8) {
				on6p8 = false;
				array6p8.push_back(-1); // mark end of 6p8 array for this strand (-1 won't get printed).
			}

			// The first and last mass on each strand should be very heavy so that they
			// don't ever move.  This pins them in space.
			if (i == 0 || i == 3 * finalStep - 3) {
				if (std::find(colorBeads.begin(), colorBeads.end(), massCount) != colorBeads.end()) {
					fprintf(ofp2, "mass %u\t%g\t%g %g %g %d\n", massCount, bigMass, finalP[i], finalP[i + 1], finalP[i + 2], 5);
				}
				else {
					fprintf(ofp2, "mass %u\t%g\t%g %g %g %d\n", massCount, bigMass, finalP[i], finalP[i + 1], finalP[i + 2], colorNumber);
				}
				ends.push_back(massCount);
			}
			else {
				float ptXY[3] = { finalP[i], finalP[i + 1], 0 };
				float zero[3] = { 0, 0, 0 };
				float currentDist = distance(ptXY, zero);
				// Check if bead is in the top half of a loop
				if (currentDist > 8.5e-8 + ((maxDistanceFromCenter - 8.5e-8) / 2)) {
					if (!onLoop) {
						onLoop = true;
					}
					halfloops.push_back(massCount);
					if (fabs(finalP[i + 2]) >= 1.65e-7) {
						onOuterLoop = true;
						outerloops.push_back(massCount);
					}
					else {
						onOuterLoop = false;
						innerloops.push_back(massCount);
					}
					if (firstLoopBead) firstLoopBead = false;
				}
				else if (onLoop) {
					onLoop = false;
					firstLoopBead = true;
					// insert negatives into bead vectors to mark the end of a loop (these won't get printed)
					halfloops.push_back(-1);
					if (onOuterLoop) {
						outerloops.push_back(-1);
					}
					else {
						innerloops.push_back(-1);
					}
				}
				// Check if bead is at the tip of an outer loop
				if (fabs(maxDistanceFromCenter - currentDist) < 5.0e-9 && (fabs(finalP[i + 2]) >= 2.3e-7 && fabs(finalP[i + 2]) <= 2.6e-7)) {
					if (finalP[i + 2] > 0) {
						if (currentDist > tip1_dist) {
							tip1_dist = currentDist;
							tip1_bead = massCount;
						}
					}
					else {
						if (currentDist > tip2_dist) {
							tip2_dist = currentDist;
							tip2_bead = massCount;
						}
					}
				}
				// Try to find a bead at the base of the outer loops
				else if (fabs(finalP[i + 2]) >= 2.45e-7 && fabs(finalP[i + 2]) <= 2.55e-7) {
					if (!base1_found && finalP[i + 2] >= 2.45e-7 && finalP[i + 2] <= 2.55e-7) {
						bases.push_back(massCount);
						base1_found = true;
					}
					else if (!base2_found && finalP[i + 2] <= -2.45e-7 && finalP[i + 2] >= -2.55e-7) {
						bases.push_back(massCount);
						base2_found = true;
					}
				}
				if (std::find(colorBeads.begin(), colorBeads.end(), massCount) != colorBeads.end()) {
					fprintf(ofp2, "mass %u\t%g\t%g %g %g %d\n", massCount, mass, finalP[i], finalP[i + 1], finalP[i + 2], 5);
				}
				else {
					fprintf(ofp2, "mass %u\t%g\t%g %g %g %d\n", massCount, mass, finalP[i], finalP[i + 1], finalP[i + 2], colorNumber);
				}
			}

			// Put a spring between this mass and the previous once we have at least two
			// masses.
			if (massCount - first_mass >= 1) {
				fprintf(ofp2, "spring %u %u %lg %lg\n", massCount - 1, massCount, mass_separation, spring_constant);
			}

			// Put a hinge between this mass and the previous two once we have at least three
			// masses.
			if (massCount - first_mass >= 2) {
				fprintf(ofp2, "hinge %u %u %u %lg\n", massCount - 2, massCount - 1, massCount,
					hinge_constant);
			}

			// If we're doing condensin staples, then see if we should attach a staple
			// here.  To avoid getting multiple staples per junction, we only put a
			// staple if we're at least 10 masses away from the previous along the chain.
			// We don't staple masses together if they are too far from the X=0 axis to
			// avoid tying rings together in their middles.  The masses also have to be
			// within a distance threshold to be stapled.
			// Note that the finalP[] array starts over at zero for first_mass.
			unsigned lastMass = 0;
			if (massCount > 0) { lastMass = massCount - 1; }
			if (doStaples && (lastMass > most_recent_staple + 10)) {
				float projectedPoint[3];

				currentPoint[0] = finalP[3 * (lastMass - first_mass)];
				currentPoint[1] = finalP[3 * (lastMass - first_mass) + 1];
				currentPoint[2] = finalP[3 * (lastMass - first_mass) + 2];

				projectedPoint[0] = 0.0f;
				projectedPoint[1] = 0.0f;
				projectedPoint[2] = currentPoint[2];

				if (distance(currentPoint, projectedPoint) < 0.15e-6) {
					for (j = most_recent_staple + 1; j < static_cast<int>(lastMass)-10; j++) {
						nextPoint[0] = finalP[3 * (j - first_mass)];
						nextPoint[1] = finalP[3 * (j - first_mass) + 1];
						nextPoint[2] = finalP[3 * (j - first_mass) + 2];
						if (distance(currentPoint, nextPoint) < 1e-8) {
							if (!missing_condensin[g_num_condensin]){
								fprintf(ofp2, "spring %u %u %lg %lg\n",
									lastMass, j,
									distance(currentPoint, nextPoint),
									spring_constant);
								g_num_condensin++;
							}
							most_recent_staple = lastMass;
							break;
						}
					}
				}
			}
		}

		tips.push_back(tip1_bead);
		tips.push_back(tip2_bead);
	}

	//rings are only created for chromosomes < 16.  
	if (doRings && current_chromosome < 16) {

		unsigned whichLoop;
		unsigned numlinks = 3;
		double xAdd = (-0.165 + cohesin_x_offset_microns)*1e-6;//-0.8;
		double yAdd[3];
		yAdd[0] = 0.41e-7;// 0.3;
		yAdd[1] = 0.04e-7;
		yAdd[2] = 0.41e-7;//0.3;
		double zAdd = cohesin_offset_microns* 1e-6;
		for (whichLoop = 0; whichLoop<numlinks; whichLoop++) {

			double r;
			int loop = 0;
			//TODO: why doesn't the loop radius that looks good (.255 is made up b/c it looked good)
			// make sense compared with what's in the SpindleB.blend file?  Probably because it
			// is multiplied by 10 to get to the Blender file?
			double loop_radius = 0.0255 * 1e-6;  //0.05 Micron diameter
			double circumference = loop_radius * 2 * acos(-1.0);
			double epsilon = mass_separation;
			double cx = loop * loop_radius * (3.0 / 2.0);
			int first_mass = massCount;	// First mass in this particular ring.
			for (r = 0.0; r < circumference - epsilon; r += mass_separation, massCount++) {
				// Figure out where this mass is on the ring.  This is a ring
				// in the X-Y plane with the specified radius and mass spacing.
				double angle = (2 * acos(-1.0)) * r / circumference;
				double x = (cos(angle) * loop_radius) + xAdd;
				double y = (sin(angle) * loop_radius) + yAdd[whichLoop];
				double z = zAdd;

				// Translate along the X axis by the ring-center coordinate.
				x += cx;

				//TODO: loops need to be rearranged, but the loop rotations are correct.
				//this simply rotates the rings created in the first chromosome to match the 
				//rotations that occur with the dna strands.
				if (current_chromosome != 0)
				{
					//ROTATION
					//initialize useful values for rotations, later.
					double number_of_chromosomes = 16;
					double inner_rad_nm = 85;
					double inner_rad_w = inner_rad_nm / 1e9;
					double sister_tip_deg = 1.6; //extra angle on sisters (- towards, + away)
					double sister_tip_rad = sister_tip_deg * 3.14159265359 / 180.0;
					double mid_angle, angleRot, new_x, new_y, rot_x, rot_y, rot_z;
					mid_angle = 2 * 3.14159265359 * current_chromosome / 16;
					angleRot = mid_angle + 2 * 3.14159265359 / (16 * 6); //0.05;
					new_x = inner_rad_w * cos(angleRot);
					new_y = inner_rad_w * sin(angleRot);
					rot_x = -angleRot + 3.14159265359 / 2 - sister_tip_rad;
					rot_y = 3.14159265359 / 2;
					rot_z = -0.09 + -4.71238898; //(small offset + adjustment in radians of 360/16)

					//These rotations are based on the math to take euclidean angles (used in Blender) to
					//the rotation matrix version used here.
					double xRot2 = (cos(rot_y)*cos(rot_z))*x + (-cos(rot_x)*sin(rot_z) + sin(rot_x)*sin(rot_y)*cos(rot_z))*y + (sin(rot_x)*sin(rot_z) + cos(rot_x)*sin(rot_y)*cos(rot_z)) *z;
					double yRot2 = (cos(rot_y)*sin(rot_z))*x + (cos(rot_x)*cos(rot_z) + sin(rot_x)*sin(rot_y)*sin(rot_z))*y + (-sin(rot_x)*cos(rot_z) + cos(rot_x)*sin(rot_y)*sin(rot_z))*z;
					double zRot2 = (-sin(rot_y))*x + (sin(rot_x)*cos(rot_y))*y + (cos(rot_x)*cos(rot_y))*z;

					//todo: why are the translations unnecessary?!
					x = xRot2; //+ new_x;
					y = yRot2; //+ new_y;
					z = zRot2;// + new_x;		
				}

				colorNumber = 5;
				if (current_chromosome == 0) {
					fprintf(ofp2, "mass %u\t%g\t%g %g %g %d\n", massCount, mass, z, y, x, colorNumber);
				}
				else {
					fprintf(ofp2, "mass %u\t%g\t%g %g %g %d\n", massCount, mass, x, y, z, colorNumber);
				}
				cohesin_beads.push_back(massCount);

				if (massCount >= first_mass + 1) {
					fprintf(ofp2, "  spring %u %u %lg %lg\n", massCount - 1, massCount,
						mass_separation, spring_constant);
				}
				if (massCount >= first_mass + 2) {
					fprintf(ofp2, "  hinge %u %u %u %lg\n", massCount - 2, massCount - 1, massCount,
						hinge_constant);
				}
			}
			xAdd += 0.165e-6;
			// insert negative into bead set to mark end of cohesin for this loop (won't get printed)
			cohesin_beads.push_back(-1);

			// Hook the last mass to the first with a spring and a hinge.
			// Also hook a hinge to the first two and the last
			fprintf(ofp2, "  spring %u %u %lg %lg\n", massCount - 1, first_mass,
				mass_separation, spring_constant);
			fprintf(ofp2, "  hinge %u %u %u %lg\n", massCount - 2, massCount - 1, first_mass,
				hinge_constant);
			fprintf(ofp2, "  hinge %u %u %u %lg\n", massCount - 1, first_mass, first_mass + 1,
				hinge_constant);
		}

	}

	delete[] x;
	delete[] nbasis;

	return massCount;
}

int main(int argc, char *argv[])
{
	const char *outfile_name = NULL;
	std::string endsfile_name;
	std::string tipsfile_name;
	std::string basesfile_name;
	std::string halfloopsfile_name;
	std::string outerloopsfile_name;
	std::string innerloopsfile_name;
	std::string cohesinfile_name;
	std::string array6p8file_name;
	bool beadOutput = true;
	bool customColors = false;
	FILE *colorfile = NULL;
	const char *colorfile_name = "";
	std::vector<int> colorBeads;
	int	doDNA = 1;	// Include the DNA itself?
	int	doStaples = 1;	// Include the condensin staples?
	int	doRings = 1;	// Include the cohesin rings?
	double	cohesin_offset_microns = 0.22;	// Radial offset from the spindle center
	double	cohesin_x_offset_microns = 0.0;	// Lengh-wise offset along the spindle center
	float loop_expansion_factor = 1.0; // Expand loops outward by this factor
	int npts, p1;
	int k = 4;	// Fourth-order splines by default
	char* m_c;	// missing condensin input
	int missing_condensin[128]; // List of missing condensin
	for (int i = 0; i < 128; i++) { missing_condensin[i] = 0; }

	std::vector<int> ends;
	std::vector<int> tips;
	std::vector<int> bases;
	std::vector<int> halfloops;
	std::vector<int> outerloops;
	std::vector<int> innerloops;
	std::vector<int> cohesinbeads;
	std::vector<int> array6p8;

	// Parse the command line to adjust default values.
	int i;
	unsigned  real_params = 0;  //< Number of non-flag parameters
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-node_separation")) {
			if (++i >= argc) { Usage(argv[0]); }
			mass_separation_microns = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-temperature")) {
			if (++i >= argc) { Usage(argv[0]); }
			temperature_Celsius = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-viscosity")) {
			if (++i >= argc) { Usage(argv[0]); }
			viscosity_centiPoise = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-damping_radius_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			damping_radius_factor = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-time_step")) {
			if (++i >= argc) { Usage(argv[0]); }
			time_step_nanoseconds = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-modulus")) {
			if (++i >= argc) { Usage(argv[0]); }
			modulus_gigaPascal = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-dna_radius")) {
			if (++i >= argc) { Usage(argv[0]); }
			dna_radius_nanometers = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-collision_radius_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			collision_radius_factor = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-collision_force_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			collision_force_factor = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-spring_damping_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			spring_damping_factor = atof(argv[i]);
		}
		else if (!strcmp(argv[i], "-random_number_seed")) {
			if (++i >= argc) { Usage(argv[0]); }
			random_number_seed = atoi(argv[i]);
		}
		else if (strcmp(argv[i], "-cohesin_offset_microns") == 0) {
			if (++i > argc) { Usage(argv[0]); }
			cohesin_offset_microns = atof(argv[i]);
		}
		else if (strcmp(argv[i], "-cohesin_x_offset_microns") == 0) {
			if (++i > argc) { Usage(argv[0]); }
			cohesin_x_offset_microns = atof(argv[i]);
		}
		else if (strcmp(argv[i], "-loop_expansion") == 0) {
			if (++i > argc) { Usage(argv[0]); }
			loop_expansion_factor = atof(argv[i]);
		}
		else if (strcmp(argv[i], "-basis_order") == 0) {
			if (++i > argc) { Usage(argv[0]); }
			k = atoi(argv[i]);
		}
		else if (strcmp(argv[i], "-nochromatin") == 0) {
			doDNA = 0;
		}
		else if (strcmp(argv[i], "-no_bead_output") == 0) {
			beadOutput = false;
		}
		else if (strcmp(argv[i], "-colorbeads") == 0) {
			customColors = true;
			if (++i > argc) { Usage(argv[0]); }
			colorfile_name = argv[i];
		}
		else if (strcmp(argv[i], "-bead_directory") == 0) {
			if (++i > argc) { Usage(argv[0]); }
			endsfile_name = argv[i];
			endsfile_name += "/";
			tipsfile_name = basesfile_name = cohesinfile_name = array6p8file_name = endsfile_name;
			halfloopsfile_name = outerloopsfile_name = innerloopsfile_name = endsfile_name;
		}
		else if (strcmp(argv[i], "-missing_condensin") == 0) {
			if (++i > argc) { Usage(argv[0]); }
			m_c = (char*)argv[i];
			char *token = strtok(m_c, ",");
			while (token != NULL) {
				if (atoi(token) < 1 || atoi(token) > 128) Usage(argv[0]);
				else {
					missing_condensin[atoi(token) - 1] = 1;
					token = strtok(NULL, ",");
				}
			}
		}
		else if (strcmp(argv[i], "-nocohesin") == 0) {
			doRings = 0;
		}
		else if (strcmp(argv[i], "-nocondensin") == 0) {
			doStaples = 0;
		}
		else {
			switch (real_params++) {
			case 0:
				outfile_name = argv[i];
				break;
			default:
				Usage(argv[0]);
			}
		}
	}

	// Compute the derived values that we feed to the simulator.
	time_step_seconds = nanoseconds_to_seconds(time_step_nanoseconds);
	node_separation_meters = microns_to_meters(mass_separation_microns);
	mass_kg = dna_mass_in_kg(node_separation_meters);
	temperature_kelvin = Celsius_to_Kelvin(temperature_Celsius);
	viscosity_Pascal_seconds = centiPoise_to_Pascal_seconds(viscosity_centiPoise);
	dna_radius_meters = nanometers_to_meters(dna_radius_nanometers);
	effective_damping_radius = damping_radius_factor * node_separation_meters;

	spring_constant = dna_tensile_spring_constant(
		node_separation_meters,
		gigaPascal_to_Pascal(modulus_gigaPascal),
		dna_radius_meters
		);
	hinge_constant = dna_bending_spring_constant(
		node_separation_meters,
		gigaPascal_to_Pascal(modulus_gigaPascal),
		dna_radius_meters
		);
	random_force = Brownian_force_equivalent(
		node_separation_meters,
		temperature_kelvin,
		viscosity_Pascal_seconds,
		time_step_seconds
		);
	mass_damping = mass_damping_equivalent(
		viscosity_Pascal_seconds,
		effective_damping_radius,
		mass_kg);
	mass_radius = collision_radius_factor * (node_separation_meters / 2.0);
	collision_spring_constant = spring_constant * collision_force_factor;


	// Figure out where to put the output file.  Open it if needed.
	FILE *out = stdout;
	if (outfile_name != NULL) {
		out = fopen(outfile_name, "wb");
		if (out == NULL) {
			perror("Could not open output file");
			return -1;
		}
	}

	// if coloring beads listed in a file, open the file and extract bead numbers
	if (customColors) {
		if (strlen(colorfile_name) == 0) {
			Usage(argv[0]);
		}
		else if (NULL == (colorfile = fopen(colorfile_name, "r"))) {
			perror("Cannot open color file for reading");
			exit(-2);
		}
		int beadNum;
		while (fscanf(colorfile, "%d", &beadNum) == 1) {
			colorBeads.push_back(beadNum);
		}
	}

	if (doRings) {
		endsfile_name += "ends_coh.csv";
		tipsfile_name += "tips_coh.csv";
		basesfile_name += "bases_coh.csv";
		halfloopsfile_name += "halfloops_coh.csv";
		outerloopsfile_name += "outerloops_coh.csv";
		innerloopsfile_name += "innerloops_coh.csv";
		cohesinfile_name += "cohesin_beads.csv";
		array6p8file_name += "6p8_coh.csv";
	}
	else {
		endsfile_name += "ends_no_coh.csv";
		tipsfile_name += "tips_no_coh.csv";
		basesfile_name += "bases_no_coh.csv";
		halfloopsfile_name += "halfloops_no_coh.csv";
		outerloopsfile_name += "outerloops_no_coh.csv";
		innerloopsfile_name += "innerloops_no_coh.csv";
		array6p8file_name += "6p8_no_coh.csv";
	}

	// Print out the header information and general parameters
	// for the simulation.  This includes medatadata lines and then
	// the beginning of the structure with its parameters
	fprintf(out, "meta temperature_Celsius %lg\n", temperature_Celsius);
	fprintf(out, "meta viscosity_centiPoise %lg\n", viscosity_centiPoise);
	fprintf(out, "meta effective_damping_radius %lg\n", effective_damping_radius);
	fprintf(out, "meta dna_modulus_gigaPascal %lg\n", modulus_gigaPascal);
	fprintf(out, "meta dna_radius_nanometers %lg\n", dna_radius_nanometers);
	fprintf(out, "meta damping_radius_factor %lg\n", damping_radius_factor);
	fprintf(out, "structure {\n");
	fprintf(out, "  random_force %lg\n", random_force);
	fprintf(out, "  mass_damping %lg\n", mass_damping);
	fprintf(out, "  mass_radius %lg\n", mass_radius);
	fprintf(out, "  time_step %lg\n", time_step_seconds);
	fprintf(out, "  collision_spring_constant %lg\n", collision_spring_constant);
	fprintf(out, "  spring_damping_factor %lg\n", spring_damping_factor);
	fprintf(out, "  random_number_seed %d\n", random_number_seed);
	fprintf(out, "  color %d\n", 1);

	// Put the masses into the system.  Attach a spring between each
	// pair of masses.  Attach a hinge between each triplet of
	// masses.

	npts = 32;	// 32 copies of the spline are created.
	p1 = 5000;   	/* Number of points on curve */

	float *b = new float[(npts + 1) * 3];
	float *h = new float[npts + 1];
	float *p = new float[(p1 + 1) * 3];
	float *finalP = new float[(p1 + 1) * 3];
	int springMasses[8];

	//initialize springMasses to be 0, for all 8
	for (i = 0; i < 8; i++) {
		springMasses[i] = 0;
	}
	for (i = 1; i <= 3 * npts; i++) {
		b[i] = 0.;
	}

	/*	set all homogeneous weighting factors to 1.0 */
	for (i = 1; i <= npts; i++) {
		h[i] = 1.0;
	}

	/*  vary the homogeneous weighting factor 0, 0.25, 1.0, 5.0 */
	h[3] = 1;

	// Set all of the elements in the p[] and finalP[] lists
	// to 0;
	for (i = 1; i <= 3 * p1; i++) {
		p[i] = 0.;
		finalP[i] = 0.;
	}

	/*
	Define the control polygon, Ex. 4.1, p. 140 in the z=1 plane because
	this is three dimensional routine. x=b[1], y=b[2], z=b[3], etc.
	*/
	FILE *ifp;
	const char *infilename = "spindle_chromatin_control_points.txt";
	ifp = fopen(infilename, "r");
	if (ifp == NULL) {
		fprintf(stderr, "Could not open %s for reading\n", infilename);
		return -1;
	}

	int numSplines = 0;
	int massCount = 0;

	//32 should be a full model
	//just 2 for one loopy thing
	while (numSplines < 32) {
		int j = 1;
		float x;
		float y;
		float z;
		char line[80];
		for (int numP = 0; numP < 32; numP++) {
			fgets(line, 80, ifp);
			sscanf(line, "%f %f %f", &x, &y, &z);
			// test if control point is part of loop, and alter it accordingly
			float xyPt[3] = { x, y, 0 };
			float zeroPt[3] = { 0, 0, 0 };
			float dist = distance(xyPt, zeroPt);
			if ((dist >= 0.46 && dist <= 0.47) || (dist >= 0.24 && dist <= 0.25)) {
				x *= loop_expansion_factor;
				y *= loop_expansion_factor;
			}
			// Convert from microns to meters.
			b[j] = x * 1e-6f;
			b[j + 1] = y * 1e-6f;
			b[j + 2] = z * 1e-6f;
			j = j + 3;
		}
		massCount = rbspline(out, npts, k, p1, b, h, p, finalP, massCount, numSplines, springMasses, missing_condensin, doDNA, doStaples, doRings,
			cohesin_offset_microns * loop_expansion_factor, cohesin_x_offset_microns, ends, tips, bases, halfloops,
			outerloops, innerloops, cohesinbeads, array6p8, colorBeads);
		numSplines++;
	}

	std::ofstream endsfile, tipsfile, basesfile, halfloopsfile, outerloopsfile,
		innerloopsfile, cohesinfile, array6p8file;
	if (beadOutput) {
		endsfile.open(endsfile_name, std::ofstream::out);
		tipsfile.open(tipsfile_name, std::ofstream::out);
		basesfile.open(basesfile_name, std::ofstream::out);
		halfloopsfile.open(halfloopsfile_name, std::ofstream::out);
		outerloopsfile.open(outerloopsfile_name, std::ofstream::out);
		innerloopsfile.open(innerloopsfile_name, std::ofstream::out);
		array6p8file.open(array6p8file_name, std::ofstream::out);
		if (doRings) {
			cohesinfile.open(cohesinfile_name, std::ofstream::out);
		}
	}

	// we subtract the bead numbers from massCount because Chromoshake's output files list
	// the positions of masses in reverse order at each timestep, so the first bead in this program
	// corresponds to the Mth bead in the chromoshake outfile timesteps, where M = massCount.
	// So we must do this subtraction to accound for the inverting of the list of masses.
	for (std::vector<int>::iterator it = ends.begin(); it != ends.end(); ++it) {
		endsfile << massCount - *it - 1 << std::endl;
	}
	for (std::vector<int>::iterator it = tips.begin(); it != tips.end(); ++it) {
		tipsfile << massCount - *it - 1 << std::endl;
	}
	for (std::vector<int>::iterator it = bases.begin(); it != bases.end(); ++it) {
		basesfile << massCount - *it - 1 << std::endl;
	}
	bool beginningOfLine = true;
	for (std::vector<int>::iterator it = halfloops.begin(); it != halfloops.end(); ++it) {
		if (*it < 0) {
			halfloopsfile << std::endl;
			beginningOfLine = true;
		}
		else if (beginningOfLine) {
			halfloopsfile << massCount - *it - 1;
			beginningOfLine = false;
		}
		else {
			halfloopsfile << ", " << massCount - *it - 1;
		}
	}
	beginningOfLine = true;
	for (std::vector<int>::iterator it = outerloops.begin(); it != outerloops.end(); ++it) {
		if (*it < 0) {
			outerloopsfile << std::endl;
			beginningOfLine = true;
		}
		else if (beginningOfLine) {
			outerloopsfile << massCount - *it - 1 - 1;
			beginningOfLine = false;
		}
		else {
			outerloopsfile << ", " << massCount - *it - 1;
		}
	}
	beginningOfLine = true;
	for (std::vector<int>::iterator it = innerloops.begin(); it != innerloops.end(); ++it) {
		if (*it < 0) {
			innerloopsfile << std::endl;
			beginningOfLine = true;
		}
		else if (beginningOfLine) {
			innerloopsfile << massCount - *it - 1;
			beginningOfLine = false;
		}
		else {
			innerloopsfile << ", " << massCount - *it - 1;
		}
	}
	beginningOfLine = true;
	for (std::vector<int>::iterator it = cohesinbeads.begin(); it != cohesinbeads.end(); ++it) {
		if (*it < 0) {
			cohesinfile << std::endl;
			beginningOfLine = true;
		}
		else if (beginningOfLine) {
			cohesinfile << massCount - *it - 1;
			beginningOfLine = false;
		}
		else {
			cohesinfile << ", " << massCount - *it - 1;
		}
	}
	beginningOfLine = true;
	for (std::vector<int>::iterator it = array6p8.begin(); it != array6p8.end(); ++it) {
		if (*it < 0) {
			array6p8file << std::endl;
			beginningOfLine = true;
		}
		else if (beginningOfLine) {
			array6p8file << massCount - *it - 1;
			beginningOfLine = false;
		}
		else {
			array6p8file << ", " << massCount - *it - 1;
		}
	}

	// Close off the structure.
	fprintf(out, "}\n");
	fclose(out);
	if (endsfile.is_open()) endsfile.close();
	if (tipsfile.is_open()) tipsfile.close();
	if (basesfile.is_open()) basesfile.close();
	if (halfloopsfile.is_open()) halfloopsfile.close();
	if (outerloopsfile.is_open()) outerloopsfile.close();
	if (innerloopsfile.is_open()) innerloopsfile.close();
	if (cohesinfile.is_open()) cohesinfile.close();

	delete[] finalP;
	delete[] p;
	delete[] h;
	delete[] b;

	return 0;
}
