// solverIMEXEuler.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "solverIMEXEuler.h"

// the name of the solver
const char * solverIMEXEuler::getName( )
{
    return "IMEX Euler Solver";
}

// destructor
solverIMEXEuler::~solverIMEXEuler( )
{
    clReleaseKernel( clKernelIntegrate );
    clReleaseMemObject( clInverseMass );
    clReleaseMemObject( clDampingScale );
}

// constructor
solverIMEXEuler::solverIMEXEuler(
    std::string sim_CLdir, // the path of directory with our openCL source code
    cl_context sim_context, // OpenCL context passed in from simulation
    cl_command_queue sim_queue, // OpenCL queue passed in from simulation
    CS_REAL sim_timeStep, // the time step of the simulation
    CS_REAL3 * sim_position, // the initial configuration of positions
    cl_ulong sim_numParticles, // the number of particles in the simulation
    cl_device_id sim_deviceId, // the device id to create the program on
    CS_REAL * mass, // the array of mass values
    CS_REAL * massDamping // the array of mass damping values
    ) : genericSolver( sim_CLdir, sim_context, sim_queue, sim_timeStep, 
        sim_position, sim_numParticles )
{
    // read errors from OpenCL
    cl_int error;

    // set up clInverseMass
    clInverseMass = clCreateBuffer(
        context, // the CL context
        CL_MEM_READ_ONLY, // we never change in kernel
        numParticles * sizeof( CS_REAL ), // space to allocate
        NULL, &error );
    checkErr( error, "clCreateBuffer (clInverseMass)" );
    // calculate the values to send to device
    CS_REAL * inverseMass = new CS_REAL[numParticles];
    for (int i = 0; i < numParticles; ++i)
    {
        inverseMass[i] = static_cast<CS_REAL>( timeStep / mass[i] );
    }
    // send to device
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // the queue to use
            clInverseMass, // write to this buffer
            CL_TRUE, // this is a blocking operation
            0, // no offset,
            numParticles * sizeof( CS_REAL ), // space to allocate
            inverseMass, // the buffer to copy from
            0, NULL, NULL // not waiting for any events
            ),
        "clEnqueueWriteBuffer (clInverseMass)"
        );
    // delete temporary buffer
    delete[] inverseMass;

    // set up clDampingScale
    clDampingScale = clCreateBuffer(
        context, CL_MEM_READ_ONLY, numParticles * sizeof( CS_REAL ),
        NULL, &error );
    checkErr( error, "clCreateBuffer (clDampingScale)" );
    // calculate values to send to device
    CS_REAL * dampingScale = new CS_REAL[numParticles];
    for (int i = 0; i < numParticles; ++i)
    {
        dampingScale[i] = static_cast<CS_REAL>( 1. / (1. + massDamping[i]*timeStep) );
    }
    // send to device
    checkErr(
        clEnqueueWriteBuffer( queue, clDampingScale, CL_TRUE,
            0, numParticles * sizeof( CS_REAL ), dampingScale,
            0, NULL, NULL
            ),
        "clEnqueueWriteBuffer (clDampingScale)"
        );
    // delete temporary buffer
    delete[] dampingScale;


    // begin setting up clKernelIntegrate
    // load the program "solverIMEXEuler.cl"
    loadProgram( sim_deviceId, "solverIMEXEuler.cl" );

    clKernelIntegrate = clCreateKernel( program, "integrateMotionIMEX", &error );
    checkErr( error, "clCreateKernel" );
    error = clSetKernelArg( clKernelIntegrate, 0, sizeof(CS_REAL), &timeStep );
    error |= clSetKernelArg( clKernelIntegrate, 1, sizeof(cl_mem), &clInverseMass );
    error |= clSetKernelArg( clKernelIntegrate, 2, sizeof(cl_mem), &clDampingScale );
    error |= clSetKernelArg( clKernelIntegrate, 3, sizeof(cl_mem), &clPosition );
    error |= clSetKernelArg( clKernelIntegrate, 4, sizeof(cl_mem), &clVelocity );
    error |= clSetKernelArg( clKernelIntegrate, 6, sizeof(cl_ulong), &numParticles );
    checkErr( error, "clSetKernelArg (arguments 0:4,6)" );

    // get optimal global item size
    clKernelIntegrateGlobalSize = getKernelPaddedGlobalSize( 
        sim_deviceId, clKernelIntegrate, numParticles );
}

// function for attaching the forces to integrate against and set up remaining arguments
void solverIMEXEuler::attachForces(
    sumForces * totalForce
    )
{
    // call base class attachForces first
    genericSolver::attachForces( totalForce ); // sets theForces and sets forcesAttached to true
    
    // get the memory objects to be passed into the integration kernel from the force objects
    cl_mem clTheForce = theForces->getCLForce();

    cl_int error;
    error = clSetKernelArg( clKernelIntegrate, 5, sizeof(cl_mem), &clTheForce );
    checkErr( error, "clSetKernelArg (arguments 5)" );
}

// go forward a time step
void solverIMEXEuler::integrateStep( )
{
    if ( !forcesAttached )
    {
        checkErr( -100, "Attempting to integrate when forces unattached." );
    }
    // first, solve for the forces
    theForces->calculateForce( );

    // TODO: synchronization? (right now we're using in order queue so doesn't matter)

    // do the integration with the updated force
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // CL queue
            clKernelIntegrate, // the integration kernel
            1, // 1d work space
            NULL, // no offsets
            &clKernelIntegrateGlobalSize, // the chosen global size of the problem
            NULL, // local size chosen by implementation
            0, NULL, NULL // not waiting for events
            ),
        "clEnqueueNDRangeKernel");

    positionUpdated = false;
    velocityUpdated = false;
}