// solverImplicitEuler.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "solverImplicitEuler.h"

// name of our solver
const char * solverImplicitEuler::getName( )
{
    return "Implicit Euler Solver";
}

// destructor
solverImplicitEuler::~solverImplicitEuler( )
{
    if ( forcesAttached )
    {
        // delete memory allocated for the viennacl objects
        delete vclMatrixSystem;
        delete vclRhsVector;
        delete vclSolutionVector;

        // release the kernels
        clReleaseKernel( kernelGenerateSystem );
        clReleaseKernel( kernelApplyVelocityUpdate );

        // release the OpenCL buffers
        // system matrix
        clReleaseMemObject( sysElementBuffer );
        clReleaseMemObject( sysColumnBuffer );
        clReleaseMemObject( sysRowBuffer );
        // rhs vector
        clReleaseMemObject( vectorRhsBuffer );
        // solution vector
        clReleaseMemObject( vectorSolutionBuffer );
    }
    
    // the inverse masses
    clReleaseMemObject( clInverseMass );
}

// constructor
solverImplicitEuler::solverImplicitEuler (
    std::string sim_CLdir, // the path of directory with our openCL source code
    cl_context sim_context, // OpenCL context passed in from simulation
    cl_command_queue sim_queue, // OpenCL queue passed in from simulation
    CS_REAL sim_timeStep, // the time step of the simulation
    CS_REAL3 * sim_position, // the initial configuration of positions
    cl_ulong sim_numParticles, // the number of particles in the simulation
    cl_device_id sim_deviceId, // the device id to create the program on
    CS_REAL * mass // the array of mass values
    ) : genericSolver( sim_CLdir, sim_context, sim_queue, sim_timeStep, sim_position, sim_numParticles )
{
    // read errors from OpenCL
    cl_int error;

    // set up clInverseMass
    clInverseMass = clCreateBuffer(
        context, // the context
        CL_MEM_READ_ONLY, // we never change this buffer in kernel
        numParticles * sizeof( CS_REAL ), // space to allocate
        NULL, &error );
    checkErr( error, "clCreateBuffer (clInverseMass)" );
    // calculate values to send to device
    CS_REAL * inverseMass = new CS_REAL[numParticles];
    for ( int i = 0; i < numParticles; ++i )
    {
        inverseMass[i] = static_cast<CS_REAL>(timeStep / mass[i]);
    }
    // send to device
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
        queue, // the queue to use
        clInverseMass, // write to this buffer
        CL_TRUE, // this is a blocking operation
        0, // no offset,
        numParticles * sizeof( CS_REAL ), // space to allocate
        inverseMass, // the buffer to copy from
        0, NULL, NULL // not waiting for any events
        ),
        "clEnqueueWriteBuffer (clInverseMass)"
        );
    // delete temporary buffer
    delete[] inverseMass;

    // load the program in preparation of setting up our kernels
    // load the program "solverImplicitEuler.cl"
    if (!loadProgram( sim_deviceId, "solverImplicitEuler.cl" )) {
	broken = true;
	return;
    }

    // set up kernels and global sizes, but do not add arguments yet
    // create kernel for generating system
    kernelGenerateSystem = clCreateKernel( program, "generateSystem", &error );
    checkErr( error, "clCreateKernel (kernelGenerateSystem)" );
    // get global size to use
    kernelGenerateSystemGlobalSize = getKernelPaddedGlobalSize(
        sim_deviceId, kernelGenerateSystem, numParticles );

    // create kernel for applying solved velocity update
    kernelApplyVelocityUpdate = clCreateKernel( program, "applyVelocityUpdate", &error );
    kernelApplyVelocityUpdateGlobalSize = getKernelPaddedGlobalSize(
        sim_deviceId, kernelApplyVelocityUpdate, numParticles );

    // set up ViennaCL context, even though we won't allocate any vectors or such
    // until we attach the forces
    // pass our context/device/queue
    viennacl::ocl::setup_context( 0, context, sim_deviceId, queue );
    viennacl::ocl::switch_context( 0 );

    // the real set up occurs when we attach the forces...
}

// attach forces to this
void solverImplicitEuler::attachForces(
    sumForces * totalForce
    )
{
    // first, call base class attachForces to update theForces and forcesAttached
    genericSolver::attachForces( totalForce );

    // use this to set up kernels and ViennaCL system of equations
    
    cl_mem clTheForce = theForces->getCLForce( );
    forceJacobian clTheJacobian = theForces->getCLJacobian( );
    cl_ulong maxInteractions = theForces->getJacobianMaxInteractions( );

    cl_int error; // keep track of errors from OpenCL API calls

    // set up buffers for system matrix
    // row buffer
    sysRowBuffer = clCreateBuffer( context, CL_MEM_READ_WRITE, // ??? could probably make this READ_ONLY.
        ((3 * numParticles) + 1) * sizeof( cl_uint ), // we have 3 * numParticles rows
        NULL, &error );
    checkErr( error, "clCreateBuffer (sysRowBuffer)" );
    // element buffer and column buffer
    sysElementBuffer = clCreateBuffer( context, CL_MEM_READ_ONLY,
        // we have 3*N position or velocity coordinates, and they interact with (3 * maxInteractions) other coordinates
        (3 * numParticles) * (3 * maxInteractions) * sizeof( CS_REAL ),
        NULL, &error );
    checkErr( error, "clCreateBuffer (sysElementBuffer)" );
    sysColumnBuffer = clCreateBuffer( context, CL_MEM_READ_ONLY,
        // we have 3*N position or velocity coordinates, and they interact with (3 * maxInteractions) other coordinates
        (3 * numParticles) * (3 * maxInteractions) * sizeof( cl_uint ),
        NULL, &error );
    checkErr( error, "clCreateBuffer (sysColumnBuffer)" );
    // element and column buffers will be updated by the kernels and don't need to have values set to anything now
    // however, row buffer remains the same throughout simulation, so set it up now
    cl_uint * rows = new cl_uint[(3 * numParticles) + 1];
    for ( cl_uint i = 0; i < (3 * numParticles) + 1; ++i )
    {
        rows[i] = static_cast<cl_uint>(i * (3 * maxInteractions));
    }
    // push rows to sysRowBuffer
    checkErr(
        clEnqueueWriteBuffer(
            queue, // the queue to use
            sysRowBuffer, // the buffer to send our data to
            CL_TRUE, // yes this is a blocking write operation
            0, // no offset
            ((3 * numParticles) + 1) * sizeof( cl_uint ), // the size of the buffer we are sending
            rows, // the buffer we are sending
            0, NULL, NULL ),
        "clEnqueueWriteBuffer (sysRowBuffer)" );
    // delete the temporarily allocated buffer
    delete[] rows;

    // set up buffers for right hand side and solution vectors
    vectorRhsBuffer = clCreateBuffer( context, CL_MEM_READ_WRITE,
        // we have 3*N coordinates
        (3 * numParticles) * sizeof( CS_REAL ),
        NULL, &error );
    checkErr( error, "clCreateBuffer (vectorRhsBuffer)" );
    vectorSolutionBuffer = clCreateBuffer( context, CL_MEM_READ_WRITE,
        (3 * numParticles) * sizeof( CS_REAL ),
        NULL, &error );
    checkErr( error, "clCreateBuffer (vectorSolutionBuffer)" );
    // we don't need to worry about setting up the values because they will be updated by
    // kernels or by ViennaCL

    // set up arguments to kernels for generating system and taking solution to update position and velocity
    // set up parameters for kernelGenerateSystem
    error = clSetKernelArg( kernelGenerateSystem, 0, sizeof(CS_REAL), &timeStep );
    error |= clSetKernelArg( kernelGenerateSystem, 1, sizeof(cl_mem), &clInverseMass );
    error |= clSetKernelArg( kernelGenerateSystem, 2, sizeof(cl_mem), &clVelocity );
    error |= clSetKernelArg( kernelGenerateSystem, 3, sizeof(cl_mem), &clTheForce );
    error |= clSetKernelArg( kernelGenerateSystem, 4, sizeof(cl_mem), &clTheJacobian.j_PosX );
    error |= clSetKernelArg( kernelGenerateSystem, 5, sizeof(cl_mem), &clTheJacobian.j_PosY );
    error |= clSetKernelArg( kernelGenerateSystem, 6, sizeof(cl_mem), &clTheJacobian.j_PosZ );
    error |= clSetKernelArg( kernelGenerateSystem, 7, sizeof(cl_mem), &clTheJacobian.j_VelX );
    error |= clSetKernelArg( kernelGenerateSystem, 8, sizeof(cl_mem), &clTheJacobian.j_VelY );
    error |= clSetKernelArg( kernelGenerateSystem, 9, sizeof(cl_mem), &clTheJacobian.j_VelZ );
    error |= clSetKernelArg( kernelGenerateSystem, 10, sizeof(cl_mem), &clTheJacobian.interactionParticleIndex );
    error |= clSetKernelArg( kernelGenerateSystem, 11, sizeof(cl_mem), &sysElementBuffer );
    error |= clSetKernelArg( kernelGenerateSystem, 12, sizeof(cl_mem), &sysColumnBuffer );
    error |= clSetKernelArg( kernelGenerateSystem, 13, sizeof(cl_mem), &vectorRhsBuffer );
    error |= clSetKernelArg( kernelGenerateSystem, 14, sizeof(cl_ulong), &maxInteractions );
    error |= clSetKernelArg( kernelGenerateSystem, 15, sizeof(cl_ulong), &numParticles );
    checkErr( error, "clSetKernelArg (kernelGenerateSystem)" );
    // set up parameters for kernelApplyVelocityUpdate
    error = clSetKernelArg( kernelApplyVelocityUpdate, 0, sizeof(CS_REAL), &timeStep );
    error |= clSetKernelArg( kernelApplyVelocityUpdate, 1, sizeof(cl_mem), &vectorSolutionBuffer );
    error |= clSetKernelArg( kernelApplyVelocityUpdate, 2, sizeof(cl_mem), &clPosition );
    error |= clSetKernelArg( kernelApplyVelocityUpdate, 3, sizeof(cl_mem), &clVelocity );
    error |= clSetKernelArg( kernelApplyVelocityUpdate, 4, sizeof(cl_ulong), &numParticles );
    checkErr( error, "clSetKernelArg(kernelApplyVelocityUpdate)" );

    // set up ViennaCL
    // pass our context/device/queue done in constructor
    // set up our system of equations to our allocated memory
    vclMatrixSystem = new viennacl::compressed_matrix<CS_REAL>( 
        sysRowBuffer, // the rows in our matrix
        sysColumnBuffer, // the buffer with the columns in our matrix
        sysElementBuffer, // the buffer containing the elements of our matrix
        3 * numParticles, // the number of rows in our matrix
        3 * numParticles, // the number of columns in our matrix
        (3 * numParticles) * (3 * maxInteractions) // the number of nonzero elements
        );
    vclRhsVector = new viennacl::vector<CS_REAL>(
        vectorRhsBuffer, // the buffer with the elements of our vector
        3 * numParticles // the length of our vector
        );
    vclSolutionVector = new viennacl::vector<CS_REAL>(
        vectorSolutionBuffer, // the buffer with the elements of our vector
        3 * numParticles // the length of our vector
        );

    viennacl::backend::finish(); // just because I saw it in an example
}

// integrate forward by one time step
void solverImplicitEuler::integrateStep( )
{
    if ( !forcesAttached ) {
        checkErr( -100, "Attempting to integrate when forces unattached." );
    }

    // first, solve for the forces and Jacobians
    theForces->calculateForceAndJacobian( );

    // use the updated force and Jacobian to set up our system of equations and integrate
    
    // synchronization?
    // make sure the step completes
    //checkErr( clFlush( queue ), "clFlush (after calculateForceAndJacobian)" );
    //checkErr( clFinish( queue ), "clFinish (after calculateForceAndJacobian)" );

    // update ViennaCL buffers using calculated force and Jacobian
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // CL queue
            kernelGenerateSystem, // the kernel for generating the (CSR sparse) matrix and rhs vector
            1, // 1d work space
            NULL, // no offsets
            &kernelGenerateSystemGlobalSize, // the global size for this kernel
            NULL, // local size chosen by implementation
            0, NULL, NULL // not waiting for events
            ),
        "clEnqueueNDRangeKernel (kernelGenerateSystem)" );

    // synchronization?
    // make sure the step completes
    //checkErr( clFlush( queue ), "clFlush (after kernelGenerateSystem)" );
    //checkErr( clFinish( queue ), "clFinish (after kernelGenerateSystem)" );

    // solve the associated linear system

    // Check the Jacobian (for debugging)
    /*
    for (int i = 0; i < 3*numParticles; ++i) {
        for (int j = 0; j < 3*numParticles; ++j) {
            std::cout << (*vclMatrixSystem)(i,j) << " ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
    */

    // apply a preconditioner?

    // solve the linear system
    *vclSolutionVector = viennacl::linalg::solve( *vclMatrixSystem, 
        *vclRhsVector, viennacl::linalg::cg_tag( ) );

    // synchronization?
    // make sure the step completes
    //checkErr( clFlush( queue ), "clFlush (after solve)" );
    //checkErr( clFinish( queue ), "clFinish (after solve)" );

    // update position and velocity using velocity update vector
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // CL queue
            kernelApplyVelocityUpdate, // the kernel for applying the velocity update
            1, // 1d work space
            NULL, // no offsets
            &kernelApplyVelocityUpdateGlobalSize, // the global size for this kernel
            NULL, // local size chosen by implementation
            0, NULL, NULL // not waiting for events
            ),
        "clEnqueueNDRangeKernel (kernelApplyVelocityUpdate)" );

    // make sure the step completes
    //checkErr( clFlush( queue ), "clFlush (after kernelApplyVelocityUpdate)" );
    //checkErr( clFinish( queue ), "clFinish (after kernelApplyVelocityUpdate)" );

    // tell the solver that the position 
    // and velocity are no longer up to date
    positionUpdated = false;
    velocityUpdated = false;
}
