// solverImplicitEuler.h
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "genericSolver.h"

#define VIENNACL_WITH_OPENCL
#include "viennacl/scalar.hpp"
#include "viennacl/vector.hpp"
#include "viennacl/compressed_matrix.hpp"
#include "viennacl/linalg/cg.hpp"
#include "viennacl/linalg/bicgstab.hpp"
#include "viennacl/linalg/gmres.hpp"


#ifndef SOLVERIMPLICITEULER_H
#define SOLVERIMPLICITEULER_H

class solverImplicitEuler : public genericSolver
{
protected :
    // We are solving the equations of motion by solving a linear system
    // say A*x = b. A is the system matrix, b is the right hand side vector
    // and x is the solution vector.
    //
    // In the Implicit Euler method, `x` = dv
    // `b` = (dt * M^-1)*(f + (dt * df/dx)*v)
    // `A` = (I - (dt * M^-1) * (df/dv + (dt*df/dx)))
    //      dv is the change in velocity from v(t) to v(t+dt)
    //      dt is the time step
    //      M is the mass matrix (diagonal)
    //      f is the force evaluated at (x(t),v(t),t) (current time)
    //      df/dx is the Jacobian of f with respect to x, evaluated at current time
    //      v is v(t) (current velocity)
    //      df/dv is the Jacobian of f with respect to v, evaluated at current time

    // define the system matrix
    viennacl::compressed_matrix<CS_REAL> * vclMatrixSystem;
    // defined by three OpenCL buffers (along with numParticles)
    cl_mem sysElementBuffer; // array with nonzero elements of vclMatrixSystem
    cl_mem sysColumnBuffer; // the column index of each element (set each time by 
    cl_mem sysRowBuffer; // the indices in element/columnBuffer at which new rows start (set once during constructor)

    // define the right hand side
    viennacl::vector<CS_REAL> * vclRhsVector;
    // defined by one OpenCL buffer (along with 3*numParticles for dimension)
    cl_mem vectorRhsBuffer;

    // define the vector for the solution of the conjugate gradient solve
    viennacl::vector<CS_REAL> * vclSolutionVector;
    // defined by one OpenCL buffer
    cl_mem vectorSolutionBuffer;

    // define the openCL buffer with the (scaled by time) inverse masses of the simulation
    cl_mem clInverseMass; // dt / mass per particle

    // define OpenCL kernel for generating the system and rhs
    cl_kernel kernelGenerateSystem;
    size_t kernelGenerateSystemGlobalSize;
    // define OpenCL kernel for taking solution and updating position and velocity
    cl_kernel kernelApplyVelocityUpdate;
    size_t kernelApplyVelocityUpdateGlobalSize;


public :
    // constructor
    solverImplicitEuler (
        std::string sim_CLdir, // the path of directory with our openCL source code
        cl_context sim_context, // OpenCL context passed in from simulation
        cl_command_queue sim_queue, // OpenCL queue passed in from simulation
        CS_REAL sim_timeStep, // the time step of the simulation
        CS_REAL3 * sim_position, // the initial configuration of positions
        cl_ulong sim_numParticles, // the number of particles in the simulation
        cl_device_id sim_deviceId, // the device id to create the program on
        CS_REAL * mass // the array of mass values
        );

    // destructor
    virtual ~solverImplicitEuler ( );

    // attach forces to this
    virtual void attachForces (
        sumForces * totalForce
        );

    // integrate forward by one time step
    virtual void integrateStep ( );

    // what's the name of this solver
    virtual const char * getName ( );

};





#endif