/************************************************************************
 *	This file contains the routines needed to simulate 		*
 * objects as masses tied together by springs and hinges.  It includes 	*
 * the routines that set up the mass spring, and hinge lists, as well as*
 * the routines that control the motion simulation.  The file		*
 * massmesh.h needs to be included by other files that use these	*
 * routines.								*
 *	The routines in this file can be used to model 1, 2, or 	*
 * 3-dimensional objects in this manner.  The difference in		*
 * dimension is only a function of how the springs and hinges are 	*
 * attatched to the masses in the list.					*
 *	The normal usage of these routines is to build a mesh of masses	*
 * connected by springs and/or hinges.  Then a loop is entered in which	*
 * the forces acting on the masses are computed and then the forces are	*
 * applied to the masses.  The application of the forces clears the old	*
 * force values so that new ones can be computed for the next iteration.*
 ************************************************************************/

#include "structureLoader.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef	M_PI
const double M_PI = 2*acos(0.0);
#endif

#ifdef WIN32
#pragma warning( disable : 4996 )
#endif

inline	double	sqr(double a) { return a*a; }

// Return the distance between the two mass nodes.
inline double mass_distance(const MASS_node *m1, const MASS_node *m2)
{
  if ( (m1 == NULL) || (m2 == NULL) ) { return 0.0; }
  double dx = m1->x - m2->x;
  double dy = m1->y - m2->y;
  double dz = m1->z - m2->z;
  return sqrt( dx*dx + dy*dy + dz*dz );
}

/****************************************************************
 *	These are utility routines that allow applications to	*
 * build the mass list, the hinge list and the spring list.	*
 ****************************************************************/

/*	This routine will add a mass to the head of a list of masses.  It gets
 * a pointer to the pointer to the head of the list.  If the head of the
 * list is null, then it is initialized.  In all cases, the new mass
 * becomes the head of the mass list. The initial force and velocity on the
 * new mass are set to zero.
 *	The function returns -1 on failure, 0 on success.
 */

int	add_mass(MASS_node **headptr, double mass, double x, double y, double z,
        int color, double damping, double radius)
{
        MASS_node *newm; 	/* Pointer to the new mass */
        if ( (newm = (MASS_node *)malloc(sizeof(MASS_node))) == NULL ) {
                perror("add_mass (cannot allocate more memory)");
                return(-1);	/* FAIL */
        }

        newm->next = *headptr;	/* Insert before head of the list */
        *headptr = newm;	/* Point the head to this entry */

        newm->mass = mass;	/* Fill in the values for this mass */
        newm->x = x;
        newm->y = y;
        newm->z = z;
        newm->colorNumber = color;
        newm->damping = damping;
        newm->radius = radius;
        newm->vx = newm->vy = newm->vz = 0.0;
        newm->fx = newm->fy = newm->fz = 0.0;

        return(0);		/* SUCCESS */

} /* end of add_mass() */


/*	This routine will add a spring to the head of a list of springs.
 * It gets a pointer to the pointer to the head of the list.  It will put
 * the new spring at the head of the list.  It works properly if the head
 * of the list is NULL to begin with.  m1 and m2 should point to the masses
 * at either end of the spring.
 *	The function returns -1 on failure and 0 on success.
 */

int	add_spring(SPRING_node **headptr, double rest, double k,
        MASS_node *m1, MASS_node *m2, double fbreak, int color)
{
        SPRING_node	*news;	/* The new spring node location */

        if ( (m1 == NULL) || (m2 == NULL) ) {
                perror("add_spring (bad mass)");
                return(-1);	/* BAD MASS */
        }
        if ( (news = (SPRING_node *)malloc(sizeof(SPRING_node))) == NULL) {
                perror("add_spring (cannot allocate more memory)");
                return(-1);	/* FAIL */
        }

        news->next = *headptr;	/* Place new entry at head of list */
        *headptr = news;	/* Make the head point to the new one */

        news->rest = rest;	/* Fill in the parameter values */
        news->k = k;
        news->fbreak = fbreak;
        news->m1 = m1;
        news->m2 = m2;
        news->colorNumber = color;

        return(0);	/* SUCCESS */

} /* end of add_spring() */

/*	This routine will add a hinge to the head of a list of hinges.
 * It gets a pointer to the pointer to the head of the list.  It will put
 * the new hinge at the head of the list.  It works properly if the head
 * of the list is NULL to begin with.  m1, m2 and m3 should point to the
 * masses connected by the hinge.  m2 is in the middle.
 *	The function returns -1 on failure and 0 on success.
 */

int	add_hinge(HINGE_node **headptr, double k,
        MASS_node *m1, MASS_node *m2, MASS_node *m3, int color)
{
        HINGE_node	*newh;	/* The new hinge node location */

        if ( (m1 == NULL) || (m2 == NULL) || (m3 == NULL) ) {
                perror("add_hinge (bad mass)");
                return(-1);	/* BAD MASS */
        }
        if ( (newh = (HINGE_node *)malloc(sizeof(HINGE_node))) == NULL) {
                perror("add_hinge (cannot allocate more memory)");
                return(-1);	/* FAIL */
        }

        newh->next = *headptr;	/* Place new entry at head of list */
        *headptr = newh;	/* Make the head point to the new one */

        newh->k = k;		/* Fill in the parameter values */
        newh->m1 = m1;
        newh->m2 = m2;
        newh->m3 = m3;
        newh->colorNumber = color;

        return(0);	/* SUCCESS */

} /* end of add_hinge() */

// Reads in a previous outfile f for the coordinates of the last time step
// and puts their coordinates into the mass nodes mh.
// returns the value of the last time step, or a negative value (-1.0) if an error occurred
// Also returns the number of steps already taken in the original file.
double parse_continue( FILE *f, MASS_node **mh, int *num_steps )
{
  // Variables used to help parse lines. the line itself is NULL-terminated in case
  // we get a line in the file that is too long and fgets() doesn't read the NULL
  // character
  char line[1024];
  line[sizeof(line)-1] = '\0';
  char ignore[1024];

  // Keeps track of the lines since the last time line that has been
  // read.  This is a buffer that we'll then parse for the actual values once
  // we reach the end of the file.  We used to use ftell() and fseek(), but
  // those fail on files over 2GB.
  std::vector<std::string>	position_lines;

  // record the last time step found
  double last_time = 0.0;

  // loop through file once to find last time
  *num_steps = 0;
  do {
    // read next line:
    fgets( line, sizeof(line) - 1, f );

    if (!(strncmp( line , "Time", 4 ))) {
      // we have reached a new time step.
      // clear out the list of positions so we can start adding new ones.
      position_lines.clear();
      (*num_steps) ++;

      // scan in our new time and check if it failed.
      if ( sscanf( line, "%s %lg", ignore, &last_time ) != 2 ) {
        // we weren't able to fill in both variables. Something went wrong.
        fprintf(stderr,"parse_continue: Line with time was improperly read.\n");
        return -1.0;
      }
    } else {
	position_lines.push_back(line);
    }
  } while( ! ( feof( f ) ) );

  // we've hit the end of our file, but we can go back to starting at the last time.
  // now, we iterate through.
  MASS_node *cur_node = *mh;
  size_t index = 0;
  const char *curline;
  // while the current node is not null, add the next line of coordinates.
  while ( cur_node != NULL ) {
    // get our next line
    curline = position_lines[index].c_str();

    // our coordinates to read into
    double x, y, z;

    // scan in our x y, and z coordinates, check for errors
    if ( sscanf( curline, "%lg %lg %lg", &x, &y, &z) != 3 ) {
      // we weren't able to fill in our coordinates. Not good.
      fprintf(stderr,"parse_continue: Old coordinates improperly read.\n");
      return -1.0;
    }

    // put our read in values into our linked list at the current node
    cur_node->x = x;
    cur_node->y = y;
    cur_node->z = z;

    // move to the next node and continue with the loop.
    cur_node = cur_node->next;
    index++;
  }

  // by now, we've either read in all the coordinates from the old file or screwed up somewhere.
  // assuming that it works, all that's left is returning the time.

  return last_time;
}

// Helper classes and functions for the next function.
class MASS_NAME {
public:
  MASS_NAME(const char *name, MASS_node *node) {
    strncpy(d_name, name, sizeof(d_name)-1);
    d_name[sizeof(d_name)-1] = '\0';
    d_node = node;
  }
  const char *name() const { return d_name; }
  MASS_node *node() const { return d_node; }

  bool operator < (const MASS_NAME &other) const {
	return strcmp(d_name, other.d_name) < 0;
  }

protected:
  char        d_name[1024];
  MASS_node   *d_node;
};

// Look up the mass node with the specified name in the vector of
// mass nodes.  Return a pointer to the mass node, or NULL if there
// was not one with this name.
MASS_node *lookup_mass_node(const std::vector<MASS_NAME> &n, const char *name)
{
  // Look starting at the most recent, which is the most likely.
  int i;
  for (i = static_cast<int>(n.size())-1; i >= 0; i--) {
    if (strcmp(name, n[i].name()) == 0) {
      // Found it!
      return n[i].node();
    }
  }

  // Didn't find it.
  return NULL;
}

// Read in a description of a single structure from a file; this includes
// reading in named masses, springs to connect masses, and hinges to connect
// triples of masses.  The format of the data in the file is:
//    meta PARAMETER VALUE  (zero or more metadata parameters)
//    structure {
//      mass_damping DAMPING  {default 5}
//      random_force FORCE {default 5}
//      time_step STEP {Must be specified}
//      collision_spring_constant K {Default 2}
//      spring_damping_factor F {Default 0}
//      random_number_seed S {Default 42}
//      color INDEX {default 1}  (Associated with all following objects until another color)
//      mass_radius RADIUS {default 0}
//      mass NAME MASS X Y Z                  (or)
//      mass NAME MASS X Y Z DAMPING          (or)
//      mass NAME MASS X Y Z DAMPING RADIUS
//      (however many masses, each with a unique name
//      spring_constant_over_length VALUE   {Defaults to 1.0}
//      rest_length_fraction VALUE   {Defaults to 1.0, 0.5 means extended to twice rest length}
//      spring NAME1 NAME2  {Computes based on spring_constant_over_length and rest_length_fraction and mass distance} (or)
//      spring NAME1 NAME2 REST K                 (or)
//      spring NAME1 NAME2 REST K BREAKING_FORCE
//      (however many springs)
//      hinge NAME1 NAME2 NAME3 K
//      (however many hinges)
//    }
// Returns true if the parsing went okay, leaving the file pointer at
// the beginning of the line following the closing brace.  Returns pointers
// to the described structure's masses, springs, and hinges in the handles,
// NULL for each if there were no entries.
bool    parse_structure_from_file(
            FILE *f                     //< File to read the structure from
            , MASS_node **mh            //< Pointer to an array of mass nodes (returned)
            , SPRING_node **sh          //< Pointer to an array of spring nodes (returned)
            , HINGE_node **hh           //< Pointer to an array of hinge nodes (returned)
            , double *randomForce       //< Random (Brownian) force to apply (returned)
            , double *timeStep          //< Time step to use in the simulation (returned)
            , double *collisionSpringConstant   //< Spring constant for collisions (returned)
            , double *springDampingFactor //< Ratio of drag constant from relative velocity of springs to spring constants
            , int *seed                 //< Random-number seed to use for first simulation run
            , std::vector<std::string>  *metaData   //< Lines describing metadata for the simulation
)
{
  printf("Parsing structure from input file\n");

  // Variables used to help parse lines.  The line itself is NULL-terminated in case
  // we get a line in the file that is too long and fgets() doesn't read the NULL
  // character.
  bool got_time_step = false;
  char  line[1024];
  line[sizeof(line)-1] = '\0';
  char s1[1024], s2[1024];

  // The linked lists used to build up the structure.  Also a name lists
  // that hold names and pointers to the objects to which they belong.

  std::vector<MASS_NAME> mass_names;
  MASS_node   *masses = NULL;
  SPRING_node *springs = NULL;

  HINGE_node  *hinges = NULL;

  // Default values for constants that can be overridden in the file
  double mass_radius = 0.2; //< Invisible masses by default.
  double mass_damping = 5.0;  //< Damping applied to masses that don't specify it.
  int color_number = 1;     //< Global color to use if color not specified for a given element
  double random_force = 5.0; // random force amount
  double time_step = 1.0;
  double spring_constant_over_length = 1.0; //< If spring constant calculated based on distance between masses.
  double rest_length_fraction = 1.0; //< Fraction of initial length that is rest length, if sccbodbm
  double hinge_constant_over_length = 1.0; //< If hinge constant calculated based on distance between masses.
  double collision_spring_constant = 2.0;
  double spring_damping_factor = 0.0;
  int random_number_seed = 42;

  // There can be zero or more "meta" lines, each of which contains metadata.
  // Store them and return them to the caller if there are any.
  metaData->clear();
  if (fgets(line, sizeof(line)-1, f) == NULL) {
    fprintf(stderr,"parse_structure_from_file: Could not read structure line\n");
    return false;
  }
  while (0 == strncmp(line, "meta", 4)) {
      metaData->push_back(line);
      if (fgets(line, sizeof(line)-1, f) == NULL) {
        fprintf(stderr,"parse_structure_from_file: Could not read structure line\n");
        return false;
      }
  }

  // The first line after metadata must have two strings in it, "structure" and "{".
  if (sscanf(line, "%s %s", s1, s2) != 2) {
    fprintf(stderr,"parse_structure_from_file: Too short structure line\n");
    return false;
  }
  if ( (strcmp(s1, "structure") != 0) || (strcmp(s2, "{") != 0) ) {
    fprintf(stderr,"parse_structure_from_file: Bad structure line:\n(%s)", line);
    return false;
  }

  // Read lines until we get to the "}" line, handling each according to its
  // initial word.
  if (fgets(line, sizeof(line)-1, f) == NULL) {
    fprintf(stderr,"parse_structure_from_file: End of structure not found\n");
    return false;
  }
  while (line[0] != '}') {

    // Check the first word in the line to see what kind of line it is,
    // then parse the rest of the line based on that word and take action
    // as appropriate.
    if (sscanf(line, "%s", s1) != 1) {
      // Empty line, which is okay.
      // Get the next line in the file, then head back around the loop.
      if (fgets(line, sizeof(line)-1, f) == NULL) {
        fprintf(stderr,"parse_structure_from_file: End of structure not found\n");
        return false;
      }
      continue;

    } else if (strcmp(s1, "mass_damping") == 0) {

      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &mass_damping) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad mass_damping line: %s\n", line);
        return false;
      }

  } else if (strcmp(s1, "random_force") == 0) {
      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &random_force) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad random_force line: %s\n", line);
        return false;
      }

  } else if (strcmp(s1, "time_step") == 0) {
      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &time_step) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad time_step line: %s\n", line);
        return false;
      }
      got_time_step = true;

  } else if (strcmp(s1, "collision_spring_constant") == 0) {
      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &collision_spring_constant) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad collision_spring_constant line: %s\n", line);
        return false;
      }

  } else if (strcmp(s1, "spring_damping_factor") == 0) {
      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &spring_damping_factor) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad spring_damping_factor line: %s\n", line);
        return false;
      }

  } else if (strcmp(s1, "random_number_seed") == 0) {
      // Find the value for this constant.
      if (sscanf(line, "%s %d", s1, &random_number_seed) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad random_number_seed line: %s\n", line);
        return false;
      }

  } else if (strcmp(s1, "color") == 0) {
      // Find the value for this constant.
      if (sscanf(line, "%s %d", s1, &color_number) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad color line: %s\n", line);
        return false;
      }

  } else if (strcmp(s1, "mass_radius") == 0) {

      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &mass_radius) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad mass_radius line: %s\n", line);
        return false;
      }

    } else if (strcmp(s1, "mass") == 0) {

      // Got a mass line.  Read its name and other parameters, then create a
      // new mass and add it to the mass list and to the mass_name list.  Remember
      // that the first string in the line will be the keyword itself, which should
      // be skipped.  Remember that radius is an optional parameter.
      char name[1024];
      double mass, x,y,z, damping;
      int color;
      double radius = 0.0;
      int ret;
      if ( (ret = sscanf(line, "%s %s %lg %lg %lg %lg %d %lg %lg",
                 s1, name, &mass, &x, &y, &z, &color, &damping, &radius)) < 6) {
        fprintf(stderr,"parse_structure_from_file: Bad mass line: %s\n", line);
        return false;
      }

      //if the color isn't specified, use default
      if (ret <= 6) {
          color = color_number;
      }

      // If the damping wasn't specified, use the default
      if (ret <= 7) {
        damping = mass_damping;
      }

      // If the radius wasn't specified, use the default
      if (ret <= 8) {
        radius = mass_radius;
      }

      // Make sure we don't already have a mass with this name.
      if (lookup_mass_node(mass_names, name) != NULL) {
        fprintf(stderr, "parse_structure_from_file: Mass name repeated: %s\n", name);
        return false;
      }

      // Good name, go ahead and add the new mass.
      if (add_mass(&masses, mass, x,y,z, color, damping, radius) != 0) {
        fprintf(stderr,"parse_structure_from_file: Could not add mass\n");
        return false;
      }
      mass_names.push_back(MASS_NAME(name, masses));

    } else if (strcmp(s1, "spring_constant_over_length") == 0) {

      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &spring_constant_over_length) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad spring_constant_over_length line: %s\n", line);
        return false;
      }

    } else if (strcmp(s1, "rest_length_fraction") == 0) {

      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &rest_length_fraction) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad rest_length_fraction line: %s\n", line);
        return false;
      }

    } else if (strcmp(s1, "spring") == 0) {

      // Got a spring line.  Read its parameters, then create a
      // new spring and add it to the spring list.  Remember
      // that the first string in the line will be the keyword itself, which should
      // be skipped.  Remember that breaking_force is an optional parameter.
      char mass1[1024], mass2[1024];
      double rest_len, k, breaking_force = 1e100;
      int ret;
      if ( (ret = sscanf(line, "%s %s %s %lg %lg %lg",
                 s1, mass1, mass2, &rest_len, &k, &breaking_force)) < 3) {
        fprintf(stderr,"parse_structure_from_file: Bad spring line: %s\n", line);
        return false;
      }
      if (ret == 4) {
        fprintf(stderr,"parse_structure_from_file: Bad spring line: %s\n", line);
        return false;
      }

      // Locate the named masses.
      MASS_node *m1, *m2;
      m1 = lookup_mass_node(mass_names, mass1);
      m2 = lookup_mass_node(mass_names, mass2);
      if ( (m1 == NULL) || (m2 == NULL) ) {
        fprintf(stderr,"parse_structure_from_file: Could not find masses: %s %s\n", mass1, mass2);
        return false;
      }

      // Compute the spring constant and rest length based on mass distance if we weren't
      // told the distance.
      if (ret == 3) {
        rest_len = mass_distance(m1, m2) * rest_length_fraction;
        k = spring_constant_over_length / mass_distance(m1, m2);
      }

      // Add the specified spring.
      if (add_spring(&springs, rest_len, k, m1, m2, breaking_force, color_number) != 0) {
        fprintf(stderr,"parse_structure_from_file: Could not add spring\n");
        return false;
      }

    } else if (strcmp(s1, "hinge_constant_over_length") == 0) {

      // Find the value for this constant.
      if (sscanf(line, "%s %lg", s1, &hinge_constant_over_length) != 2) {
        fprintf(stderr,"parse_structure_from_file: Bad hinge_constant_over_length line: %s\n", line);
        return false;
      }

    } else if (strcmp(s1, "hinge") == 0) {

      // Got a hinge line.  Read its parameters, then create a
      // new hinge and add it to the hinge list.  Remember
      // that the first string in the line will be the keyword itself, which should
      // be skipped.  Remember that K is an optional parameter.
      char mass1[1024], mass2[1024], mass3[1024];
      double k;
      int ret;
      if ( (ret = sscanf(line, "%s %s %s %s %lg",
                 s1, mass1, mass2, mass3, &k)) < 4) {
        fprintf(stderr,"parse_structure_from_file: Bad hinge line: %s\n", line);
        return false;
      }

      // Locate the named masses.
      MASS_node *m1, *m2, *m3;
      m1 = lookup_mass_node(mass_names, mass1);
      m2 = lookup_mass_node(mass_names, mass2);
      m3 = lookup_mass_node(mass_names, mass3);
      if ( (m1 == NULL) || (m2 == NULL) || (m3 == NULL) ) {
        fprintf(stderr,"parse_structure_from_file: Could not find masses: %s %s %s\n", mass1, mass2, mass3);
        return false;
      }

      // Compute the hinge spring constant based on mass distance if we weren't
      // told the distance.  We use the distance from the first mass to the
      // third.
      if (ret == 4) {
        k = hinge_constant_over_length / mass_distance(m1, m3);
      }

      // Add the specified hinge.
      if (add_hinge(&hinges, k, m1, m2, m3, color_number) != 0) {
        fprintf(stderr,"parse_structure_from_file: Could not add hinge\n");
        return false;
      }
      //printf("XXX Warning: Hinge code is broken and likely to blow up.\n");

    } else {
      // Didn't recognize the keyword starting this line.
      fprintf(stderr,"parse_structure_from_file: Unrecognized keyword on line: %s\n", line);
      return false;
    }

    // Get the next line in the file.
    if (fgets(line, sizeof(line)-1, f) == NULL) {
      fprintf(stderr,"parse_structure_from_file: End of structure not found\n");
      return false;
    }
  }

  if (!got_time_step) {
      fprintf(stderr,"parse_structure_from_file: Time step not specified in file.\n");
      return false;
  }

  // Put the array values into the return parameters.
  printf("Done parsing input file\n");
  *mh = masses;
  *sh = springs;
  *hh = hinges;
  *randomForce = random_force;
  *timeStep = time_step;
  *collisionSpringConstant = collision_spring_constant;
  *springDampingFactor = spring_damping_factor;
  *seed = random_number_seed;
  return true;
}
