/************************************************************************
 *      This file contains routines to parse configuration files and fill
 * in parameters, including building structures of masses connected by
 * springs and hinges.  It also includes a routine to read in a simulation
 * output file so that we can continue from the end of a previous run.
 *	This file was originally copied from
 * the massmesh.c/h files from the WebSlinger application.
 ************************************************************************/

#pragma once
#include <stdio.h>
#include <vector>
#include <string>

/*	The masses are stored in a linked list.  The last mass has a null
 * next pointer.  Each mass points stores its mass, its location, its
 * velocity when last computed, the damping factor (the percent of velocity
 * countered during one time step due to friction, etc), and a pointer to the
 * next mass in the list.  Each point also includes an area where forces on
 * the mass point can be accumulated by the objects (springs) acting on this
 * point.  This allows an arbitrary number of springs to be attatched to one
 * point and they will have their forces summed appropriately.
 *	The ordering of the masses in the list is not important, as any
 * spring attached to a mass point will have a pointer that will allow it
 * to reference the point.  Note that once the list is created and springs
 * have been attached, no mass point can be deleted or moved if it has a
 * spring pointing to it, or the spring will act incorrectly.
 */

typedef	struct	MASS {	/* Used to allow a pointer to this type in struct */
          double mass;		/* The mass of this point */
          double radius;	/* The radius of the mass for collision purposes */
          double x,y,z;		/* The coordinates in space of this point */
          double vx,vy,vz;	/* The velocity of this point */
          double fx,fy,fz;	/* The forces acting on this point */
          double damping;	/* The damping of movement of this point */
          int colorNumber;      // The color Number associated with this entity for rendering
          size_t   sofaID;        /* SOFA index of the particle associated with this mass */
          struct MASS *next;	/* The next mass in the list */
        } MASS_node;

/*	This struct defines a node in the list of springs.  A spring goes
 * between two masses and applies forces to them based on the distance and
 * the spring constant of the spring.  If the distance between the masses
 * is equal to the rest length of the spring, then there is no force generated
 * by the spring.  If the distance is too short, then a force that tends to
 * separate the masses is generated.  If the distance is too long, then a
 * force pulling them together will be generated.
 *	The order of springs in the list does not matter.  The last node has
 * a null next pointer.
 */

typedef struct	SPRING { /* Used to allow a pointer to this type in struct */
          double rest;		/* The rest length of the spring */
          double k;		/* The spring constant of the spring */
          double fbreak;	/* The force which will break the spring */
          MASS_node *m1;	/* One mass attatched to the spring */
          MASS_node *m2;	/* The other mass attatched to the spring */
          int colorNumber;      // The color Number associated with this entity for rendering
          struct SPRING	*next;	/* The next spring in the list */
        } SPRING_node;

/*	This structure defines a hinge node.  A hinge relates three masses
 * and tends to keep the second on a line between the first and third.  A
 * hinge will apply torque to the arms between the masses to make this so.
 * The hinge constant k, multiplied by the angle between the arms, determines
 * the strength of this torque.  The torque is applied both to the center
 * node and to the end nodes, being converted to force by the arm lengths.
 *	The order of hinges in the hinge list does not matter.  The last node
 * has a null next pointer.
 */

typedef	struct	HINGE {	/* Used to allow a pointer to this type in struct */
         double	k;		/* The hinge constant of this hinge */
         MASS_node *m1,*m2,*m3;	/* The masses this hinge is connected to */
         int colorNumber;      // The color Number associated with this entity for rendering
         struct HINGE *next;	/* The next hinge in the list */
        } HINGE_node;

/*	The return types of all functions in massmesh.c that are made to be
 * used by other routines can be found here.  They return -1 on failure.
 */

extern	int	add_mass(MASS_node **headptr,
                        double mass,
                        double x, double y, double z,
                        int color,
                        double damping,
                        double radius = 0.0);

extern	int	add_spring(SPRING_node **headptr,
                        double rest,
                        double k,
                        MASS_node *m1, MASS_node *m2,
                        double fbreak = 1e100,
                        int color = 1);

extern	int	add_hinge(HINGE_node **headptr,
                        double k,
                        MASS_node *m1, MASS_node *m2, MASS_node *m3,
                        int color = 1);

// Reads in a previous outfile f for the coordinates of the last time step
// and puts their coordinates into the mass nodes mh.
// returns the value of the last time step, or a negative value (-1.0) if
// an error occurred.
// Also returns the number of steps already taken in the original file.

extern double parse_continue( FILE *f, MASS_node **mh, int *num_steps );

// Read in a description of a single structure from a file; this includes
// reading in named masses, springs to connect masses, and hinges to connect
// triples of masses.  The format of the data in the file is:
//    meta PARAMETER VALUE  (zero or more metadata parameters)
//    structure {
//      mass_damping DAMPING  {default 5}
//      random_force FORCE {default 5}
//      time_step STEP {Must be specified}
//      collision_spring_constant K {Default 2}
//      spring_damping_factor F {Default 0}
//      random_number_seed S {Default 42}
//      mass_radius RADIUS {default 0}
//      color INDEX {default 1}  (Associated with all following objects until another color)
//      mass NAME MASS X Y Z                  (or)
//      mass NAME MASS X Y Z COLOR_INDEX DAMPING          (or)
//      mass NAME MASS X Y Z COLOR_INDEX DAMPING RADIUS
//      (however many masses, each with a unique name
//      spring_constant_over_length VALUE   {Defaults to 1.0}
//      rest_length_fraction VALUE   {Defaults to 1.0, 0.5 means extended to twice rest length}
//      spring NAME1 NAME2  {Computes based on spring_constant_over_length and rest_length_fraction and mass distance} (or)
//      spring NAME1 NAME2 REST K                 (or)
//      spring NAME1 NAME2 REST K BREAKING_FORCE
//      (however many springs)
//      hinge NAME1 NAME2 NAME3 K
//      (however many hinges)
//    }
// Returns true if the parsing went okay, leaving the file pointer at
// the beginning of the line following the closing brace.  Returns pointers
// to the described structure's masses, springs, and hinges in the handles,
// NULL for each if there were no entries.
extern  bool    parse_structure_from_file(
                    FILE *f                     //< File to read the structure from
                    , MASS_node **mh            //< Pointer to an array of mass nodes (returned)
                    , SPRING_node **sh          //< Pointer to an array of spring nodes (returned)
                    , HINGE_node **hh           //< Pointer to an array of hinge nodes (returned)
                    , double *randomForce       //< Random (Brownian) force to apply (returned)
                    , double *timeStep          //< Time step to use in the simulation (returned)
                    , double *collisionSpringConstant   //< Spring constant for collisions (returned)
                    , double *springDampingFactor //< Ratio of drag constant from relative velocity of springs to spring constants
                    , int *seed                 //< Random-number seed to use for first simulation run
                    , std::vector<std::string>  *metaData   //< Lines describing metadata for the simulation
                );

//-----------------------------------------------------------------------
// Functions to move structures.

extern	void  translate_masses(MASS_node *mlist, double x, double y, double z);
