// sumForces.cl
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II


#include <chromoTypes.cl>

// add forces to the sum forces buffer
__kernel void addForce(
    __global CS_REAL3 * sumForce, // the current sum of the force
    __global CS_REAL3 * force, // the force to add
    const ulong columns, // the number of columns in force
    const ulong numParticles // the number of particles
    )
{
    ulong ndx = get_global_id(0);

    if (ndx >= numParticles)
    {
        return;
    }

    CS_REAL3 f = force[ndx];
    for (ulong i = 1; i < columns; ++i)
    {
        f = f + force[ndx + i*numParticles];
    }

    sumForce[ndx] = sumForce[ndx] + f;
}

// add Jacobians to the sum Jacobian buffer
__kernel void addJacobian(
    __global CS_REAL3 * sumJPosX, // result Jacobian wrt position of result
    __global CS_REAL3 * sumJPosY,
    __global CS_REAL3 * sumJPosZ,
    __global CS_REAL3 * sumJVelX, // result Jacobian wrt velocity of result
    __global CS_REAL3 * sumJVelY,
    __global CS_REAL3 * sumJVelZ,
    __global ulong * sumInteractionIndex, // particle index something in summand goes to
    __global ulong * sumNumInteractions, // the number of unique particles we've had interactions with
    __global CS_REAL3 * addJPosX, // result Jacobian wrt position of result
    __global CS_REAL3 * addJPosY,
    __global CS_REAL3 * addJPosZ,
    __global CS_REAL3 * addJVelX, // result Jacobian wrt velocity of result
    __global CS_REAL3 * addJVelY,
    __global CS_REAL3 * addJVelZ,
    __global ulong * addInteractionIndex, // particle index something in added goes to
    __global ulong * addNumInteractions, // the number of unique particles we have interactions with
    const ulong numParticles
    )
{
    ulong ndx = get_global_id(0);

    if (ndx >= numParticles)
    {
        return;
    }

    ulong addedCt = 0; // how many new interactions have we added
    ulong overLastSumNdx = sumNumInteractions[ndx] * numParticles; // over the index of the last Jacobian/particle interaction we put in for this particle
    // loop over the interactions that we are adding
    ulong overLastAddNdx = addNumInteractions[ndx] * numParticles; // none of the Jacobians we want to use are above this index

    for (ulong add_ndx = ndx; add_ndx < overLastAddNdx; add_ndx += numParticles)
    {
        ulong particle = addInteractionIndex[add_ndx]; // this is the index of the block that we are adding
        // loop over the interactions we already have to see if we should add to an already "existing" block
        ulong sum_ndx; // we declare the iterating integer beforehand so we can use whatever it ends on
        for (sum_ndx = ndx; sum_ndx < overLastSumNdx; sum_ndx += numParticles) // loop over existing blocks
        {
            if (particle == sumInteractionIndex[sum_ndx])
            {
                // we have seen this particle before, and our address to it is sum_ndx
                break;
            }
        }
        // so either we broke out of the loop where the index already exists, or we got to the next one, which is where we should add our Jacobian...
        sumJPosX[sum_ndx] += addJPosX[add_ndx];
        sumJPosY[sum_ndx] += addJPosY[add_ndx];
        sumJPosZ[sum_ndx] += addJPosZ[add_ndx];
        sumJVelX[sum_ndx] += addJVelX[add_ndx];
        sumJVelY[sum_ndx] += addJVelY[add_ndx];
        sumJVelZ[sum_ndx] += addJVelZ[add_ndx];

        // take note if this is a new one
        if (sum_ndx >= overLastSumNdx)
        {
            sumInteractionIndex[sum_ndx] = particle; // we added the particle
            overLastSumNdx += numParticles; // we need to increment this because we have one more particle
            ++addedCt;
        }
    }

    // we are done with everything except that we need to update sumNumInteractions...
    sumNumInteractions[ndx] += addedCt;
}

// kernel that sets the indices for the interactions that weren't present but so that
// they do not conflict with the existing ones in each row
__kernel void setIndexJacobian(
    __global ulong * sumInteractionIndex, // particle index something in summand goes to/interacts with
    __global ulong * sumNumInteractions, // the number of unique particles we've had interactions with
    const ulong maxInteractions, // the maximum number of interactions in the sum
    const ulong numParticles
    )
{
    ulong ndx = get_global_id(0);

    if (ndx >= numParticles)
    {
        return;
    }

    
    ulong maxIndex = numParticles * maxInteractions;
    ulong interactedIndex = numParticles * sumNumInteractions[ndx]; // these are for accessing the matrix structure

    ulong cur_ndx = interactedIndex + ndx;
    if (cur_ndx >= maxIndex) // this row is full.
    {
        return;
    }
    // otherwise fill up the rest
    for (ulong tryIndex = 0; tryIndex < numParticles; ++tryIndex)
    {
        bool alreadyInteracting = false; // is particle tryIndex already interacting with particle ndx?
        for (ulong loop_ndx = ndx; loop_ndx < interactedIndex; loop_ndx += numParticles)
        {
            if (tryIndex == sumInteractionIndex[loop_ndx])
            {
                alreadyInteracting = true; // if we are interacting, then we can skip ahead to go to next try index
                break;
            }
        }
        // are we are not already interacting with tryIndex? otherwise just go on to next try
        if (!alreadyInteracting)
        {
            // we add this to the cur_ndx and increment cur_ndx. if over maxIndex we can quit because we are done
            sumInteractionIndex[cur_ndx] = tryIndex;
            cur_ndx += numParticles;
            if (cur_ndx >= maxIndex)
            {
                return;
            }
        }
    }
}