// sumForces.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "sumForces.h"

// name of the force
const char * sumForces::getName( )
{
    return "Total Force";
}

// destructor
sumForces::~sumForces( )
{
    clReleaseMemObject( zeroForce );
    if (jacobianMaxInteractions > 0)
    {
        clReleaseMemObject( zeroJacobianRow );
        clReleaseMemObject( zeroJacobianInteractions );
        clReleaseKernel( kernelFixIndexJacobian );
    }

    // unload the kernels in addForce and addJacobian
    for (int i = 0; i < addForce.size(); ++i)
    {
        clReleaseKernel( addForce[i] );
    }
    for (int i = 0; i < addJacobian.size(); ++i)
    {
        clReleaseKernel( addJacobian[i] );
    }
}

// constructor
sumForces::sumForces( 
    std::string sim_CLdir, // the path of directory with our openCL source code
    cl_context sim_context, // OpenCL context passed in from simulation
    cl_command_queue sim_queue, // OpenCL queue passed in from simulation
    cl_ulong sim_numParticles, // number of particles in simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    cl_device_id sim_deviceId, // the deviceId we are running on
    const std::vector<genericForce*> & sim_forces // vector of the forces to add up
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, 
        sim_totalMaxInteractions, true )
{
    // copy the vector of force class pointers
    theForces = sim_forces;

    // set the "columns" of this force (1)
    columns = 1;
    // set the maximum number of interactions for this Jacobian
    // jacobianMaxInteractions = sim_totalMaxInteractions; // this is the place where they all will go
    jacobianMaxInteractions = 0;
    for (int i = 0; i < theForces.size(); ++i)
    {
        cl_ulong force_i_max = theForces[i]->getJacobianMaxInteractions( );
        if (force_i_max > jacobianMaxInteractions)
        {
            jacobianMaxInteractions = force_i_max; // set the maximum number of interactions to the max of added forces
        }
    }

    // zero out our starting force and Jacobian
    initForcesZero( );
    initJacobianZero( );
    // save copy of relevant stuff for zeroing out each time later
    cl_int error;
    zeroForce = clCreateBuffer(
        context, CL_MEM_READ_ONLY,
        numParticles * sizeof(CS_REAL3),
        NULL, &error );
    checkErr( error, "clCreateBuffer (zeroForce)" );
    CS_REAL3 * zf = new CS_REAL3[numParticles];
    for (int i = 0; i < numParticles; ++i)
    {
        zf[i].s[0] = 0;
        zf[i].s[1] = 0;
        zf[i].s[2] = 0;
    }
    checkErr(
        clEnqueueWriteBuffer( queue, zeroForce,
            CL_TRUE, 0, numParticles * sizeof(CS_REAL3), zf,
            0, NULL, NULL
            ),
        "clEnqueueWriteBuffer (zeroForce)"
        );
    delete[] zf;
    if (jacobianMaxInteractions > 0)
    {
        zeroJacobianRow = clCreateBuffer(
            context, CL_MEM_READ_ONLY,
            jacobianMaxInteractions * numParticles * sizeof(CS_REAL3),
            NULL, &error );
        checkErr( error, "clCreateBuffer (zeroJacobianRow)" );
        zf = new CS_REAL3[jacobianMaxInteractions * numParticles];
        for (int i = 0; i < numParticles * jacobianMaxInteractions; ++i)
        {
            zf[i].s[0] = 0;
            zf[i].s[1] = 0;
            zf[i].s[2] = 0;
        }
        checkErr(
            clEnqueueWriteBuffer( queue, zeroJacobianRow,
                CL_TRUE, 0, jacobianMaxInteractions *numParticles * sizeof(CS_REAL3), 
                zf,
                0, NULL, NULL
                ),
            "clEnqueueWriteBuffer (zeroJacobianRow)"
            );
        delete[] zf;
        zeroJacobianInteractions = clCreateBuffer(
            context, CL_MEM_READ_ONLY,
            numParticles * sizeof(cl_ulong),
            NULL, &error );
        checkErr( error, "clCreateBuffer (zeroJacobianInteractions)" );
        cl_ulong * zi = new cl_ulong[numParticles];
        for (int i = 0; i < numParticles; ++i)
        {
            zi[i] = 0;
        }
        checkErr(
            clEnqueueWriteBuffer( queue, zeroJacobianInteractions,
                CL_TRUE, 0, numParticles * sizeof( cl_ulong ), zi,
                0, NULL, NULL
                ),
            "clEnqueueWriteBuffer (zeroJacobianInteractions)"
            );
    }

    // build program/kernels
    
    // load the program "sumForces.cl"
    loadProgram( sim_deviceId, "sumForces.cl" );

    // set up the kernels:
    // the vectors addForce and addJacobian should be currently empty

    // loop through the forces
    for (int i = 0; i < theForces.size(); ++i)
    {
        // set up add forces
        cl_ulong addColumns = theForces[i]->getColumns();
        cl_mem addFmem = theForces[i]->getCLForce();
        cl_kernel kernelForceAdd = clCreateKernel( program, "addForce", &error );
        checkErr( error, "clCreateKernel addForce" );

        // pass arguments into kernel
        error = clSetKernelArg( kernelForceAdd, 0, sizeof(cl_mem), &clForce );
        error |= clSetKernelArg( kernelForceAdd, 1, sizeof(cl_mem), &addFmem );
        error |= clSetKernelArg( kernelForceAdd, 2, sizeof(cl_ulong), &addColumns );
        error |= clSetKernelArg( kernelForceAdd, 3, sizeof(cl_ulong), &numParticles );
        checkErr( error, "clSetKernelArg addForce" );

        // get the global size we should have
        size_t globalSizeForce = getKernelPaddedGlobalSize(
            sim_deviceId, kernelForceAdd, numParticles );

        // push into our vector of add force kernels
        addForce.push_back( kernelForceAdd );
        addForceGlobalSize.push_back( globalSizeForce );

        // set up add Jacobians
        cl_ulong addMaxInteractions = theForces[i]->getJacobianMaxInteractions();
        if (addMaxInteractions > 0)
        {
            // only if we have a nonzero Jacobian
            forceJacobian addJmem = theForces[i]->getCLJacobian();

            cl_kernel kernelJacobianAdd = clCreateKernel( program, "addJacobian", &error );
            checkErr( error, "clCreateKernel addJacobian" );

            // pass arguments into kernel
            error = clSetKernelArg( kernelJacobianAdd, 0, sizeof(cl_mem), &clJacobian.j_PosX );
            error |= clSetKernelArg( kernelJacobianAdd, 1, sizeof(cl_mem), &clJacobian.j_PosY );
            error |= clSetKernelArg( kernelJacobianAdd, 2, sizeof(cl_mem), &clJacobian.j_PosZ );
            error |= clSetKernelArg( kernelJacobianAdd, 3, sizeof(cl_mem), &clJacobian.j_VelX );
            error |= clSetKernelArg( kernelJacobianAdd, 4, sizeof(cl_mem), &clJacobian.j_VelY );
            error |= clSetKernelArg( kernelJacobianAdd, 5, sizeof(cl_mem), &clJacobian.j_VelZ );
            error |= clSetKernelArg( kernelJacobianAdd, 6, sizeof(cl_mem), &clJacobian.interactionParticleIndex );
            error |= clSetKernelArg( kernelJacobianAdd, 7, sizeof(cl_mem), &clJacobian.numInteractions );
            error |= clSetKernelArg( kernelJacobianAdd, 8, sizeof(cl_mem), &addJmem.j_PosX );
            error |= clSetKernelArg( kernelJacobianAdd, 9, sizeof(cl_mem), &addJmem.j_PosY );
            error |= clSetKernelArg( kernelJacobianAdd, 10, sizeof(cl_mem), &addJmem.j_PosZ );
            error |= clSetKernelArg( kernelJacobianAdd, 11, sizeof(cl_mem), &addJmem.j_VelX );
            error |= clSetKernelArg( kernelJacobianAdd, 12, sizeof(cl_mem), &addJmem.j_VelY );
            error |= clSetKernelArg( kernelJacobianAdd, 13, sizeof(cl_mem), &addJmem.j_VelZ );
            error |= clSetKernelArg( kernelJacobianAdd, 14, sizeof(cl_mem), &addJmem.interactionParticleIndex );
            error |= clSetKernelArg( kernelJacobianAdd, 15, sizeof(cl_mem), &addJmem.numInteractions );
            error |= clSetKernelArg( kernelJacobianAdd, 16, sizeof(cl_ulong), &numParticles );
            checkErr( error, "clSetKernelArg addJacobian" );

            // get global size
            size_t globalSizeJacobian = getKernelPaddedGlobalSize(
                sim_deviceId, kernelJacobianAdd, numParticles );

            // push back into our vector of add Jacobian kernels
            addJacobian.push_back( kernelJacobianAdd );
            addJacobianGlobalSize.push_back( globalSizeJacobian );
        }
    }

    // if we do have a Jacobian, set up kernel to fix indices for no interactions
    if (jacobianMaxInteractions > 0)
    {
        kernelFixIndexJacobian = clCreateKernel( program, "setIndexJacobian", &error );
        checkErr( error, "clCreateKernel fixIndex" );

        // pass arguments into kernel
        error = clSetKernelArg( kernelFixIndexJacobian, 0, sizeof(cl_mem), &clJacobian.interactionParticleIndex );
        error |= clSetKernelArg( kernelFixIndexJacobian, 1, sizeof(cl_mem), &clJacobian.numInteractions );
        error |= clSetKernelArg( kernelFixIndexJacobian, 2, sizeof(cl_ulong), &jacobianMaxInteractions );
        error |= clSetKernelArg( kernelFixIndexJacobian, 3, sizeof(cl_ulong), &numParticles );
        checkErr( error, "clSetKernelArg fixIndex" );

        // get global size
        kernelFixIndexJacobianGlobalSize = getKernelPaddedGlobalSize(
            sim_deviceId, kernelFixIndexJacobian, numParticles );
    }
}

void sumForces::helperAddForces( )
{
    cl_int error;
    // zero out the force
    error = clEnqueueCopyBuffer( queue, zeroForce, clForce, 
        0, 0, numParticles * sizeof( CS_REAL3 ),
        0, NULL, NULL );
    // TODO event to wait?
    for (int i = 0; i < addForce.size(); ++i)
    {
        error |= clEnqueueNDRangeKernel(
            queue, // queue we are running on
            addForce[i], // the kernel we are running on
            1, // 1d work space
            NULL, // no offsets
            &addForceGlobalSize[i], // padded global work space size
            NULL, // local items chosen by implementation
            0, NULL, NULL // no waiting for events
            );
            checkErr( clFlush( queue ), "clFlush (after addForce)" );
            checkErr( clFinish( queue ), "clFinish (after addForce)" );
    }

    checkErr( error, "calculateForce" );

    updated = false;
}

void sumForces::calculateForce ( )
{
    for (int i = 0; i < theForces.size(); ++i)
    {
        theForces[i]->calculateForce();
        checkErr( clFlush( queue ), "clFlush (after calculateForce)" );
        checkErr( clFinish( queue ), "clFinish (after calculateForce)" );
    }

    helperAddForces( );
}

void sumForces::calculateForceAndJacobian ( )
{
    for (int i = 0; i < theForces.size(); ++i)
    {
        theForces[i]->calculateForceAndJacobian();
        checkErr( clFlush( queue ), "clFlush (after calculateForceAndJacobian)" );
        checkErr( clFinish( queue ), "clFinish (after calculateForceAndJacobian)" );
    }
    // TODO wait for the force calculations to complete?
    helperAddForces( ); // we still have to add the forces
    
    if (jacobianMaxInteractions > 0)
    {
        cl_int error;
        // zero out the Jacobian
        error = clEnqueueCopyBuffer( queue, zeroJacobianRow, 
            clJacobian.j_PosX, 0, 0, 
            jacobianMaxInteractions * numParticles * sizeof( CS_REAL3 ),
            0, NULL, NULL );
        error |= clEnqueueCopyBuffer( queue, zeroJacobianRow, 
            clJacobian.j_PosY, 0, 0, 
            jacobianMaxInteractions * numParticles * sizeof( CS_REAL3 ),
            0, NULL, NULL );
        error |= clEnqueueCopyBuffer( queue, zeroJacobianRow, 
            clJacobian.j_PosZ, 0, 0, 
            jacobianMaxInteractions * numParticles * sizeof( CS_REAL3 ),
            0, NULL, NULL );
        error |= clEnqueueCopyBuffer( queue, zeroJacobianRow, 
            clJacobian.j_VelX, 0, 0, 
            jacobianMaxInteractions * numParticles * sizeof( CS_REAL3 ),
            0, NULL, NULL );
        error |= clEnqueueCopyBuffer( queue, zeroJacobianRow, 
            clJacobian.j_VelY, 0, 0, 
            jacobianMaxInteractions * numParticles * sizeof( CS_REAL3 ),
            0, NULL, NULL );
        error |= clEnqueueCopyBuffer( queue, zeroJacobianRow, 
            clJacobian.j_VelZ, 0, 0, 
            jacobianMaxInteractions * numParticles * sizeof( CS_REAL3 ),
            0, NULL, NULL );
        // zero out the number of interactions for each particle
        error |= clEnqueueCopyBuffer( queue, zeroJacobianInteractions,
            clJacobian.numInteractions, 0, 0, 
            numParticles * sizeof(cl_ulong),
            0, NULL, NULL );
        // TODO wait for each event to finish?
        for (int i = 0; i < addJacobian.size(); ++i)
        {
            error |= clEnqueueNDRangeKernel(
                queue, // queue we are running on
                addJacobian[i], // the kernel we are running on
                1, // 1d work space
                NULL, // no offsets
                &addJacobianGlobalSize[i], // padded global work space size
                NULL, // local items chosen by implementation
                0, NULL, NULL // no waiting for events
                );
            // TODO event to wait?
            //checkErr( clFlush( queue ), "clFlush (after addJacobian)" );
            //checkErr( clFinish( queue ), "clFinish (after addJacobian)" );
        }

        // check if any of those adds yielded an error
        checkErr( error, "calculateJacobian" );

        // now fix the remaining indices that we didn't get to to
        // ensure that we don't have any repeated particle interactions
        // for ease transferring to sparse matrix classes
        error = clEnqueueNDRangeKernel(
            queue, kernelFixIndexJacobian, 1, NULL,
            &kernelFixIndexJacobianGlobalSize, NULL,
            0, NULL, NULL );
        checkErr( error, "clEnqueueNDRangeKernel (fixIndex)" );
    }
}