// By Russell M Taylor II, Fall 2014

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sstream>
#include "structureLoader.h"
#include "chromoSim.h"
#include "unitConversion.h"
#include "analysis.h"

// Globals
bool g_verbose = false;

void Usage(const char *s)
{
    fprintf( stderr, "Usage: %s [-verbose] [-CPU] [-num_masses N] [-mass_separation_microns S] [-viscosity_centiPoise] [-temperature_Celsius T] [-temperature_Kelvin T] [-timestep_ns T] [-burst N] [-stride N] [-repeat N] [-openCL_dir directory] [-seed S] [-check_range]\n", s );
    fprintf( stderr, "       -CPU: Ask OpenCL to use the CPU instead of a GPU\n" );
    fprintf( stderr, "       -num_masses: Use this many masses (default 5)\n" );
    fprintf( stderr, "       -mass_separation_microns: Use this mass separation (default 0.01)\n" );
    fprintf( stderr, "       -viscosity_centiPoise: Use this as the viscosity (default 1)\n" );
    fprintf( stderr, "       -temperature_Celsius: Use this as the temperature (default 25)\n" );
    fprintf( stderr, "       -temperature_Kelvin: Use this as the temperature\n" );
    fprintf( stderr, "       -timestep_ns: Use this as the basic time step (default 9)\n" );
    fprintf( stderr, "       -burst: Run a burst of this many timesteps between reads (default 10)\n" );
    fprintf( stderr, "       -stride: When computing MSD, use bursts that are this many apart (default 1)\n" );
    fprintf( stderr, "       -repeat: Run this many bursts total (default 100)\n" );
    fprintf( stderr, "       -openCL_dir: Open the specified directory to load the openCL kernels (default is ./)\n" );
    fprintf( stderr, "       -seed: Use this random-number seed (default 42)\n" );
    fprintf( stderr, "       -check_range: Check a range of parameters around the selected and print csv.\n" );
    exit(-1);
}

int test_configuration(
	int number_of_masses
	, double node_separation_meters
	, double mass_radius
	, double mass_kg
	, double mass_damping
	, double time_step_seconds
	, double random_force
	, double temperature_Kelvin
	, double viscosity_Pascal_seconds
	, double spring_damping_factor
	, int seed
	, const char *openCL_file_dir
	, int tauIterations
	, int readIterations
	, int stride
	, bool useGPU
)
{
    double startTime = 0.0; // start time of the simulation
    int collisionScheme = -1;    // Which collision-detection approach to use?
    double  collision_spring_constant = 0.0;    // No collision response

    int i;

    // Initialize structures for simulation.  We add ten masses, which start at the
    // origin and at distances separated by the node_separation_meters.
    // They must have consistent damping parameters because random force is the
    // same for all of them.
    MASS_node * mlist = NULL; // points to first node in mass list
    SPRING_node * slist = NULL; // points to first node in spring list
    HINGE_node * hlist = NULL; // points to first hinge in the hinge list
    for (i = 1; i <= number_of_masses; i++) {
        add_mass( &mlist, mass_kg,
            i * node_separation_meters, 0, 0,
            1,
            mass_damping,
            mass_radius);
    }

    // Get a simulation to use.
    chromoSim *simulation = new chromoSim( mlist, slist, hlist, // structures from configuration files
              static_cast<CS_REAL>(random_force), 
              static_cast<CS_REAL>(spring_damping_factor), // the default parameter in SOFA
              static_cast<CS_REAL>(collision_spring_constant), // the spring constant for collisions
              static_cast<CS_REAL>(startTime), // what is the starting time of our simuation?
              static_cast<CS_REAL>(time_step_seconds), // what is the time step of our simulation
              collisionScheme, // we haven't implemented cylindrical collisions yet
              useGPU, // are we using the GPU or CPU?
              seed, // use the random seed
              openCL_file_dir);

    // If we couldn't get a simulation, then try again but this time asking for a CPU-based
    // one (default was GPU-based).
    if (useGPU && (!simulation || !simulation->doing_okay()) ) {
	cerr << "Could not open GPU simulator, trying to obtain a CPU-based one" << endl;
	useGPU = false;
    	simulation = new chromoSim( mlist, slist, hlist, // structures from configuration files
              static_cast<CS_REAL>(random_force), 
              static_cast<CS_REAL>(spring_damping_factor), // the default parameter in SOFA
              static_cast<CS_REAL>(collision_spring_constant), // the spring constant for collisions
              static_cast<CS_REAL>(startTime), // what is the starting time of our simuation?
              static_cast<CS_REAL>(time_step_seconds), // what is the time step of our simulation
              collisionScheme, // we haven't implemented cylindrical collisions yet
              useGPU, // are we using the GPU or CPU?
              seed, // use the random seed
              openCL_file_dir);
    }
    if (!simulation || !simulation->doing_okay()) {
	cerr << "Could not open simulator." << endl;
	return 1;
    }
    
    // Structures to record the state of all the masses returned at one
    // simulation time step.
    std::vector<CS_REAL3> pos, vel;

    // Structures to record the trajectories of individual masses.
    // Make one empty vector for each mass.
    std::vector< std::vector<CS_REAL3> >  trajectories;
    std::vector<CS_REAL3> empty;
    for (i = 0; i < number_of_masses; i++) {
        trajectories.push_back(empty);
    }

    // Find the initial positions of the masses and store them.
    simulation->readState( pos, vel );
    for (i = 0; i < number_of_masses; i++) {
        trajectories[i].push_back(pos[i]);
    }

    // Take a number of iterations, each of which runs a specified
    // number of simulation steps.  Read the positions of the masses
    // after each of them.
    for ( int t = 0; t < readIterations; t++ ) {
        simulation->simulateTimeSteps( tauIterations );
        simulation->readState( pos, vel );
        for (i = 0; i < number_of_masses; i++) {
            trajectories[i].push_back(pos[i]);
        }
    }

    // Compute the MSD of each of the beads.  Print it and the expected
    // result for that bead.
    if (g_verbose) for (i = 0; i < number_of_masses; i++) {
        printf("Bead %d MSD: %lg", static_cast<int>(i),
            compute_MSD(trajectories[i], stride) );
        printf(" (expected %lg)\n",
            expected_MSD(temperature_Kelvin,
                        tauIterations * time_step_seconds * stride,
                        viscosity_Pascal_seconds,
                        mass_radius));
    }

    // See if we have more than 1/10th the number of masses on each side of
    // what we expected.  Since this is a random process, we can't be sure
    // of exactly how many will be on each side, but it is very unlikely that
    // less than 10% will be on the wrong side.
    int above = 0, below = 0;
    double expected = expected_MSD(temperature_Kelvin,
                        tauIterations * time_step_seconds * stride,
                        viscosity_Pascal_seconds,
                        mass_radius);
    for (i = 0; i < number_of_masses; i++) {
	if (compute_MSD(trajectories[i], stride) < expected) {
		below++;
	} else {
		above++;
	}
    }
    int threshold = number_of_masses / 10;
    if ( (below <= threshold) || (above <= threshold) ) {
	if (g_verbose) fprintf(stderr,"Not enough results on both sides of expected value\n");
	return 2;
    }

    delete simulation;
    return 0;
}

int main(int argc, const char *argv[])
{
    // Program version number.  Major, minor, bugfix
    const char *version_string = "01.00.00";

    // variables changed by command line argument
    bool useGPU = true; // Try to use a GPU for the calculations?
    bool check_range = false; // Try a range of values and print what worked?
    int seed = 42; // magic number for seed

    // Parameters controlling time step, iterations, and stride when computing
    double  time_step_nanoseconds = 9;
    int tauIterations = 10; // the number of steps that we take each time before reading
    int readIterations = 100; // the total number of times we iterate tauIterations times
    int stride = 1;  // How many bursts to skip when computing MSD

    int number_of_masses = 100;
    double  mass_separation_microns = 0.01;
    double  temperature_Celsius = 25;
    double  viscosity_centiPoise = 1;

    // Constants (could become command-line arguments).
    double  spring_damping_factor = 0.0;

    // Name of the openCL command file to load
    const char * openCL_file_dir = "./";

    // Command-line parsing
    int  i;
    unsigned  real_params = 0;  //< Number of non-flag parameters
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i],"-help")) {
            Usage(argv[0]);
        } else if (!strcmp(argv[i],"-verbose")) {
            g_verbose = true;
        } else if (!strcmp(argv[i],"-check_range")) {
            check_range = true;
        } else if (!strcmp(argv[i],"-CPU")) {
            useGPU = false;
        } else if (!strcmp(argv[i],"-num_masses")) {
            if (++i >= argc) { Usage(argv[0]); }
            number_of_masses = atoi(argv[i]);
        } else if (!strcmp(argv[i],"-mass_separation_microns")) {
            if (++i >= argc) { Usage(argv[0]); }
            mass_separation_microns = atof(argv[i]);
        } else if (!strcmp(argv[i],"-seed")) {
            if (++i >= argc) { Usage(argv[0]); }
            seed = atoi(argv[i]);
        } else if (!strcmp(argv[i],"-viscosity_centiPoise")) {
            if (++i >= argc) { Usage(argv[0]); }
            viscosity_centiPoise = atof(argv[i]);
        } else if (!strcmp(argv[i],"-temperature_Celsius")) {
            if (++i >= argc) { Usage(argv[0]); }
            temperature_Celsius = atof(argv[i]);
        } else if (!strcmp(argv[i],"-temperature_Kelvin")) {
            if (++i >= argc) { Usage(argv[0]); }
            temperature_Celsius = Kelvin_to_Celsius(atof(argv[i]));
        } else if (!strcmp(argv[i],"-timestep_ns")) {
            if (++i >= argc) { Usage(argv[0]); }
            time_step_nanoseconds = atof(argv[i]);
        } else if (!strcmp(argv[i],"-burst")) {
            if (++i >= argc) { Usage(argv[0]); }
            tauIterations = atoi(argv[i]);
        } else if (!strcmp(argv[i],"-stride")) {
            if (++i >= argc) { Usage(argv[0]); }
            stride = atoi(argv[i]);
        } else if (!strcmp(argv[i],"-repeat")) {
            if (++i >= argc) { Usage(argv[0]); }
           readIterations = atoi(argv[i]);
        } else if (!strcmp(argv[i],"-openCL_dir")) {
            if (++i >= argc) { Usage(argv[0]); }
            openCL_file_dir = argv[i];
        } else if (argv[i][0] == '-') {
            Usage(argv[0]);
        } else {
            switch (real_params) {
            case 1:
            case 2:
            default:
                Usage(argv[0]);
            }
        }
    }

    // Tell the version number.
    if (g_verbose) std::cout << "test_diffusion version " << version_string << endl;

    // Compute the derived values that we feed to the simulator.
    double  time_step_seconds = nanoseconds_to_seconds(time_step_nanoseconds);
    double  node_separation_meters = microns_to_meters(mass_separation_microns);
    double  mass_kg = dna_mass_in_kg(node_separation_meters);
    double  temperature_Kelvin = Celsius_to_Kelvin(temperature_Celsius);
    double  viscosity_Pascal_seconds = centiPoise_to_Pascal_seconds(viscosity_centiPoise);
    double  mass_radius = node_separation_meters;

    double  random_force = Brownian_force_equivalent(
                    mass_radius,
                    temperature_Kelvin,
                    viscosity_Pascal_seconds,
                    time_step_seconds
                    );
    double  mass_damping = mass_damping_equivalent(
		viscosity_Pascal_seconds,
		mass_radius,
		mass_kg);
    if (g_verbose) {
	printf("Mass = %lg\n", mass_kg);
	printf("Temperature = %lg K\n", temperature_Kelvin);
	printf("Viscosity = %lg Pascal-seconds\n", viscosity_Pascal_seconds);
	printf("Mass damping = %lg\n", mass_damping);
	printf("Random force = %lg\n", random_force);
	printf("Time step = %lg s\n", time_step_seconds);
    }

    if (check_range) {
	printf("Mass_sep_meters, Viscosity_Pas_sec, Time_step_sec, Temp_Kelvin, Success\n");
	double ms, vis, ts, temp;
	for (ms = node_separation_meters / 100; ms <= node_separation_meters * 100; ms *= 100)
	 for (vis = viscosity_Pascal_seconds / 100; vis <= viscosity_Pascal_seconds * 100; vis *= 100)
	  for (ts = time_step_seconds / 100; ts <= time_step_seconds * 100; ts *= 100)
	    for (temp = temperature_Kelvin - 100; temp <= temperature_Kelvin + 100; temp += 100) {
		bool result = (0 == test_configuration(
			number_of_masses, ms, ms, mass_kg, mass_damping
			, ts, random_force, temp
			, vis, spring_damping_factor,
			seed, openCL_file_dir,
			tauIterations, readIterations, stride, useGPU));
		printf("%lg, %lg, %lg, %lg, %s\n", ms, vis, ts, temp, result ? "Y" : "N");
	    }
    } else {
	bool ret = (0 == test_configuration(
		number_of_masses, node_separation_meters, mass_radius, mass_kg, mass_damping
		, time_step_seconds, random_force, temperature_Kelvin
		, viscosity_Pascal_seconds, spring_damping_factor,
		seed, openCL_file_dir,
		tauIterations, readIterations, stride, useGPU));
	if (ret) {
		printf("Success!\n");
		return 0;
	} else {
		printf("Failure.\n");
		return 3;
	}
    }
}

