// By Russell M Taylor II, Fall 2014

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sstream>
#include "structureLoader.h"
#include "chromoSim.h"
#include "unitConversion.h"
#include "analysis.h"

// Globals
bool g_verbose = false;

void Usage(const char *name)
{
	fprintf(stderr,"Usage: %s [-CPU] [-verbose] [-option value]\n", name);
        fprintf(stderr, "    -CPU: Ask OpenCL to use the CPU instead of a GPU\n" );
        fprintf(stderr, "    -verbose: Print a lot of debugging info\n" );
        fprintf(stderr, "    -timestep_ns: Use this as the basic time step (default 9)\n" );
        fprintf(stderr, "    -burst: Run a burst of this many timesteps between reads (default 10)\n" );
        fprintf(stderr, "    -repeat: Run this many bursts total (default 100)\n" );
	fprintf(stderr,"     -chain_length (microns) (default 1)\n");
	fprintf(stderr,"     -node_separation (microns) (default 0.01\n");
	fprintf(stderr,"     -temperature (Celsius) (default 25)\n");
	fprintf(stderr,"     -viscosity (centiPoise) (default 1.5e-5)\n");
        fprintf(stderr,"         (Water = 1, but we use this to scale time)\n");
        fprintf(stderr,"         (Smaller runs faster but can make the simulation unstable)\n");
	fprintf(stderr,"     -damping_radius_factor: Multiplied by node_separation to determine effective radius for damping (ratio) (1)\n");
	fprintf(stderr,"     -time_step (nanoseconds) (default 2)\n");
        fprintf(stderr,"         (Longer time steps run faster but can make the simulation unstable\n");
	fprintf(stderr,"     -modulus: Young's modulus of DNA (gigaPascal) (default 2)\n");
	fprintf(stderr,"     -dna_radius: Radius of B-form DNA (nanometers) (default 0.6)\n");
        fprintf(stderr,"         (The collision/visual radius will be larger than this)\n");
        fprintf(stderr,"     -collision_approach: Which collision-detection scheme to use (default -1, none)\n");
	fprintf(stderr,"     -collision_radius_factor: Fraction of distance collision spheres will cover (default 0.9)\n");
        fprintf(stderr,"     -collision_force_factor: Number of times the spring constant to use for collision response (default 0.25)\n");
        fprintf(stderr,"     -spring_damping_factor: Ratio of drag constant from relative velocity of springs to spring constants (default 0)\n");
        fprintf(stderr, "    -seed: Specify the seed value for the random number generator\n" );
        fprintf(stderr, "        (default is 42)\n" );
        fprintf(stderr, "    -openCL_dir: Open the specified directory to load the openCL kernels (default is ./)\n" );
	exit(-1);
}

int test_configuration(
	double chain_length_meters
	, double node_separation_meters
	, double mass_radius
	, double mass_kg
	, double mass_damping
        , double spring_constant
        , double hinge_constant
	, double time_step_seconds
	, double random_force
	, double temperature_Kelvin
	, double viscosity_Pascal_seconds
	, double spring_damping_factor
        , double modulus_gigaPascal
        , double dna_radius_meters
	, int seed
	, const char *openCL_file_dir
	, int tauIterations
	, int readIterations
	, bool useGPU
        , int collisionScheme
        , double collision_spring_constant
)
{
    double startTime = 0.0; // start time of the simulation

    int i;

    i = 0;
    double x;
    double epsilon = node_separation_meters / 10.0;

    // Initialize structures for simulation.
    MASS_node * mlist = NULL; // points to first node in mass list
    SPRING_node * slist = NULL; // points to first node in spring list
    HINGE_node * hlist = NULL; // points to first hinge in the hinge list
    for (x = 0.0; x < chain_length_meters+epsilon; x += node_separation_meters, i++) {
        add_mass( &mlist, mass_kg,
            x, 0, 0,
            1,
            mass_damping,
            mass_radius);
        // If we have at least two masses, add a spring between the most-recently-
        // added two masses at the head of the list.
	if (i >= 1) {
            add_spring( &slist, node_separation_meters, spring_constant,
                mlist, mlist->next );
        }
        // If we have at least three masses, add a hinge between the most-recently-
        // added three masses at the head of the list.
	if (i >= 2) {
            add_hinge( &hlist, hinge_constant, mlist, mlist->next,
                mlist->next->next);
        }
    }
    int number_of_masses = i;

    // Get a simulation to use.
    if (g_verbose) {
        printf("Constructing simulation\n");
    }
    chromoSim *simulation = new chromoSim( mlist, slist, hlist, // structures from configuration files
              static_cast<CS_REAL>(random_force), 
              static_cast<CS_REAL>(spring_damping_factor), // the default parameter in SOFA
              static_cast<CS_REAL>(collision_spring_constant), // the spring constant for collisions
              static_cast<CS_REAL>(startTime), // what is the starting time of our simuation?
              static_cast<CS_REAL>(time_step_seconds), // what is the time step of our simulation
              collisionScheme, // we haven't implemented cylindrical collisions yet
              useGPU, // are we using the GPU or CPU?
              seed, // use the random seed
              openCL_file_dir);

    // If we couldn't get a simulation, then try again but this time asking for a CPU-based
    // one (default was GPU-based).
    if (useGPU && (!simulation || !simulation->doing_okay()) ) {
	cerr << "Could not open GPU simulator, trying to obtain a CPU-based one" << endl;
	useGPU = false;
    	simulation = new chromoSim( mlist, slist, hlist, // structures from configuration files
              static_cast<CS_REAL>(random_force), 
              static_cast<CS_REAL>(spring_damping_factor), // the default parameter in SOFA
              static_cast<CS_REAL>(collision_spring_constant), // the spring constant for collisions
              static_cast<CS_REAL>(startTime), // what is the starting time of our simuation?
              static_cast<CS_REAL>(time_step_seconds), // what is the time step of our simulation
              collisionScheme, // we haven't implemented cylindrical collisions yet
              useGPU, // are we using the GPU or CPU?
              seed, // use the random seed
              openCL_file_dir);
    }
    if (!simulation || !simulation->doing_okay()) {
	cerr << "Could not open simulator." << endl;
	return 1;
    }
    
    // Structures to record the state of all the masses returned at one
    // simulation time step.
    std::vector<CS_REAL3> pos, vel;

    // Structure to record the trajectories of individual masses.
    std::vector< std::vector<CS_REAL3> >  trajectories;

    // Find the initial positions of the masses and store them.
    if (g_verbose) {
        printf("Storing initial positions\n");
    }
    simulation->readState( pos, vel );
    std::vector<CS_REAL3> empty;
    trajectories.push_back(empty);
    for (i = 0; i < number_of_masses; i++) {
        trajectories[0].push_back(pos[i]);
    }

    // Take a number of iterations, each of which runs a specified
    // number of simulation steps.  Read the positions of the masses
    // after each of them.
    for ( int t = 0; t < readIterations; t++ ) {
        trajectories.push_back(empty);
        if (g_verbose) {
            printf("Starting step %d out of %d (%lg percent)\n",
                t, readIterations, static_cast<double>(t)/readIterations*100.0);
        }
        simulation->simulateTimeSteps( tauIterations );
        simulation->readState( pos, vel );
        for (i = 0; i < number_of_masses; i++) {
            trajectories[t+1].push_back(pos[i]);
        }
    }

    // Compute the expected radius of gyration given the parameters we used.
    // This needs to change if we have collision detection turned on.
    double expected = expected_RoG( expected_persistence_length(temperature_Kelvin,
                        gigaPascal_to_Pascal(modulus_gigaPascal), dna_radius_meters),
                        chain_length_meters);
    if (collisionScheme >= 0) {
        if (g_verbose) {
            printf("Extending RoG because of collisions\n");
        }
        expected *= Rog_collision_multiplier(number_of_masses-1);
    }

    // Compute the radius of gyration for the last few times in the trajectories.
    // Print them and the expected.
    int start = readIterations - 10;
    if (start < 0) { start = 0; }
    if (g_verbose) {
        printf("Expected RoG: %lg\n", expected);
        for (i = start; i < readIterations; i++) {
            printf(" RoG: %lg\n", compute_RoG(trajectories[i]) );
        }
    }

    // See if we have more than 1/10th the number of RoG on each side of
    // what we expected.  Since this is a random process, we can't be sure
    // of exactly how many will be on each side, but it is very unlikely that
    // less than 10% will be on the wrong side.  We only check the ones
    // that are more than halfway through the simulation process, because
    // we need to wait until the chain settles down from its straight
    // configuration.
    int above = 0, below = 0;
    for (i = readIterations/2 + 1; i < readIterations; i++) {
	if (compute_RoG(trajectories[i]) < expected) {
		below++;
	} else {
		above++;
	}
    }
    int threshold = (readIterations/2) / 10;
    if ( (below <= threshold) || (above <= threshold) ) {
	if (g_verbose) fprintf(stderr,"Not enough results on both sides of expected value\n");
	return 2;
    }

    delete simulation;
    return 0;
}

int main (int argc, char *argv[])
{
	// Constants (could become command-line arguments).
        bool useGPU = true;
        double  time_step_nanoseconds = 2;
        int tauIterations = 10000; // the number of steps that we take each time before reading
        int readIterations = 6000; // the total number of times we iterate tauIterations times
        const char * openCL_file_dir = "./";
	double	chain_length_microns = 1;
	double	mass_separation_microns = 0.01;
        double  temperature_Celsius = 25;
	double  viscosity_centiPoise = 1;
        double  modulus_gigaPascal = 2;
        double  dna_radius_nanometers = 0.6;
        int collisionScheme = -1;    // Which collision-detection approach to use?
        double  collision_radius_factor = 0.9;
        double  collision_force_factor = 0.25;
        double  spring_damping_factor = 0.0;
        double  damping_radius_factor = 1.0;
        int random_number_seed = 42;

	// Parse the command line to adjust default values.
	int i;
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-chain_length")) {
			if (++i >= argc) { Usage(argv[0]); }
			chain_length_microns = atof(argv[i]);
		} else if (!strcmp(argv[i], "-node_separation")) {
			if (++i >= argc) { Usage(argv[0]); }
			mass_separation_microns = atof(argv[i]);
		} else if (!strcmp(argv[i], "-temperature")) {
			if (++i >= argc) { Usage(argv[0]); }
			temperature_Celsius = atof(argv[i]);
		} else if (!strcmp(argv[i], "-viscosity")) {
			if (++i >= argc) { Usage(argv[0]); }
			viscosity_centiPoise = atof(argv[i]);
		} else if (!strcmp(argv[i], "-damping_radius_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			damping_radius_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-time_step")) {
			if (++i >= argc) { Usage(argv[0]); }
			time_step_nanoseconds = atof(argv[i]);
		} else if (!strcmp(argv[i], "-modulus")) {
			if (++i >= argc) { Usage(argv[0]); }
			modulus_gigaPascal = atof(argv[i]);
		} else if (!strcmp(argv[i], "-dna_radius")) {
			if (++i >= argc) { Usage(argv[0]); }
			dna_radius_nanometers = atof(argv[i]);
		} else if (!strcmp(argv[i], "-collision_approach")) {
			if (++i >= argc) { Usage(argv[0]); }
			collisionScheme = atoi(argv[i]);
		} else if (!strcmp(argv[i], "-collision_radius_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			collision_radius_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-collision_force_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			collision_force_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-spring_damping_factor")) {
			if (++i >= argc) { Usage(argv[0]); }
			spring_damping_factor = atof(argv[i]);
		} else if (!strcmp(argv[i], "-random_number_seed")) {
			if (++i >= argc) { Usage(argv[0]); }
			random_number_seed = atoi(argv[i]);
                } else if (!strcmp(argv[i],"-openCL_dir")) {
                        if (++i >= argc) { Usage(argv[0]); }
                        openCL_file_dir = argv[i];
                } else if (!strcmp(argv[i],"-timestep_ns")) {
                        if (++i >= argc) { Usage(argv[0]); }
                        time_step_nanoseconds = atof(argv[i]);
                } else if (!strcmp(argv[i],"-burst")) {
                        if (++i >= argc) { Usage(argv[0]); }
                        tauIterations = atoi(argv[i]);
                } else if (!strcmp(argv[i],"-repeat")) {
                        if (++i >= argc) { Usage(argv[0]); }
                       readIterations = atoi(argv[i]);
                } else if (!strcmp(argv[i],"-verbose")) {
                        g_verbose = true;
                } else if (!strcmp(argv[i],"-CPU")) {
                        useGPU = false;
		} else {
			Usage(argv[0]);
		}
	}

        // Compute the derived values that we feed to the simulator.
        double  time_step_seconds = nanoseconds_to_seconds(time_step_nanoseconds);
        double  chain_length_meters = microns_to_meters(chain_length_microns);
        double  node_separation_meters = microns_to_meters(mass_separation_microns);
	double	mass_kg = dna_mass_in_kg(node_separation_meters);
        double  temperature_kelvin = Celsius_to_Kelvin(temperature_Celsius);
        double  viscosity_Pascal_seconds = centiPoise_to_Pascal_seconds(viscosity_centiPoise);
        double  dna_radius_meters = nanometers_to_meters(dna_radius_nanometers);
	double	effective_damping_radius = damping_radius_factor * node_separation_meters;

	double	spring_constant = dna_tensile_spring_constant(
                        node_separation_meters,
                        gigaPascal_to_Pascal(modulus_gigaPascal),
                        dna_radius_meters
                        );
	double	hinge_constant = dna_bending_spring_constant(
                        node_separation_meters,
                        gigaPascal_to_Pascal(modulus_gigaPascal),
                        dna_radius_meters
                        );
	double	random_force = Brownian_force_equivalent(
                        node_separation_meters,
                        temperature_kelvin,
                        viscosity_Pascal_seconds,
                        time_step_seconds
                        );
	double	mass_damping = mass_damping_equivalent(
			viscosity_Pascal_seconds,
			effective_damping_radius,
			mass_kg);
	double	mass_radius = collision_radius_factor * (node_separation_meters / 2.0);
        double  collision_spring_constant = spring_constant * collision_force_factor;

        // Run the test, given the parameters.
        if (g_verbose) {
            printf("Starting test\n");
        }
	bool ret = (0 == test_configuration(
		chain_length_meters, node_separation_meters, mass_radius, mass_kg, mass_damping
                , spring_constant, hinge_constant
		, time_step_seconds, random_force, temperature_kelvin
		, viscosity_Pascal_seconds, spring_damping_factor
                , modulus_gigaPascal, dna_radius_meters
		, random_number_seed, openCL_file_dir
		, tauIterations, readIterations, useGPU
                , collisionScheme, collision_spring_constant));
	if (ret) {
		printf("Success!\n");
		return 0;
	} else {
		printf("Failure.\n");
		return 3;
	}
}
