/*
Model we want to test:
        The radius of gyration is the root mean of the squared
        distance from the center of mass to each point on the chain.

	The expected end-to-end distance is sqrt(Lp*Lc), where Lc
	is the contour length (the distance traced along the chain
	from one end to the other) and Lp is the persistence length
	(about 50nm for B-form DNA).

	The expected radius of gyration is sqrt(Lp*Lc/6).

        Rg = sqrt(<R(from center of mass)^2>)
        Lp = persistence length.  Lc = contour length.
        Lp*Lc = <R(end to end)^2>,  Also, Lp = E*I / (kB*T),
        which turns out to be when the correlation of the cosine of the angles goes below 1/e.
        It turns out that Rg^2 = Lp*Lc / 6.
	We want to compute Rg using the two-endpoints method.

        We'd like to have Lp much smaller than Lc.
        Lp = 50nm for B-form DNA, so we want Lc to be >= 1 micron.
        The end-to-end expected is 223nm, which is much less than this.
        Also try 10 microns to see if it works.
        We need to compute for a while to get this average,
        and let the system settle in before starting the average.
        We'll look at the time average over time and see when it settles down.
        It could take a lot of steps.
*/

#include <stdio.h>
#include <math.h>
#include <vector>

typedef std::vector<double> POINT;
typedef std::vector<POINT>  POINTS_FOR_A_GIVEN_TIME;

static double dist2(const POINT &a, const POINT &b)
{
    double dx = a[0] - b[0];
    double dy = a[1] - b[1];
    double dz = a[2] - b[2];
    return dx*dx + dy*dy + dz*dz;
}

static double dist(const POINT &a, const POINT &b)
{
    return sqrt( dist2(a,b) );
}

static POINT find_mean(const POINTS_FOR_A_GIVEN_TIME &pts)
{
    // If we have no points, return zero.
    POINT mean;
    mean.push_back(0);
    mean.push_back(0);
    mean.push_back(0);
    if (pts.size() == 0) {
	return mean;
    }

    // Compute the mean location of the all the points
    size_t i;
    for (i = 0; i < pts.size(); i++) {
	mean[0] += pts[i][0];
	mean[1] += pts[i][1];
	mean[2] += pts[i][2];
    }
    mean[0] /= pts.size();
    mean[1] /= pts.size();
    mean[2] /= pts.size();

    return mean;
}

// This computes the radius of gyration as the mean squared distance
// from the mean of all of the points in the polymer.
static double radius_of_gyration(const POINTS_FOR_A_GIVEN_TIME &pts)
{
    // If we don't have any points, we return a negative radius,
    // which indicates an error.
    if (pts.size() == 0) { return -1; }

    // Compute the mean location of the points.
    POINT mean = find_mean(pts);

    // Compute the average squared distance from the mean
    size_t i;
    double sum = 0.0;
    for (i = 0; i < pts.size(); i++) {
	sum += dist2(mean, pts[i]);
    }
    return sqrt( sum / pts.size() );
}

// This computes the squared end-to-end distance of the polymer.
static double squared_end_to_end(const POINTS_FOR_A_GIVEN_TIME &pts)
{
    // If we don't have at least two points, return -1, which
    // indicates an error.
    if (pts.size() < 2) { return -1; }

    return dist2(pts[0], pts[pts.size()-1]);
}

// This computes the contour length of the polymer.
static double contour_length(const POINTS_FOR_A_GIVEN_TIME &pts)
{
    // If we don't have at least two points, return -1, which
    // indicates an error.
    if (pts.size() < 2) { return -1; }

    double length = 0;
    size_t i;
    for (i = 1; i < pts.size(); i++) {
	length += dist(pts[i-1], pts[i]);
    }

    return length;
}


int main(int argc, const char *argv[])
{
    // Get the input and output files (for now, just standard in and out)
    FILE *infile = stdin;
    FILE *outfile = stdout;

    // We repeat reading in blocks of points, each of which is preceded
    // by a line that says "Time T", where T is a floating-point number.
    // Then there are a number of lines of X Y Z which are floating-point
    // coordinates. Then there is a blank line after each block of points.
    std::vector<POINTS_FOR_A_GIVEN_TIME>    times;

    // we want to also record the actual times at which these were calculated
    // for accurate comparison between different runs (since SaveTrajectory
    // doesn't actually save every 0.1 seconds but only close to it)
    std::vector<double> t_vals;

    char    line[4096];
    line[sizeof(line)-1] = '\0';    // Ensure null-termination
    while (true) {
        // Read and parse the "Time" line.
        if (fgets(line, sizeof(line)-1, infile) == NULL) {
            // Done reading lines, we got them all.
            break;
        }
        double t;
        if (sscanf(line, "Time %lf", &t) <= 0) {
            fprintf(stderr, "Bad Time line: %s\n", line);
            return(-1);
        }
        // push the time to back of t_vals
        t_vals.push_back( t );

        // Read and parse point lines, adding them to our time.
        POINTS_FOR_A_GIVEN_TIME points;
        times.push_back(points);
        while (true) {
            // Read and parse the next line, which may be a points line.
            if (fgets(line, sizeof(line)-1, infile) == NULL) {
                // Done reading lines, we got them all.
                break;
            }
            double x,y,z;
            if (sscanf(line, "%lf %lf %lf", &x, &y, &z) != 3) {
                // No more points for this line
                break;
            }
            POINT p;
            p.push_back(x);
            p.push_back(y);
            p.push_back(z);
            times.back().push_back(p);
        }
    }
    fclose(infile);

    // Compute the radius of gyration by looking at the motion of the points
    // in the chain over time.  Print out the radius of gyration at each time
    // step.
    size_t i;
    fprintf(outfile, "Time, RoG, EndToEnd, Contour\n");
    for (i = 0; i < times.size(); i++) {
	fprintf(outfile, "%lf, %lf, %lf, %lf\n",
                t_vals[i],
		radius_of_gyration(times[i]),
		sqrt(squared_end_to_end(times[i])),
		contour_length(times[i]) );
    }

    fclose(outfile);
    return 0;
}

